/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.Layout;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.FontUIResource;
import javax.swing.plaf.metal.*; // Java Look & Feel


/**
 * Die Klasse Application verwaltet das Applikationsfenster, in welchem alle
 * Fenster einer Hoverball-Applikation dargestellt werden.
 * <p>
 * Alle Hoverball-Fenster werden als <i>internal frames</i> (Klasse JInternalFrame)
 * repräsentiert. Einzig das Applikationsfenster ist ein <i>top-level frame</i>
 * (Klasse JFrame), das alle Hoverball-Fenster auf seinem Desktop
 * <a href="#desktop">desktop</a>
 * (Klasse JDesktopPane) darstellt.
 * Beachte hierzu:
 *
 * <menu>
 * <li>Die Standard-Fenster (die Fenster von
 *     <a href="Simulator.html">Simulator</a>
 *     und
 *     <a href="Controller.html">Controller</a>
 *     samt Screen, sowie die
 *     <a href="Hovlet.html">Hovlet</a>-Browser)
 *     werden automatisch auf dem Desktop abgelegt.
 *     Lediglich selbst definierte Fenster, die im Applikationsfenster von Hoverball
 *     angezeigt werden sollen, müssen mit
 *     <p>
 *     <tt>Application.desktop.add(</tt>...<tt>);</tt>
 *     <p>
 *     auf dem Desktop hinzugefügt werden.
 *     <p></li>
 * <li>Die Position und die Größe des Applikationsfensters kann mit den Methoden
 *     <p>
 *     <tt>Application.setBounds(</tt>...<tt>);</tt>
 *     </p>
 *     festgelegt werden. Soll nur ein einziges Fenster in einer Hoverball-Applikation
 *     angezeigt werden (wenn etwa nur der Simulator gestartet wird),
 *     kann das Applikationsfenster mit
 *     <p>
 *     <tt>Application.setBounds();</tt>
 *     </p>
 *     direkt an das darzustellende Fenster angepasst werden. Das Applikationsfenster
 *     übernimmt dann sozusagen die Attribute des darzustellenden inneren Fensters.
 *     <p></li>
 * <li>Das Applikationsfenster öffnet sich automatisch, sobald ein Standard-Fenster
 *     sichtbar gemacht wird. Bei Bedarf kann das Applikationsfenster mit
 *     <p>
 *     <tt>Application.show();</tt>
 *     </p>
 *     aber auch "manuell" geöffnet werden.
 *     <p></li>
 * <li>Das automatische Sichtbarmachen kann mit dem Befehl
 *     <p>
 *     <tt>Application.hide();</tt>
 *     </p>
 *     zu Beginn der Anwendung unterbunden werden.
 *     <p></li>
 * </menu>
 */
public abstract class Application
{
   private static final String TITLE = "Hoverball"; // "Hoverball "+About.VERSION;

   // ---- AWEFUL BAD WORKAROUND FOR SYNC CONNECTING TO HOVERBALL SERVER ------

   static String servertext_with_hash = null;

   static java.util.Vector<Browser.ConnectionActionListener> listeners = new java.util.Vector<Browser.ConnectionActionListener>();

	/**
	 * Verbindet die Hoverball Application mit der angegebenen Adresse.
	 *
	 * @param address
	 *            Netzwerk-Adresse
	 * @return <b><code>true</code></b>, falls Verbindung erfolgte
	 */
   public static void connect (String address) {
	  SwingUtilities.invokeLater(new Runnable() { // ouch!! :-(
	     public void run () { 
	        new Thread(new Runnable() {
		       public void run () { try { Thread.sleep(200); } catch(InterruptedException e) {}
      servertext_with_hash = null;
      if (address == null || address.equals("")) return;
      for (Browser.ConnectionActionListener listener : listeners) listener.connect(address);
		       }
	        }).start();
		 }
	  });
   }

	/**
	 * Trennt die Verbindung der Hoverball Application.
	 */
   public static void disconnect () {
      SwingUtilities.invokeLater(new Runnable() {
		 public void run () {                     
      for (Browser.ConnectionActionListener listener : listeners) listener.disconnect();
		 }
      });
   }


   // ---- Hoverball Bounds ---------------------------------------------------

   private static final int MINIMUM_WIDTH  = 640;
   private static final int MINIMUM_HEIGHT = 480;

   private static String title = "";
   private static Rectangle bounds = new Rectangle();
   private static Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();

   /**
    * Liefert den Bereich des Applikationsfensters.
    *
    * @return den Bereich
    */
   public static Rectangle getBounds ()
   {
      return (bounds == null)? null : new Rectangle(bounds);
   }

   /**
    * Passt das Applikationsfenster an ein allein darzustellendes Fenster an.
    * <p>
    * Wenn das Applikationsfenster geöffnet wird, werden die Attribute
    * des ersten sichtbaren Fensters in der Liste des Desktops übernommen.
    */
   public static void setBounds ()
   {
      Application.bounds = null;
   }

   /**
    * Definiert den Bereich des Applikationsfensters.
    *
    * @param ratio Bruchteil des Bildschirm-Bereichs zum Gesamt-Bildschirm (zwischen 0 und 1)
    */
   public static void setBounds (double ratio)  // 0..1
   {
      Dimension screen = screensize;
      setBounds(new Dimension((int)(screen.width*ratio),(int)(screen.height*ratio)));
   }

   /**
    * Definiert den Bereich des Applikationsfensters.
    *
    * @param size Größe des Fensters
    */
   public static void setBounds (Dimension size)
   {
      setBounds(Layout.center(new Rectangle(screensize),size));
   }

   /**
    * Definiert den Bereich des Applikationsfensters.
    *
    * @param bounds Bildschirm-Bereich
    */
   public static void setBounds (Rectangle bounds)
   {
      Rectangle r = bounds.union(Layout.center(bounds,new Dimension(MINIMUM_WIDTH,MINIMUM_HEIGHT)));
      Application.bounds = new Rectangle(Layout.fit(new Rectangle(screensize),r));
   }

   /**
    * Liefert den Titel des Applikationsfensters.
    *
    * @return den Titel
    */
   public static String getTitle ()
   {
      return ""+title;
   }

   /**
    * Definiert den Titel des Applikationsfensters.
    *
    * @param title Titel
    */
   public static void setTitle (String title)
   {
      if (title == null || title.length() == 0) {
         Application.title = "";
         frame.setTitle(TITLE);
      } else {
         Application.title = title;
         frame.setTitle(title);
//         frame.setTitle(title+" - "+TITLE);
      }
   }


   // -------------------------------------------------------------------------

   /**
    * Desktop des Applikationsfensters
    */
   public static final JDesktopPane desktop = new JDesktopPane();

   /**
    * Frame des Applikationsfensters
    */
   public static final JFrame frame = new JFrame(TITLE);
   
   private static int FRAME_SUPPLEMENT_L = 4;
   private static int FRAME_SUPPLEMENT_R = 4;
   private static int FRAME_SUPPLEMENT_T = 32;
   private static int FRAME_SUPPLEMENT_B = 4;

   private static boolean opened = false;

   static {
	   /*
	   for ( java.util.Enumeration e = UIManager.getDefaults().keys(); e.hasMoreElements(); )
	   {
	       Object key   = e.nextElement();
	       Object value = UIManager.get( key );

	       if ( value instanceof Font )
	       {
	           Font f = (Font) value;

	           UIManager.put( key, new FontUIResource( f.getName(), f.getStyle(), 14 ) );
	       }
	   }
	   */

	  //if (Applet.applet == null)
      {
      setBounds(Layout.getApplicationBounds());

      Layout.loadResources(frame);

      try { // Fix Ocean Theme
         Class ocean = Class.forName("javax.swing.plaf.metal.OceanTheme");
         if (ocean != null)
            MetalLookAndFeel.setCurrentTheme((MetalTheme)ocean.newInstance());
      } catch(Exception e) {}

      //javax.swing.UIManager.put("Label.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.PLAIN, 20)));
      //javax.swing.UIManager.put("Button.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.BOLD, 20)));
      //javax.swing.UIManager.put("TextField.font", new javax.swing.plaf.FontUIResource(new Font("Dialog", Font.PLAIN, 20)));
      /*
      for (Entry<Object, Object> entry : javax.swing.UIManager.getDefaults().entrySet()) {
    	    Object key = entry.getKey();
    	    Object value = javax.swing.UIManager.get(key);
    	    if (value != null && value instanceof javax.swing.plaf.FontUIResource) {
    	        javax.swing.plaf.FontUIResource fr=(javax.swing.plaf.FontUIResource)value;
    	        javax.swing.plaf.FontUIResource f = new javax.swing.plaf.FontUIResource("Noto Sans",fr.getStyle(),fr.getSize()+4); // Font.BOLD,16); //(fr.getFamily(), fr.getStyle(), 20);
    	        javax.swing.UIManager.put(key, f);
    	    }
    	}
      */
      desktop.setBackground(Layout.colorDesktop);

      if (Layout.icon_png != null) frame.setIconImage(new ImageIcon(Layout.icon_png).getImage());
      frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
      frame.getContentPane().add(desktop);

      frame.addWindowListener(new WindowAdapter() {
         public void windowClosed (WindowEvent e) { Application.exit(); }
      });
      }
   }

   /**
    * Schließt das Applikationsfenster und beendet das Programm.
    */
   public static void exit ()                                   // exit -> dispose
   {
      synchronized(desktop) { if(exited) return; exited = true; }

      JInternalFrame[] frames = desktop.getAllFrames();
      for (int i=0; i<frames.length; ++i) frames[i].dispose();  // dispose -> close
      frame.dispose();
      
      if (Simulator.headless_simulator != null) Simulator.headless_simulator.close();
   }
   private static boolean exited = false;

   /**
    * Zeigt das Applikationsfenster an (expliziter Aufruf nicht notwendig).
    */
   public static void show ()
   {
      if (opened) return;
      opened = true;

      if (bounds != null) {
         frame.setBounds(bounds);
         frame.setVisible(true); // .show(); [deprecated, since 1.5]
      } else {
         JInternalFrame[] frames = desktop.getAllFrames();
         JInternalFrame f = null;
         for (int i=0; i<frames.length; ++i) if (frames[i].isVisible()) { f = frames[i]; break; }
         if (f == null) return;

         Rectangle b = f.getBounds();  // expand bounds (toplevel frame border)
         b.x -= FRAME_SUPPLEMENT_L;
         b.y -= FRAME_SUPPLEMENT_T;
         b.width  += FRAME_SUPPLEMENT_L + FRAME_SUPPLEMENT_R;
         b.height += FRAME_SUPPLEMENT_T + FRAME_SUPPLEMENT_B;

         frame.setBounds(Layout.fit(new Rectangle(screensize),b));
         frame.setResizable(f.isResizable());
         frame.setVisible(true); // .show(); [deprecated, since 1.5]

         f.setFrameIcon(null);
         f.setBounds(desktop.getBounds());
         f.setMaximizable(true);
         try { f.setMaximum(true); } catch(java.beans.PropertyVetoException e) {}
         f.addInternalFrameListener(new InternalFrameAdapter() {
            public void internalFrameClosed(InternalFrameEvent e) { Application.exit(); }
         });
         desktop.setSelectedFrame(f);
      }
   }

   /**
    * Versteckt das Applikationsfenster. Wird dieser Befehl zu Beginn aufgerufen, unterbleibt das automatische Anzeigen
    * und die Application bleibt versteckt.
    */
   public static void hide ()
   {
	  Layout.auto_show = false;;

      if (!opened) return;
      opened = false;

	  frame.setVisible(false);
   }
}
