/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

import java.util.Vector;


// ---- The Hovlet Browser ------------------------------------- [Browser] ----


class Browser extends MyInternalFrame
{
   // ---- class Browser ------------------------------------------------------

   private static final String HOVLET_SEPARATOR_1  = " > ";
   private static final String HOVLET_SEPARATOR_2  = "  >  ";

   private static final String BUTTON_CONNECT    = "-> <-";
   private static final String BUTTON_DISCONNECT = "--|--";

   private static final String TOOLTIP_CONNECT   = "Connect";
   private static final String TOOLTIP_DISCONNECT= "Disconnect";

   private static final String STATUS_CONNECTION_FAILED = "Connection failed";
   private static final String STATUS_DISCONNECTING     = "Disconnecting";
   private static final String STATUS_DISCONNECTED      = "Disconnected";
   private static final String STATUS_CONNECTING        = "Connecting";
   private static final String STATUS_CONNECTED         = "Connected";

   private static final int REMEMBER_HOSTS = 5;

   // ----

   private final JMenuBar  menubar;                  // Browser's Menu Bar

   private final JMenu     menuConnect;
   private final JMenuItem menuDisconnect;
   private final JMenu     menuShow;

   private final JTextField textfield;               // Server TextField
   private final JButton button;                     // Connect Button
   private final JComboBox combo;                    // Hovlet Choice
   private final JButton up;                         // Hovlet Up Button
   private final JLabel statusline;                  // Status Line

   private final JPanel icon;                        // Hovlet Icon
   private final CardLayout icon_manager;            // Hovlet Icon's CardLayout-Manager
   private final JPanel main;                        // Hovlet Area
   private final CardLayout main_manager;            // Hovlet Area's CardLayout-Manager

   private Hovlet hovlet = null;                     // current Hovlet
   private Vector hovlets = new Vector(10,5);        // all Hovlets (with sub-hovlets)


   // ---- Menu's ConnectListener ---------------------------------------------

   private class ConnectListener
      implements ActionListener, Runnable
   {
      private String host;
      ConnectListener (String host) { this.host = host; }

      public void run () {
          if (host == null) hovlet.disconnect();
          else hovlet.connect(host);
      }
      public void actionPerformed (ActionEvent e)
      {
//APPLET THREAD BUG//
    	  new Thread(this,"Hoverball Connect/Disconnect[Menu]").start(); //new Thread(Hovlet.group,this,"Connect/Disconnect[Menu]").start();
      }
   }

   // ---- Textfield's & Button's ActionListener ------------------------------

   /*private*/ class ConnectionActionListener
      implements ActionListener, Runnable
   {
      public ConnectionActionListener() { Application.listeners.add(this); }

      public void connect (String address) {
         if (hovlet == null) return; // should not happen, but ... :-(
         // Neu: Connect/Disconnect-Button wirkt sich auf die komplette Hovlet-Struktur ab dem Root.Hovlet aus!
         Hovlet root = hovlet; while (root.parent != null) root = root.parent;
         root.connect(address);
      }

      public void disconnect () {
         if (hovlet == null) return; // should not happen, but ... :-(
         // Neu: Connect/Disconnect-Button wirkt sich auf die komplette Hovlet-Struktur ab dem Root.Hovlet aus!
         Hovlet root = hovlet; while (root.parent != null) root = root.parent;
         root.disconnect();
      }

      public void run () {
         if (hovlet == null) return; // should not happen, but ... :-(
         // Neu: Connect/Disconnect-Button synchron für die ganze Application! ("AWEFUL BAD WORKAROUND")
         if (hovlet.state == Hovlet.DISCONNECTED) Application.connect(textfield.getText());
         else if (hovlet.state == Hovlet.CONNECTED) Application.disconnect();
      }

      public void actionPerformed(ActionEvent e) {
//APPLET THREAD BUG//
    	  new Thread(this,"Hoverball Connect/Disconnect[Button]").start(); //new Thread(Hovlet.group,this,"Connect/Disconnect[Button]").start();
      }
   }

   // ---- Hovlet Choice's ItemListener ---------------------------------------

   private class HovletComboListener
      implements ItemListener
   {
      public void itemStateChanged (ItemEvent e)
      {
         if (e.getStateChange() == e.DESELECTED) return;
         show((Hovlet)hovlets.elementAt(combo.getSelectedIndex()));
      }
   }

   // ---- Hovlet Up Button's ActionListener ----------------------------------

   private class HovletUpListener
      implements ActionListener
   {
      public void actionPerformed(ActionEvent e) {
         if(hovlet.parent==null) return;
         show(hovlet.parent);
      }
   }

   // ---- Hovlet Show-Menu's ItemListener ------------------------------------

   private class HovletMenuListener
      implements ActionListener
   {
      private final int n;
      HovletMenuListener (int n) { this.n = n; }

      public void actionPerformed (ActionEvent e)
      {
         show((Hovlet)hovlets.elementAt(n));
      }
   }

   // ---- private class ComboRenderer (list cell renderer) -------------------

   private class ComboRenderer
         implements ListCellRenderer
   {
      JLabel label = new JLabel();

      ComboRenderer () { label.setFont(Layout.fontCombo); label.setOpaque(true); }

      public Component getListCellRendererComponent(
            JList list,
            Object value,            // value to display
            int index,               // cell index
            boolean isSelected,      // is the cell selected
            boolean cellHasFocus)    // the list and the cell have the focus
      {
         label.setText(""+value);
         return Layout.finishListCell(label,list,isSelected);
      }
   }


   // ---- Constructor --------------------------------------------------------


   Browser (Hovlet hovlet)
   {
      super("",true,true,false,false);

      // ---- Window Listener: Application.exit on close on every Hovlet-Browser!

      addInternalFrameListener(new InternalFrameAdapter() {
         public void internalFrameClosing(InternalFrameEvent e) { Application.exit(); }
      });

      // ---- Frame...

      setSize(Layout.BROWSER_WIDTH,Layout.BROWSER_HEIGHT);

      // ---- Components...

      Container j = getContentPane();
      j.setLayout(new GridBagLayout());

      int padx = 8;
      int pady = 6;

      JPanel header = new JPanel(new GridBagLayout()); header.setOpaque(false);

      JPanel textfield_wrapper = new JPanel(new GridBagLayout()); textfield_wrapper.setOpaque(true);
      textfield_wrapper.add(textfield = new JTextField(),
                            new Constraints(0,0,1,1,1.0,0.0,new Insets(0,0,0,0)));

      header.add(new JLabel("Server:"), // { public boolean isFocusTraversable() { return true; }},
                 new Constraints(0,1,1,1,0.0,0.0,new Insets(0,0,0,padx)));
      header.add(textfield_wrapper,
                 new Constraints(1,1,1,1,1.0,0.0,new Insets(0,0,0,0)));
      header.add(button = new MyButton("",true,50),
                 new Constraints(2,1,1,1,0.0,0.0,new Insets(0,padx,0,0)));
      if(Layout.TOOLTIPS) textfield.setToolTipText("<server>:<port>");

      JPanel combo_wrapper = new JPanel(new GridBagLayout()); combo_wrapper.setOpaque(false);
      combo_wrapper.add(combo = new MyComboBox(),
                        new Constraints(0,0,1,1,1.0,0.0,new Insets(0,0,0,0)));
      combo_wrapper.add(up = new MyButton("^",true,21),
                        new Constraints(1,0,1,1,0.0,0.0,new Insets(0,4,0,0)));
      combo.setRenderer(new ComboRenderer());
      combo.setMaximumRowCount(10);
      JPanel icon_wrapper = new JPanel(new GridBagLayout()); icon_wrapper.setOpaque(false);
      icon_wrapper.add(icon = new JPanel(icon_manager = new CardLayout()),
                       new Constraints(0,0,1,1,1.0,1.0,new Insets(2,1,2,1)));
      //if(Layout.TOOLTIPS) combo.setToolTipText("");
      if(Layout.TOOLTIPS) up.setToolTipText("Up");

      up.setFont(up.getFont().deriveFont(Font.PLAIN)); // Symbol Font non-italic!

      header.add(new JLabel("Hovlet:"),
                 new Constraints(0,3,1,1,0.0,0.0,new Insets(1,0,0,padx)));
      header.add(combo_wrapper,
                 new Constraints(1,3,1,1,1.0,0.0,new Insets(1,0,0,0)),1);  // insert at 1!
      header.add(icon_wrapper,
                 new Constraints(2,3,1,1,0.0,0.0,new Insets(1,padx,0,0)));

      j.add(header,
          new Constraints(0,1,1,1,1.0,0.0,new Insets(pady,padx,pady,padx)));


      JPanel status = new JPanel(new GridBagLayout()); status.setOpaque(false);
      status.add(statusline = new JLabel(),
                 new Constraints(0,1,1,1,1.0,0.0,new Insets(2,padx,2,padx)));

      j.add(status,
          new Constraints(0,3,1,1,1.0,0.0,new Insets(0,0,0,0)));


      ActionListener connect = new ConnectionActionListener();
      textfield.addActionListener(connect);
      button.addActionListener(connect);

      up.addActionListener(new HovletUpListener());

      // ---- Main Panel (Vertical Filler!)

      main = new JPanel(main_manager = new CardLayout());

      j.add(main,
          new Constraints(0,2,1,1,1.0,1.0,new Insets(0,0,0,0)));


      // ---- Menu "Hovlet"

      JMenu menuHovlet = new JMenu("Hovlet");
         menuConnect = new JMenu("Connect");
            JMenuItem menuLocalhost = new JMenuItem("localhost"+hoverball.net.Address.PORT+Simulator.DEFAULT_PORT);
         menuDisconnect = new JMenuItem("Disconnect");
         JMenuItem menuClose = new JMenuItem("Close");
      menuShow = new JMenu("Show");

      menuConnect.add(menuLocalhost);

      menuHovlet.add(menuConnect);
      menuHovlet.add(menuDisconnect); menuDisconnect.setEnabled(false);
      menuHovlet.addSeparator();
      menuHovlet.add(menuShow);
      menuHovlet.addSeparator();
      menuHovlet.add(menuClose);

      menuLocalhost.addActionListener(new ConnectListener(menuLocalhost.getText()));
      menuDisconnect.addActionListener(new ConnectListener(null));
      menuClose.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { dispose(); }
      });


      // ---- Menu Bar

      menubar = new JMenuBar();
      menubar.add(menuHovlet);

      adaptFontInMenu();

      setJMenuBar(menubar);


      // ---- Init

      Hovlet root = hovlet;  // Search Root Hovlet
      while (root.parent_ != null) root = root.parent_;
      add(root);             // recursive addition!

final Hovlet root_ = root;                   // !!
final Browser this_ = this;
SwingUtilities.invokeLater(new Runnable() {  // !! WORKAROUND: JAVA 1.8 NEEDS THIS
   public void run () {                      // !!
      if (root_.location != null) setLocation(hovlet.location);  // "soft" location property
   }                                         // !!
});                                          // !!

      root_.browse(this_);

      // ---- Combo listener last... (otherwise: show() while adding!)

      combo.addItemListener(new HovletComboListener());

   }

   void adaptFontInMenu()
   {
      setFontInMenuBar(this.menubar,Layout.fontMenu);
   }

   private static void setFontInMenuBar (JMenuBar menubar, Font font)
   {
      for(int k=0; k<menubar.getMenuCount(); ++k)
         setFontInMenu(menubar.getMenu(k),font);
   }

   private static void setFontInMenu (JMenuItem item, Font font)
   {
      item.setFont(font);

      if(item instanceof JMenu) {
         JMenu menu = (JMenu)item;
         Component[] submenus = menu.getMenuComponents();
         for (int k = 0; k < submenus.length; ++k)
            if (submenus[k] instanceof JMenuItem)
               setFontInMenu((JMenuItem)submenus[k], font);
      }
   }

   private void add (Hovlet hovlet)  // recursive addition
   {
      main.add(hovlet.main,"Center");                              // add Main Panel
      main_manager.addLayoutComponent(hovlet.main,""+hovlets.size());
      icon.add(hovlet.icon,"Center");                              // add Icon Panel
      icon_manager.addLayoutComponent(hovlet.icon,""+hovlets.size());
                                                                   // add to Menu
      JMenuItem item = new JMenuItem(hovlet.title(HOVLET_SEPARATOR_2));
      item.addActionListener(new HovletMenuListener(hovlets.size()));
//      if (hovlets.size() < 10) item.setMnemonic(KeyEvent.VK_0+hovlets.size());
      menuShow.add(item); adaptFontInMenu();

      hovlets.addElement(hovlet);                                  // add to internal List
      hovlet.browser = this;                                       // link to Browser

      // Gleichnamige Combo-Items werden nicht unterschieden!
      // eigentlich: combo.addItem(hovlet.title(HOVLET_SEPARATOR_2));             // add to Choice
      //
      String titlex = hovlet.title(HOVLET_SEPARATOR_2);
      do {
         boolean contains = false;
         for(int i=0; i<combo.getItemCount(); ++i)
            if(combo.getItemAt(i).equals(titlex)) contains = true;
         if(contains) { titlex += " "; continue; } // füge Jokerzeichen hinzu!
      } while(false);
      combo.addItem(titlex);                                       // add to Choice
      //

      for (int i=0; i<hovlet.hovlets_.length; ++i) add(hovlet.hovlets_[i]);
   }

   public void dispose ()  // dispose -> close
   {
	   close();
      super.dispose();
   }

   private boolean closed = false;
   void close () // war "synchronized" => Bug! (Hängenbleiben bei exit)
   {
      synchronized(this) { if (closed) return; closed = true; }

      Hovlet root = (Hovlet)hovlets.firstElement();                // start with Root Hovlet!
      root.close();
   }


   // ---- Hovlet Commands ----------------------------------------------------

   synchronized void show (final Hovlet hovlet)
   {
      final Browser this_ = this;
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      if (this_.hovlet != hovlet) {
         int n = hovlets.indexOf(hovlet);                             // get Hovlet Index

         main_manager.show(main,""+n);                                // Main Panel
         icon_manager.show(icon,""+n);                                // Icon
         combo.setSelectedIndex(n);                                   // Choice
         up.setEnabled(hovlet.parent != null);                        // Up
         setTitle(Layout.getTitle(hovlet.title));                     // Title
         textfield.setText(hovlet.servertext);                        // Server-Text
         setConnectionState(hovlet.state);
         if (this_.hovlet != null)                                    // MenuBar
            move(menubar,this_.hovlet.menubar);
         move(hovlet.menubar,menubar); adaptFontInMenu();

         this_.hovlet = hovlet;                                       // updated.
      }
      status(hovlet,null);                                            // update status
      show();                                                         // show Frame!
      
      Application.desktop.setSelectedFrame(this_);
         }
      });
   }

   private Vector connecting = new Vector();  // remember connecting Hovlets for "Connection failed."

   synchronized void refresh (final Hovlet hovlet)
   {
      final Browser this_ = this;
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      if (hovlet.state == Hovlet.CONNECTING)       // xxx-ing: set Status Line!
         status(null,STATUS_CONNECTING+" "+hovlet.title(HOVLET_SEPARATOR_1)+"...");
      else if (hovlet.state == Hovlet.DISCONNECTING)
         status(null,STATUS_DISCONNECTING+" "+hovlet.title(HOVLET_SEPARATOR_1)+"...");
      else if (hovlet.state == Hovlet.DISCONNECTED && connecting.contains(hovlet))
         status(null,STATUS_CONNECTION_FAILED+"."); 
      else
         status(null,null);

      if (hovlet.state == Hovlet.CONNECTING) connecting.addElement(hovlet);
      else connecting.removeElement(hovlet);

      if (hovlet.state == Hovlet.DISCONNECTED)     //  disable Disconnect-Menu
         menuDisconnect.setEnabled(false);
      else if (hovlet.state == Hovlet.CONNECTED) { //  enable Disconnect-Menu
         menuDisconnect.setEnabled(true);

         String host = hovlet.servertext;
         boolean found = false;                    //  add host to Menu
         for (int i=0; i<menuConnect.getItemCount(); ++i) {
            JMenuItem m = menuConnect.getItem(i);
            if (m != null && m.getText() != null &&  // m.getText()!=null does not happen.
                m.getText().equals(""+host)) { found = true; break; }
         }
         if (!found) {
            JMenuItem item = new JMenuItem(""+host);
            item.addActionListener(new ConnectListener(""+host));
            if (menuConnect.getItemCount() == 1)
               menuConnect.insertSeparator(0);
            if (menuConnect.getItemCount() == REMEMBER_HOSTS+2)
               menuConnect.remove(REMEMBER_HOSTS-1);
            menuConnect.insert(item,0); adaptFontInMenu();
         }
      }

      if (this_.hovlet != hovlet) return;      // not shown => no Header-refresh.

      textfield.setText(hovlet.servertext);                  // Server-Text
      setConnectionState(hovlet.state);
      repaint();
         }
      });
   }

   private void move (JMenuBar source, JMenuBar destination)   // move Menus recursively!
   {
      if (source.getMenuCount() < 2) return;                 // (except "Hovlet")

      JMenu menu = source.getMenu(source.getMenuCount()-1);
      source.remove(menu);

      move(source,destination);

      destination.add(menu);
   }

   private synchronized void setConnectionState (final int state)
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      if (state == Hovlet.DISCONNECTED) {
         button.setText(BUTTON_CONNECT); if(Layout.TOOLTIPS) button.setToolTipText(TOOLTIP_CONNECT);
         button.setEnabled(true);
         textfield.setEnabled(true);
         menuDisconnect.setEnabled(false);
      }
      else if (state == Hovlet.CONNECTING || state == Hovlet.DISCONNECTING) {
         button.setText(BUTTON_DISCONNECT); if(Layout.TOOLTIPS) button.setToolTipText(TOOLTIP_DISCONNECT);
         button.setEnabled(false);
         textfield.setEnabled(false);
         menuDisconnect.setEnabled(false);
      }
      else if (state == Hovlet.CONNECTED) {
         button.setText(BUTTON_DISCONNECT); if(Layout.TOOLTIPS) button.setToolTipText(TOOLTIP_DISCONNECT);
         button.setEnabled(true);
         textfield.setEnabled(false);
         menuDisconnect.setEnabled(true);
      }
      repaint();
         }
      });
   }

   synchronized void status (final Hovlet hovlet, String status)
   {
      final Browser this_ = this;
      final String status_ = status;
      SwingUtilities.invokeLater(new Runnable() {
         String status = status_;
         public void run () {
      if (this_.hovlet == null) return;                     // before showing
      if (this_.hovlet != hovlet && hovlet != null) return; // Hovlet mismatch (hovlet==null: always!)

      if (status == null) status = this_.hovlet.status;     // null? read Hovlet's status

      if (status != null)  // if status defined: set status
         statusline.setText(status);
      else {               // else: standard-status
         int connected = 0;
         for (int i=0; i<hovlets.size(); ++i) {
            Hovlet hovlet_ = (Hovlet)hovlets.elementAt(i);
            if (hovlet_.state == Hovlet.CONNECTED) ++ connected;
         }
         if (connected == 0)
            statusline.setText(STATUS_DISCONNECTED+".");
         else if (connected == hovlets.size())
            statusline.setText(STATUS_CONNECTED+".");
         else
            statusline.setText(STATUS_CONNECTED+" "+connected+" of "+hovlets.size()+".");
      }
         }
      });
   }
}
