/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.simulator.Constants;

import hoverball.net.*;
import hoverball.serial.*;
import hoverball.layout.*;
import hoverball.debug.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.datatransfer.*;  // Copy to Clipboard
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

import java.io.IOException;


// ---- Controller ------------------------------------------ [Controller] ----


/**
 * Der Controller, basierend auf der Klasse Hovlet, steuert den Simulator, mit
 * dem er verbunden ist. Er meldet sich als Control Client beim Simulator an
 * und kann somit die gesamte Vielfalt der netzwerkbasierten Steuerung des
 * Simulators nutzen.
 * <p>
 * Zu den Funktionen des Controllers gehören:
 * <menu>
 * <li>die Verwaltung der Channels,</li>
 * <li>die Verwaltung der Optionen,</li>
 * <li>die Steuerung der Simulation,</li>
 * <li>die Visualisierung der Hoverball-Sphäre auf dem Screen.</li>
 * </menu>
 * <p>
 * Für eine genauere Beschreibung der Funktionen des Controllers
 * sei auf das <b>Hoverball Manual</b> verwiesen.
 * <p>
 * Weiterhin hat der Controller die wichtige Aufgabe, das <b>Graphical Debugging</b>
 * der Java-Units, die bei ihm mit dem Befehl
 * <a href="#debug(hoverball.Unit, boolean)">debug(...)</a>
 * registriert wurden, anzuzeigen. Dieses kann dann über das Menu <i>Debug</i>
 * des Controllers für die einzelnen Units ausgewählt oder durch Click
 * auf die Hovlet-Icons der Units ein- und ausgeblendet werden.
 * Siehe hierzu auch <a href="debug/Debug.html">hoverball.debug.Debug</a>.
 * <p>
 * Auf dem Screen ist der Human-KeyListener installiert. Alle Human Units (in derselben
 * Java-Instanz) sind also steuerbar, wenn der Screen aktiv ist. Die Farben der Sphäre
 * und des Hintergrunds können mit den Methoden setForeground(...) und setBackground(...)
 * des Screens geändert werden.
 * <p>
 * Der Controller ist als integriertes Hovlet oder als Stand-Alone (Root-Hovlet)
 * ausführbar, jedoch kann kein Sub-Hovlet an den Controller angeknüpft werden.
 * <p>
 * <hr>
 * <p>
 * Die Klasse Controller kann auch als Application gestartet werden:
 * <p>
 * <tt>java hoverball.Controller </tt>[<i>host</i>][<tt>:</tt><i>port</i>][<tt>#</tt><i>hash</i>]
 * <p>
 * <menu>
 * <li>(ohne) - Öffnet einen Controller.</li>
 * <li><i>host</i><tt>:</tt><i>port</i><tt>#</tt><i>hash</i>  - Öffnet einen Controller und verbindet ihn
 *     mit der angegebenen Adresse.</li>
 * </menu>
 */
public final class Controller extends Hovlet
{
   // ---- static Launcher ----------------------------------------------------

   /**
    * Startet den Controller (Application!).
    */
   public static void main (String[] args)
   {
      Session session = new Session("Hoverball",
                                    (args.length==0)? null : (Object)args) {
         public void browse (JInternalFrame browser) {
            super.browse(browser);
            browser.setLocation(Application.getBounds().width,0);
            session_frame = browser;
         }
      };
      session.show();
      
SwingUtilities.invokeLater(new Runnable() {  // !! WORKAROUND: JAVA 1.8 NEEDS THIS
   public void run () {                      // !!
      session_frame.setLocation(Toolkit.getDefaultToolkit().getScreenSize().width,0); // invisible!

if (session.controller.frame != null) {
      session.controller.frame.setClosable(true);
      session.controller.frame.addInternalFrameListener(new InternalFrameAdapter() {
         public void internalFrameClosed(InternalFrameEvent e) { Application.exit(); }
      });
}
      
      session.controller.viewer.show();
   }                                         // !!
});                                          // !!
   }
   private static JInternalFrame session_frame = null;


   // ---- class Controller ---------------------------------------------------

   /**
    * Frame des Controllers (falls Stand-Alone)
    */
   public JInternalFrame frame;

   /**
    * Frame des Hoverball-Screens.
    */
   public final JInternalFrame viewer;


   // ---- Controller

   private static final String NAME = "Hoverball Controller";

   private static final Icon ICON_INIT    = new TextIcon("<<|",30,0);
   private static final Icon ICON_CHECKIN = new TextIcon(">>|",30,0);
   private static final Icon ICON_START   = new TextIcon(">"  ,30,0);
   private static final Icon ICON_SUSPEND = new TextIcon("||" ,30,0);

   private static final String TOOLTIP_INIT     = "Initialize";
   private static final String TOOLTIP_CHECKIN  = "Check in";
   private static final String TOOLTIP_START    = "Run";
   private static final String TOOLTIP_SUSPEND  = "Pause";

   private static final String S = "                "; // 16_
   private static final String SPACE = S+S+S; // 48_
   private static final String ICON_EMPTY       = ""+SPACE;
   private static final String ICON_INITIALIZED = "Initialized."+SPACE;
   private static final String ICON_COMPLETE    = "Complete."+SPACE;
   private static final String ICON_PAUSED      = "Paused."+SPACE;
   private static final String ICON_RUNNING     = "Running..."+SPACE;

   private NetAgent client = null;
   private final Viewer viewer_;
   private final Options options_;
   private final About about;
   
   private Address host = null;                           // Remember host (Follow & Graphical Debugging!)

   private Constants constants = new Constants();         // Constant Table...
   private Vector channels = new Vector(20,10);           // Channels...
   private View view = null;                              // View... (for Clipboard only!)

   private JInternalFrame frame_ = null;                  // Frame if browsing

   private final JButton button1;                         // Options/Init
   private final JButton button2;                         // Checkin
   private final JButton button3;                         // Run/Pause
   private final JButton label;                           // Simulator's State

   private final DefaultListModel model = new DefaultListModel();
   private final JList list = new JList(model);           // List
   private final JScrollPane list_;
   private final JPanel options;
   private final JPanel card;
   private final CardLayout card_manager;

   private Dimension size = null;                         // Accessing View Menu-Commands...
   private int list_status = 0;
   private Component statusline = null;

   private int state = 0;                                 // current state
   private boolean complete = false;                      // complete-flag


   // ---- private class ButtonListener ---------------------------------------

   private class ButtonListener extends AbstractAction
   {
      private int what;
      ButtonListener (int what) { this.what = what; }

      public void actionPerformed (ActionEvent e) {
    	  control(what);
    	  try { viewer_.frame.setSelected(true); } catch(java.beans.PropertyVetoException v) {}
          viewer_.screen.requestFocus(); // Focus to Screen (!) after Button pressed...    			  
      }
   }

   // ---- private class ConnectListener --------------------------------------
/*
   private class ConnectListener implements ActionListener
   {
      private String host;
      ConnectListener (String host) { this.host = host; }

      public void actionPerformed (ActionEvent e)
      {
         if (client != null) disconnect();
         connect(host);
      }
   }
*/
   // ---- private class StatusComment ----------------------------------------

   private class ZoomListener implements ActionListener
   {
      private double zoom;
      ZoomListener (double zoom) { this.zoom = zoom; }

      public void actionPerformed (ActionEvent e) { zoom(zoom); }
   }

   // ---- private static class DebugMenuItem ---------------------------------

   private class DebugMenuItem extends JCheckBoxMenuItem
   {
      public final Unit unit;

      DebugMenuItem (Unit unit, boolean debug)
      {
         super(unit.name()+"  ",debug);
         this.unit = unit;

         addActionListener(new ActionListener() {
            public void actionPerformed (ActionEvent e) {
               select(isSelected());
            }
         });
      }

      private void select (boolean selected)
      {
         debugging.put(unit,new Boolean(selected));       // toggle debugging-flag

         if (viewer_.visible()) viewer_.screen.update();  // Screen.update!

         if (this.isSelected() != selected) this.setSelected(selected);
      }
   }

   // ---- private class ChannelRenderer (list cell renderer) -----------------

   private class ChannelRenderer
         implements ListCellRenderer
   {
      JLabel separator = new JLabel(" ");
      JLabel team = new JLabel();
      JPanel unit = new JPanel(new GridBagLayout());

      JLabel x = new JLabel("",SwingConstants.RIGHT);
      JLabel a = new JLabel();
      JLabel b = new JLabel();
      JLabel c = new JLabel();

      ChannelRenderer () {
         separator.setFont(Layout.fontPlain); separator.setOpaque(true);
         team.setFont(Layout.fontBold); team.setOpaque(true);
         team.setBorder(BorderFactory.createEmptyBorder(2,0,2,0));

         x.setFont(Layout.fontBold);  x.setPreferredSize(new Dimension(30,0));
         a.setFont(Layout.fontPlain); a.setPreferredSize(new Dimension(80,0));
         b.setFont(Layout.fontPlain);
         c.setFont(Layout.fontPlain); c.setPreferredSize(new Dimension(100,0));
         unit.add(x,new Constraints(0,0,1,1,0.0,0.0,new Insets(0,0,0,0)));
         unit.add(a,new Constraints(1,0,1,1,0.0,0.0,new Insets(0,0,0,0)));
         unit.add(b,new Constraints(2,0,1,1,1.0,0.0,new Insets(0,0,0,0)));
         unit.add(c,new Constraints(3,0,1,1,0.0,0.0,new Insets(0,0,0,0)));
      }

      public Component getListCellRendererComponent(
            JList list,
            Object value,            // value to display
            int index,               // cell index
            boolean isSelected,      // is the cell selected
            boolean cellHasFocus)    // the list and the cell have the focus
      {
         Channel channel = (Channel)value;
         if (channel.team == null) {         // Separator
            return Layout.finishListCell(separator,list,false);
         } else if (channel.name == null) {  // Team
            team.setText("   "+channel.team);
            if(Layout.TOOLTIPS) team.setToolTipText("Double click to reject channels");
            return Layout.finishListCell(team,list,isSelected);
         } else {                            // Unit
        	if(viewer_.channels == null) x.setText(""); // nicht eingecheckt: kein *
        	else {
        	   x.setText("*   ");                       // sonst: findet er's? kein *, sonst *
        	   for (int i=0; i<viewer_.channels.length; ++i)
                  if (viewer_.channels[i] == channel) { x.setText(""); break; }
        	}
            a.setText("(unit "+channel.t+" "+channel.n+")");
            b.setText(channel.name);
            c.setText(channel.host);
            if(Layout.TOOLTIPS) unit.setToolTipText("Double click to reject channels");
            return Layout.finishListCell(unit,list,isSelected);
         }
      }
   }

   // ---- TextIcon -----------------------------------------------------------

   static class TextIcon implements Icon
   {
      private final String text;
      private final int width;
      private final int align;

      public TextIcon (String text, int width, int align)
      {
         this.text = text;
         this.width = width;
         this.align = align;
      }

      public int getIconWidth ()  { return width;  }
      public int getIconHeight () { return 12; }

      public void paintIcon (Component c, Graphics g, int x, int y)
      {
         int dx = (align >  0)? 0 :
                  (align == 0)? (width - c.getFontMetrics(c.getFont()).stringWidth(text))/2 :
                                 width - c.getFontMetrics(c.getFont()).stringWidth(text);
         g.setColor(c.getForeground());
         g.drawString(text,x+dx,y+getIconHeight()-2);
      }
   }

   // ---- Constructors ------------------------------------- [Controller] ----

   /**
    * öffnet einen Controller.
    */
   public Controller () { this(true); }


   /**
    * öffnet einen Controller und ermöglicht oder verbietet die Kontroll-Tastaturkürzel
    * im Screen.
    */
   public Controller (boolean shortcuts)
   {
      super("Controller",true,false);

      // ---- Control Buttons

      toolbar = new JToolBar("Hoverball Controls");
      // toolbar.setRollover(true);

      button1 = toolbar.add(new ButtonListener(1));
      button1.setIcon(ICON_INIT); if(Layout.TOOLTIPS) button1.setToolTipText(TOOLTIP_INIT);
      button1.setEnabled(false);
      button1.setFont(Layout.fontSymbol);
      button1.setFocusPainted(false);

      toolbar.addSeparator(new Dimension(4,4));

      button2 = toolbar.add(new ButtonListener(2));
      button2.setIcon(ICON_CHECKIN); if(Layout.TOOLTIPS) button2.setToolTipText(TOOLTIP_CHECKIN);
      button2.setEnabled(false);
      button2.setFont(Layout.fontSymbol); button2.setFocusPainted(false);

      toolbar.addSeparator(new Dimension(4,4));

      button3 = toolbar.add(new ButtonListener(3));
      button3.setIcon(ICON_START); if(Layout.TOOLTIPS) button3.setToolTipText(TOOLTIP_START);
      button3.setEnabled(false);
      button3.setFont(Layout.fontSymbol);
      button3.setFocusPainted(false);

      toolbar.addSeparator(new Dimension(12,4));

      label = toolbar.add(new AbstractAction(){ public void actionPerformed(ActionEvent e){} });
      //label.setSize(120,1); // [klappt nicht.]
      //label.setHorizontalAlignment(SwingConstants.LEFT);
      label.setText(ICON_EMPTY); label.setEnabled(true);
      label.setFont(Layout.fontControl); label.setFocusPainted(false); label.setBorderPainted(false);
      if(Layout.TOOLTIPS) label.setToolTipText("Switch Channels and Options");

      // Transparency for Ocean Theme!
      button1.setContentAreaFilled(false);
      button2.setContentAreaFilled(false);
      button3.setContentAreaFilled(false);
      label.setContentAreaFilled(false);

      JPanel buttons = new JPanel(new BorderLayout());
      buttons.add(toolbar,BorderLayout.CENTER);
      buttons.setBorder(new EmptyBorder(0,0,2,0));

      // ---- Channel List

      list.addMouseListener(new MouseAdapter() {
         public void mouseClicked (MouseEvent e) {
            if (e.getClickCount() != 2) return;
            int k = list.locationToIndex(e.getPoint());
            if (k < 0 || k >= channels.size()) return;
            bye((Channel)channels.elementAt(k));
         }
      });
      list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      list.addFocusListener(new FocusAdapter() {
         public void focusLost (FocusEvent e) { list.clearSelection(); }
      });
      list.setCellRenderer(new ChannelRenderer());

      list_ = new JScrollPane(list,
            JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
            JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

      // ---- Options

      options_ = new Options(constants,new Options.Listener() {
         public void listen (Object message) {
            if (client != null) client.send(message);
         }
      });
      options = options_.panel;

      // ---- Cards: Channels / Options

      card = new JPanel(card_manager = new CardLayout());
      card.add(list_,"Center"); card_manager.addLayoutComponent(list_,"channels");
      card.add(options,"Center"); card_manager.addLayoutComponent(options,"options");
      card_manager.show(card,"channels");

      // ---- Main Panel: Buttons & List

      JPanel main_ = new JPanel(new BorderLayout());
      main_.add(buttons,BorderLayout.NORTH);
      main_.add(card,BorderLayout.CENTER);
      main.add(main_);

      // ---- Viewer

      viewer_ = new Viewer(new Screen(shortcuts,debugging,new Screen.Listener() {
         public void listen (final Object message) {
            SwingUtilities.invokeLater(new Runnable() {
               public void run () {
            if (message instanceof Unit) {  // toggle Graphical Debugging
               DebugMenuItem item = getDebugMenuItem((Unit)message);
               item.setSelected(!item.isSelected());
               return;
            } else if (message instanceof Boolean) {  // toggle Follow Mode
               menuFollowMode.setSelected(((Boolean)message).booleanValue());
               return;
            } else if (message instanceof Integer) {  // access view memory
               view(((Integer)message).intValue());
               return;
            }
            if (client != null) client.send(message);
               }
            });
         }
      }));
      viewer = viewer_.frame;

      follow(null,1);  // Default: Follow Shot-Ball

      // ---- About

      about = new About(null);

      // ---- Menu

      JMenu menuControl = new JMenu("Control");
         menuReset = new JMenuItem("Initialize",KeyEvent.VK_DELETE);
         menuCheckin = new JMenuItem("Check in",KeyEvent.VK_INSERT);
         menuRunPause = new JMenuItem("Run / Pause",KeyEvent.VK_SPACE);
      menuView = new JMenu("View");
         JMenuItem menuScreen = new JMenuItem("Screen");
         menuFollowMode = new JCheckBoxMenuItem("Follow Mode",false);
         menuAntialiasing = new JCheckBoxMenuItem("Antialiasing",true);
         JMenuItem menuFullScreen = new JMenuItem("Full Screen");
         menuZoom = new JMenu("Zoom");
            JMenuItem menuZoomTo1 = new JMenuItem("... to 1.00");
            JMenuItem menuZoomTo2 = new JMenuItem("... to 2.00");
            JMenuItem menuZoomTo4 = new JMenuItem("... to 4.00");
         menuCopyToClipboard = new JMenuItem("Copy to Clipboard");
         JMenuItem menuChannels = new JMenuItem("Channels");
         JMenuItem menuOptions = new JMenuItem("Options");
         menuCollapse = new JMenuItem("None");

      menuDebug = new JMenu("Debug");
         menuDebugAll = new JMenuItem("Debug All");
         menuDebugNone = new JMenuItem("Debug None");

      JMenu menuX = new JMenu("?");
         JMenuItem menuAbout = new JMenuItem("About...");

      menuControl.add(menuReset);
      menuControl.add(menuCheckin);
      menuControl.add(menuRunPause);

      menuZoom.add(menuZoomTo1); menuZoomTo1.addActionListener(new ZoomListener(1));
      menuZoom.add(menuZoomTo2); menuZoomTo2.addActionListener(new ZoomListener(2));
      menuZoom.add(menuZoomTo4); menuZoomTo4.addActionListener(new ZoomListener(4));

      menuView.add(menuScreen);
      menuView.add(menuFollowMode); menuFollowMode.setEnabled(false);
      menuView.add(menuAntialiasing); menuAntialiasing.setEnabled(true);
      menuView.add(menuFullScreen); menuFullScreen.setEnabled(GraphicsEnvironment
            .getLocalGraphicsEnvironment()
            .getDefaultScreenDevice()
            .isFullScreenSupported());
      menuView.add(menuZoom); menuZoom.setEnabled(false);
      menuView.addSeparator();
      menuView.add(menuCopyToClipboard); menuCopyToClipboard.setEnabled(false);
      menuView.addSeparator();   // menuFollowMode enabled if (viewer_.visible() && state > 0 && connected)
      menuView.add(menuChannels);
      menuView.add(menuOptions);
      menuView.add(menuCollapse); menuCollapse.setVisible(false);

      menuDebug.setEnabled(false);
      menuDebug.add(menuDebugAll);
      menuDebug.add(menuDebugNone);
      menuDebug.addSeparator();

      menuX.add(menuAbout);

      menuReset.addActionListener(new ButtonListener(1));
      menuCheckin.addActionListener(new ButtonListener(2));
      menuRunPause.addActionListener(new ButtonListener(3));

      menuReset.setEnabled(false);
      menuCheckin.setEnabled(false);
      menuRunPause.setEnabled(false);

      menuScreen.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { viewer_.frame.show(); }
      });

      menuCopyToClipboard.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { view_export(); }
      });

      menuFollowMode.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            boolean mode = menuFollowMode.isSelected();
            if (mode) viewer_.screen.follow_enter();
            else viewer_.screen.follow_leave();
         }
      });

      menuAntialiasing.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            boolean mode = menuAntialiasing.isSelected();
            viewer_.screen.setAntialiasing(mode);
         }
      });

      menuFullScreen.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { viewer_.screen.setFullScreen(true); }
      });

      viewer_.frame.addComponentListener(new ComponentAdapter() {
         public void componentShown  (ComponentEvent e) {
            if (state != 0 && client != null) { menuFollowMode.setEnabled(true); menuZoom.setEnabled(true); }
         }
         public void componentHidden (ComponentEvent e) {
            menuFollowMode.setEnabled(false); menuZoom.setEnabled(false);
         }
      });

      menuChannels.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { card_manager.show(card,"channels"); showCard(true); }
      });

      menuOptions.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { card_manager.show(card,"options"); showCard(true); }
      });

      menuCollapse.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) { showCard(false); }
      });

      menuDebugAll.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            int n = menuDebug.getItemCount();
            for (int i=0+3; i<n; ++i) {
               DebugMenuItem item = (DebugMenuItem)menuDebug.getItem(i);
               item.unit.setDebugState(true);
            }
         }
      });

      menuDebugNone.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            int n = menuDebug.getItemCount();
            for (int i=0+3; i<n; ++i) {
               DebugMenuItem item = (DebugMenuItem)menuDebug.getItem(i);
               item.unit.setDebugState(false);
            }
         }
      });

      MouseListener toggle = new MouseAdapter() {  // Quick Channels/Options Toggler
         public void mousePressed (MouseEvent e) {
            if (frame_ != null) {  // stand-alone!
               if (!card.isVisible()) {
                  card_manager.show(card,"channels");
                  showCard(true);
               } else if (list_.isVisible())
                  card_manager.show(card,"options");
               else
                  showCard(false);
            } else {               // integrated Hovlet
               if (list_.isVisible())
                  card_manager.show(card,"options");
               else
                  card_manager.show(card,"channels");
            }
         }
      };
      label.addMouseListener(toggle);

      menuAbout.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            if (!about.frame.isVisible()) center(about.frame);
            about.frame.show();
         }
      });

      menubar.add(menuControl);
      menubar.add(menuView);
      menubar.add(menuDebug);
      menubar.add(menuX);
   }

   // ---- UGLY BUG of Swing
   
   synchronized void viewer_setSize (int width, int height)
   {
	   if (viewer_.screen.getFullScreen())
		   viewer_.screen.VIEWER_SIZE = new Dimension(width,height);
	   else
		   viewer.setSize(width,height);
   }
   
   // ---- center on desktop...

   private static void center (JInternalFrame f)  // frame must be shown...
   {
      JInternalFrame[] frames = Application.desktop.getAllFrames();
      for (int i=0; i<frames.length; ++i)  if (frames[i] == f)
         f.setBounds(Layout.center(Application.desktop.getBounds(),f.getSize()));
   }


   // ---- Global Menu Variables ----------------------------------------------

   private final JMenuItem menuReset;
   private final JMenuItem menuCheckin;
   private final JMenuItem menuRunPause;

   private final JMenu menuView;
   private final JMenu menuZoom;
   private final JCheckBoxMenuItem menuFollowMode;
   private final JCheckBoxMenuItem menuAntialiasing;
   private final JMenuItem menuCopyToClipboard;
   private final JMenuItem menuCollapse;

   private final JMenu menuDebug;
   private final JMenuItem menuDebugAll;
   private final JMenuItem menuDebugNone;
   private final JToolBar toolbar;


   // ---- Accessories --------------------------------------------------------

   /**
    * Schließt den Controller.
    */
   public void close ()
   {
      viewer_.close();
      super.close();
   }


   /**
    * [Implementierung]
    */
   public void browse (JInternalFrame frame)
   {
      this.frame = this.frame_ = frame;

      // ---- Modify installed Components

      toolbar.setFloatable(false);
      toolbar.setBorder(new EmptyBorder(0,8,0,0));

      JPanel header = (JPanel)frame_.getContentPane().getComponent(0);
      GridBagLayout layout = (GridBagLayout)header.getLayout();
      Component[] components = header.getComponents();
      for (int i=0; i<components.length; ++i) {
         GridBagConstraints constraints = layout.getConstraints(components[i]);
         if (constraints.gridy > 1) components[i].setVisible(false);
      }

      statusline = frame_.getContentPane().getComponent(1);

      // ---- Frame

      frame_.setResizable(false);
SwingUtilities.invokeLater(new Runnable() {  // !! WORKAROUND: JAVA 1.8 NEEDS THIS
   public void run () {                      // !!
      frame_.setSize(Layout.BROWSER_WIDTH,Layout.CONTROLLER_HEIGHT);  // size of "un-channeled" Controller

      size = frame_.getSize();

      card.setVisible(false);
      statusline.setVisible(false);
      list_status = 180;
   }                                         // !!
});                                          // !!
  }

   private void showCard (final boolean visible)
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      if (card.isVisible() == visible) return;     // nothing to do.

      if (!visible) {
         card.setVisible(false);
         menuCollapse.setVisible(false);
         statusline.setVisible(false);
         size.height -= list_status;
      } else if (visible) {
         card.setVisible(true);
         menuCollapse.setVisible(true);
         statusline.setVisible(true);
         size.height += list_status;
      }

      if (frame_.getSize().height == size.height) return;  // not changed.
      frame_.setSize(size);
      frame_.revalidate();
      frame_.repaint();
         }
      });
   }


   // ---- Style --------------------------------------------------------------

   /**
    * Setzt den Anzeigestil.
    *
    * @param antialiasing <tt>true</tt> für Antialiasing
    */
// fetch
   // fetch
   public void style (boolean antialiasing)
   {
      viewer_.screen.setAntialiasing(antialiasing);
   }


   // ---- Full Screen --------------------------------------------------------

   /**
    * Setzt Vollbildschirm an oder aus.
    *
    * @param fullscreen <tt>true</tt> für Vollbildschirm
    */
// fetch
   // fetch
   public void fullscreen (boolean fullscreen)
   {
      SwingUtilities.invokeLater(new Thread() {
    	  public void run () {
	    		  
	   viewer_.screen.setFullScreen(fullscreen);
 	  
    	  }
      });
   }

   /**
    * Gibt zurück, ob Vollbildschirm ist.
    *
    * @return <tt>true</tt> bei Vollbildschirm
    */
// fetch
   // fetch
   public boolean fullscreen ()
   {
	   return viewer_.screen.getFullScreen();
   }

   
   // ---- View Manager -------------------------------------------------------

   private void view_export ()
   {
      if (view == null) return;

      String objects = "";
      for (int i=0; i<view.disks.length; ++i) objects += view.disks[i];

      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      StringSelection selection = new StringSelection(objects);
      clipboard.setContents(selection,selection);
   }

   private void view_ (View view)
   {
      if (view == null) return;

      Serial message = view;
      if (state == 0 || state == 2) { message = new CS.State(1); message.next = view; }
      if (client != null) client.send(message);
   }

   /**
    * Speichert einen View.
    * <p>
    * Ein View enthält alle Informationen einer Momentaufnahme eines Hoverball-Spiels.
    * Die Objekte werden im View allerdings nur durch die <code>(<i>what t n</i>)</code>-Kennung
    * charakterisiert. Bei veränderten Channels bekommt der View also eine
    * andere Bedeutung. Die Nutzung dieser Methode ist demnach nur dann sinnvoll,
    * wenn gewährleistet ist, dass sich jedesmal dieselben Units in derselben Reihenfolge
    * mit dem Simulator verbinden.
    * Dies ist etwa in einer Hoverball-Session auf einem lokalen Rechner der Fall.
    * <p>
    * Als String ist der Puck-Teil einer <code>(view...)</code>-Nachricht aus dem
    * <b>Control Client to Server</b>-Protokoll zu übergeben. Es ist also nicht möglich, mit
    * diesem Befehl die Zeit oder den Punktestand zu definieren.
    * <p>
    * <b>Tipp:</b> Ein solcher String kann leicht zur Laufzeit des Spiels erzeugt werden:
    * Das Menu <i>View</i> &gt; <i>Copy to Clipboard</i> kopiert den erforderlichen
    * String für den aktuellen View ins System-Clipboard, so dass dieser anschließend
    * leicht in den Programm-Code eingefügt werden kann.
    * <p>
    * Mit <b><code>null</code></b> wird der View in dem Speicherplatz gelöscht.
    *
    * @param memory Speicherplatz
    * @param pucks Puck-Teil der <code>(view...)</code>-Nachricht
    */
   public void view (int memory, String pucks)  // store
   {
      if (pucks == null)
         viewer_.screen.views.remove(""+memory);
      else {
         Serial view = (Serial)CS.deserialize("(view 0 "+pucks+")");
         if (view != null && view instanceof View) viewer_.screen.views.put(""+memory,view);
      }
      viewer_.screen.update();
   }

   /**
    * Versendet einen View aus dem Speicher.
    *
    * @param memory Speicherplatz
    */
// fetch
   // fetch
   public void view (int memory)
   {
      view_((View)viewer_.screen.views.get(""+memory));
   }

/*
   public void view (String pucks)              // directly!
   {
      Serial view = (Serial)CS.deserialize("(view 0 "+pucks+")");  // null will become null!
      if (view != null && view instanceof View) view_((View)view);
   }
*/


   // ---- Following ----------------------------------------------------------

   /**
    * Verfolgt eine Unit auf dem Screen (Follow-Mode).
    * <p>
    * Hat <code>unit</code> den Wert <b><code>null</code></b>,
    * wird der Follow-Mode deaktiviert.
    *
    * @param unit Unit
    */
   public void follow (Unit unit)
   {
      viewer_.screen.followDefaultUnit = unit;
      viewer_.screen.followDefaultTeam = null;
      viewer_.screen.followDefaultBall = 0;
      viewer_.screen.followDefault();
   }

   /**
    * Verfolgt einen Ball auf dem Screen (Follow-Mode).
    * <p>
    * Hat <code>team</code> den Wert <b><code>null</code></b>,
    * so wird ein Shot-Ball, ansonsten ein Team-Ball des angegebenen Teams verfolgt.
    * <p>
    * Die Ball-Nummern werden von 1 hochgezählt.
    *
    * @param team Team des Balls (<b><code>null</code></b> = Shot-Ball)
    * @param n Ball-Nummer
    */
   public void follow (Team team, int n)
   {
      viewer_.screen.followDefaultUnit = null;
      viewer_.screen.followDefaultTeam = team;
      viewer_.screen.followDefaultBall = n;
      viewer_.screen.followDefault();
   }


   // ---- Zoom ---------------------------------------------------------------

   /**
    * Zoomt den Screen.
    *
    * @param zoom Zoom-Faktor.
    */
   public void zoom (double zoom)
   {
      viewer_.screen.zoom(zoom);
   }


   // ---- Graphical Debugging ------------------------------------------------

   private Hashtable debugging = new Hashtable();  // Unit -> Boolean

   synchronized Boolean getDebugState (Unit unit)  // Access of class Unit
   {
      DebugMenuItem item = getDebugMenuItem(unit);
      if (item == null) return null;

      return new Boolean(item.getState());
   }

   /**
    * Registriert alle Units eines Teams für Graphical Debugging.
    * <p>
    * Ist <code>debug</code> gleich <b><code>true</code></b>,
    * so ist die Grafik-Ausgabe dieser Unit von Beginn an eingeschaltet.
    *
    * @param team Team
    * @param debug Grafik-Ausgabe ein?
    */
   public synchronized void debug (Team team, boolean debug)
   {
      for (int i=0; i<team.hovlets.length; ++i)
         if (team.hovlets[i] instanceof Unit)
            debug((Unit)team.hovlets[i],debug);
   }

   /**
    * Entfernt alle Units eines Teams aus der Registrierung für Graphical Debugging.
    *
    * @param team Team
    */
   public synchronized void debug (Team team)
   {
      for (int i=0; i<team.hovlets.length; ++i)
         if (team.hovlets[i] instanceof Unit)
            debug((Unit)team.hovlets[i]);
   }

   /**
    * Registriert eine Unit für Graphical Debugging.
    * <p>
    * Ist <code>debug</code> gleich <b><code>true</code></b>,
    * so ist die Grafik-Ausgabe dieser Unit von Beginn an eingeschaltet.
    *
    * @param unit Unit
    * @param debug Grafik-Ausgabe ein?
    */
   public synchronized void debug (Unit unit, boolean debug)
   {
      final Unit unit_ = unit;
      final boolean debug_ = debug;
      final Controller this_ = this;
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      DebugMenuItem item = getDebugMenuItem(unit_);

      if (item == null) {  // new?
         if (unit_.debugger != null) unit_.debugger.debug(unit_);  // remove old debuggers' registry

         item = new DebugMenuItem(unit_,debug_);  // Setup new DebugMenuItem
         menuDebug.add(item);
         if (menuDebug.getItemCount() == 1+3) menuDebug.setEnabled(true);  // false -> true

         unit_.debugger = this_;
      }
      item.select(debug_);                       // Select debugging
         }
      });
   }

   /**
    * Entfernt eine Unit aus der Registrierung für Graphical Debugging.
    *
    * @param unit Unit
    */
   public synchronized void debug (Unit unit)
   {
      final Unit unit_ = unit;
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      DebugMenuItem item = getDebugMenuItem(unit_);
      if (item == null) return;                 // now: registered here.

      item.setSelected(false);                  // Unselect debugging
      debugging.remove(unit_);                  // and remove from hashtable!

      unit_.debugger = null;

      menuDebug.remove(item);
      if (menuDebug.getItemCount() == 0+3) menuDebug.setEnabled(false);  // true -> false
         }
      });
   }


   private DebugMenuItem getDebugMenuItem (Unit unit)
   {
      int n = menuDebug.getItemCount();
      for (int i=0+3; i<n; ++i) {
         DebugMenuItem item = (DebugMenuItem)menuDebug.getItem(i);
         if (item.unit == unit) return item;
      }
      return null;
   }

   synchronized void addDebug (Address host, int t, int n, Debug debug, Color color)
   {
      if (this.host == null || !this.host.equals(host)) return;        // Server mismatch!
      if (state == 0) return;

      if (viewer_.visible()) viewer_.screen.addDebug(t,n,debug,color); // debug==null: new Session!

      if (debug instanceof Break && state == 2) state(1);
   }


   // ---- Control Buttons ----------------------------------------------------

   private synchronized void control (int what)
   {
      if (what == 1) {
         if (state == 0) {}
         else if (state == 1) { state(0); }
         else if (state == 2) { state(1); state(0); }
      } else if (what == 2) {
         if (state == 0)      { state(1); }
         else if (state == 1) { state(0); state(1); }
         else if (state == 2) { state(1); state(0); state(1); }
      } else if (what == 3 && !complete) {
         if (state == 0)      { state(1); state(2); }
         else if (state == 1) { state(2); }
         else if (state == 2) { state(1); }
      }
   }

   /**
    * Setzt eine Simulator-Variable.
    * <p>
    * Mit <code>value</code> gleich <b><code>null</code></b> kann die Variable gelöscht werden.
    *
    * @param key Simulator-Variable
    * @param value Wert
    */
   public void set (String key, String value) 
   { 
	  if (client != null) client.send(new CS.Set(key,value)); 
   }

   /**
    * Setzt eine Simulator-Variable.
    * <p>
    * Der <code>double</code>-Wert wird in einen <code>String</code> konvertiert.
    *
    * @param key Simulator-Variable
    * @param value Wert
    */
   public void set (String key, double value)
   {
      set(key,(value==Math.round(value))? ""+(int)value : ""+value);
   }

   /**
    * Liefert den Wert einer Simulator-Variablen.
    *
    * @param key Simulator-Variable
    * @return den Wert
    */
   public String get (String key)
   {
      return constants.get(key);
   }

   /**
    * Liefert den Wert einer Hoverball-Option.
    * <p>
    * Falls es sich bei <code>key</code> um eine Simulator-Variable handelt,
    * die keine Option ist, liefert die Funktion den Wert 0.
    *
    * @param key Hoverball-Option
    * @return den Wert
    */
   public double option (String key)
   {
      int i = constants.indexOf(key);
      if (i < 0) return 0;

      return constants.getConstant(i);
   }

   /**
    * Ändert den Zustand des verbundenen Simulators.
    *
    * @param state neuer Zustand
    */
   public void state (int state) 
   { 
	  if (client != null) client.send(new CS.State(state)); 
   }

   /**
    * Gibt den Zustand des verbundenen Simulators zurück.
    *
    * @return Zustand
    */
   public int state () 
   { 
      return state; 
   }
   
   /**
    * Gibt den Zeit des Spiels im Simulators zurück.
    *
    * @return Zeit
    */
   public double time () 
   {
      if (view != null) return view.time; else return 0;   
   }
   
   /**
    * Gibt zurück, ob das Spiel beendet ist ("Complete").
    *
    * @return beendet
    */
   public boolean complete () 
   {
      return complete;   
   }



   private synchronized void setState (final int state, final boolean complete)
   {
      if (this.state == 0 && state != 0) {          // checkin
         Vector units = new Vector();
         for (Enumeration<Channel> e = channels.elements(); e.hasMoreElements(); ) {
            Channel channel = (Channel)e.nextElement();
            if (channel.name != null) units.addElement(channel);
         }
         Channel[] channels = new Channel[units.size()];
         units.copyInto(channels);
         double frequency= constants.getConstant(Constants.SIMULATOR_FREQUENCY);
         double duration = constants.getConstant(Constants.GAME_DURATION);
         double sphere   = constants.getConstant(Constants.WORLD_RADIUS);
         viewer_.checkin(host,channels,frequency,duration,sphere);
         SwingUtilities.invokeLater(new Runnable() {
            public void run () {
         if (viewer_.visible() && client != null) { menuFollowMode.setEnabled(true); menuZoom.setEnabled(true); }
         if                      (client != null) menuCopyToClipboard.setEnabled(true);
            }
         });
      } else if (this.state != 0 && state == 0) {   // checkout
         viewer_.checkout();
         SwingUtilities.invokeLater(new Runnable() {
            public void run () {
         menuFollowMode.setEnabled(false); menuZoom.setEnabled(false);
         menuCopyToClipboard.setEnabled(false);
            }
         });
         view = null;
      }

      this.state = state;
      this.complete = complete;

      if (state == 1) viewer_.pause(true);
      if (state == 2) viewer_.pause(false);

      model_update(); // für das * in der Channel-Liste, wenn Unit nicht beim Spiel dabei ist

      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      if (state == 0) {
         button3.setIcon(ICON_START); if(Layout.TOOLTIPS) button3.setToolTipText(TOOLTIP_START);
         label.setText(ICON_INITIALIZED);
      } else if (complete) {
         button3.setIcon(ICON_START); if(Layout.TOOLTIPS) button3.setToolTipText(TOOLTIP_START);
         label.setText(ICON_COMPLETE);
      } else if (state == 1) {
         button3.setIcon(ICON_START); if(Layout.TOOLTIPS) button3.setToolTipText(TOOLTIP_START);
         label.setText(ICON_PAUSED);
      } else if (state == 2) {
         button3.setIcon(ICON_SUSPEND); if(Layout.TOOLTIPS) button3.setToolTipText(TOOLTIP_SUSPEND);
         label.setText(ICON_RUNNING);
      }
         }
      });
   }


   // ---- Connect & Disconnect -----------------------------------------------

   java.util.Timer pinger = null;

   /**
    * [Implementierung]
    */
   protected synchronized final boolean connecting (String server)
   {
      Address address = new Address(server);
      if (!address.isValid()) return false;

      host = address;  // Remember Host before connecting! (Follow & Graphical Debugging)
      boolean success = false;

      try
      {
	   	 client = new ControlClient(host);  // must be defined for first received serials
	   	 success = client.connect(host,new CS.Connect(host.getHash(),NAME));
	   	 // host's #hash will be changed externally by 'client.connect'
      }
      catch (IOException e) {}

      if (success) {
         servertext = ""+host; // ggf. mit zugewiesenem #hash vom Server
         Application.servertext_with_hash = servertext;
      }

      if (success) { // install Pinger
    	  pinger = new java.util.Timer("Hoverball Ping (Controller)");
    	  pinger.schedule(new java.util.TimerTask() {
    		  public void run () { client.send(new US.Ping()); }
    	  },0,Socket.PING_EVERY);
      }

      if (success) {
       	 SwingUtilities.invokeLater(new Runnable() {
            public void run () {
         button1.setEnabled(true);
         button2.setEnabled(true);
         button3.setEnabled(true);
         menuReset.setEnabled(true);
         menuCheckin.setEnabled(true);
         menuRunPause.setEnabled(true);
         if (viewer_.visible() && state != 0) { menuFollowMode.setEnabled(true); menuZoom.setEnabled(true); }
         if                      (state != 0) menuCopyToClipboard.setEnabled(true);
         options_.connected(true);
            }
         });
         return true;
      } else {
      	if (client!=null)
      	{
      	   client.close();
      	   client = null;
      	}
      	host = null;
      	return false;
      }
   }

   /**
    * [Implementierung]
    */
   protected synchronized final void disconnecting ()
   {
      pinger.cancel(); // cancel Pinger
      pinger = null;

      client.close();
      client = null;
      host = null;
      removeAll();  // Channels & List
      SwingUtilities.invokeLater(new Runnable() {
         public void run () {
      button1.setEnabled(false);
      button2.setEnabled(false);
      button3.setEnabled(false);
      menuReset.setEnabled(false);
      menuCheckin.setEnabled(false);
      menuRunPause.setEnabled(false);
      menuFollowMode.setEnabled(false); menuZoom.setEnabled(false);
      menuCopyToClipboard.setEnabled(false);
      options_.connected(false);
      constants.setup();
      label.setText(ICON_EMPTY);
         }
      });
      viewer_.checkout();
   }


   // ---- Network Protocol: Server to Control Client ---- [ControlClient] ----

   private final Controller this_ = this;

   private class ControlClient extends NetAgent
   {
      // ---- Bye on close() --------------------------------------------------

      private boolean bye = true;  // send bye on close()
      private boolean closed = false;

      public ControlClient(Address serverAddress) throws IOException
      {
    	  super(serverAddress);
      }

      public synchronized void close ()
      {
         if (closed) return;
         if (bye) this.send(new CS.Bye());
         super.close();
         closed = true;
      }

      // ---- Receive ---------------------------------------------------------

      public void receive (Address sender, String message)
      {
         long arrival = System.currentTimeMillis();  // for screen.view() (Graphical Debugging)
         Serial serials = SC.deserialize(message);

         Serial serial_ = serials;
         while (serial_ != null) synchronized(this_) {
            Serial serial = serial_;
            serial_ = serial_.next;
            serial.next = null;
            if (client == null) continue;  // disconnected? ignore!

            // ---- (connect #hash name) --------------------------------------

            if (serial instanceof SC.Connect)
            {
               host.setHash(((SC.Connect)serial).hash);
               setState(0,false);
            }

            // ---- (set parameter_key value) ---------------------------------

            if (serial instanceof SC.Set)
            {
               SC.Set set = (SC.Set)serial;

               if (set.value != null) constants.put(set.key,set.value);
               else constants.remove(set.key);

               options_.set(set.key,set.value);
            }

            // ---- (channel 2 3 MyTeam Bob 23456 25) -------------------------

            if (serial instanceof Channel)
            {
               Channel channel = (Channel)serial;
               if (channel.name != null) add(channel);
               else {
                  remove(channel.t,channel.n);
                  viewer_.remove(channel.t,channel.n);  // case state==0 will be caught in viewer.remove
               }
            }

            // ---- (state 1) -------------------------------------------------

            if (serial instanceof SC.State)
            {
               SC.State state = (SC.State)serial;
               setState(state.state,complete && state.state != 0);
            }

            // ---- (view 23.7 ...) -------------------------------------------

            if (serial instanceof View)
            {
               view = (View)serial;
               view.arrival = arrival;

               double duration = constants.getConstant(Constants.GAME_DURATION);
               if     (!complete && view.time == duration) { setState(state,true);  }
               else if (complete && view.time != duration) { setState(state,false); }

               viewer_.view(view);
            }

            // ---- (bye) -----------------------------------------------------

            else if (serial instanceof SC.Bye)
            {
               bye = false;  // don't send bye yourself!
               disconnect();
            }
         }
      }
   }


   // ---- Channel List -------------------------------------------------------

   private synchronized void bye (Channel channel)
   {
      if (channel.team == null) return;  // separator
      else if (channel.name == null) {   // team
         int k = channels.indexOf(channel);
         while(true) {
            ++ k;
            channel = (Channel)channels.elementAt(k);
            if (channel.name == null) break; // separator
            bye(channel);
         }
      } else                             // unit
         client.send(new CS.Channel(channel.t,channel.n));
   }



   private synchronized void add (Channel channel)
   {
      int k = searchInsert(channel.t,channel.n);
      // hier könnte noch rein: Abbruch, falls (t,n) schon existiert (das könnte vielleicht passieren, auch wenn es das nicht sollte)

      boolean newteam = (k == channels.size() ||    // list over ||
         ((Channel)channels.elementAt(k)).n == 0);  // team?

      if (!newteam) {
         addToList(k,channel);
      } else {
         addToList(k,new Channel(channel.t,0,channel.team,null,0,null));  // team
         addToList(k+1,channel);                                          // unit
         addToList(k+2,new Channel(channel.t,999999,null,null,0,null));   // separator
      }
      model_update();
   }

   private int searchInsert (int t, int n)  // search index to insert!
   {
      int k = channels.size();

      while(true) {
         if (k == 0) break;
         Channel channel = (Channel)channels.elementAt(k-1);
         if (t > channel.t || (t == channel.t && n > channel.n)) break;
         -- k;
      }

      return k;
   }

   private void addToList (final int k, final Channel channel) { channels.add(k,channel); }



   private synchronized void remove (final int t, final int n)
   {
      int k = searchRemove(t,n);                 // sollte es eigentlich geben...
      if (k<0) return;                           // (könnte aber sein, dass es schon weg ist)
      channels.remove(k);

      k = searchRemove(t,0);  // must be found!
      if (((Channel)channels.elementAt(k+1)).team == null) {  // separator?
         channels.remove(k);  // last unit, remove team!
         channels.remove(k);  // remove separator!
      }
      model_update();
   }

   private int searchRemove (int t, int n)  // search *exact* index to remove!
   {
      int k = channels.size();

      while(true) {
         if (k == 0) break;
         Channel channel = (Channel)channels.elementAt(k-1);
         if (t == channel.t && n == channel.n) break;
         -- k;
      }

      return k-1;
   }



   private synchronized void removeAll ()  // invoked by disconnected()
   {
      channels.removeAllElements();
      model_update();
   }



   // ---- ListModel must will be updated thread-safe!

   private void model_update ()
   {
      SwingUtilities.invokeLater(new Runnable() {
         public void run () { copy_channels_to_model(); }
      });
   }

   private synchronized void copy_channels_to_model ()
   {
      model.removeAllElements();;
      for (int i=0; i<channels.size(); ++i)
         model.addElement(channels.elementAt(i));
   }
}
