/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.layout.*;
import hoverball.net.Address;

import java.awt.*;
import javax.swing.*;

// ---- The Hovlet ---------------------------------------------- [Hovlet] ----

/**
		 * Die Klasse Hovlet (= "kleines Hoverball-Programm") stellt die
		 * Basis-Klasse für alle Programm-Elemente dar, die sich mit dem
		 * Simulator verbinden können.
		 * <p>
		 * Hovlets sind rekursiv verschachtelbar. Jedes Hovlet besitzt ein
		 * Super-Hovlet <a href="#parent">parent</a> und eine Liste von
		 * Sub-Hovlets <a href="#hovlets">hovlets</a>, die mit der Methode <a
		 * href="#add(hoverball.Hovlet)">add(...)</a> an das Hovlet angeknüpft
		 * werden können. Dabei lässt sich die Verknüpfbarkeit eines Hovlets
		 * nach oben und nach unten im Konstruktor definieren. Ein Hovlet ohne
		 * Super-Hovlet heißt Root-Hovlet.
		 * <p>
		 * Die so entstehenden Hovlet-Bäume werden in sogenannten <i>Browsern</i>
		 * angezeigt, die je nach Auswahl die einzelnen Hovlets des Baumes in
		 * den Vordergrund bringen. Mit der Methode <a href="#show()">show()</a>
		 * wird für den ganzen Hovlet-Baum, sofern dies nicht schon geschehen
		 * ist, ein neuer Browser geöffnet, der anfänglich das aufrufende Hovlet
		 * anzeigt. (Die Erscheinung des Browsers kann dabei vom Root-Hovlet mit
		 * der Methode <a href="#browse(javax.swing.JInternalFrame)">browse(...)</a>
		 * bestimmt werden.) Andernfalls wird das Hovlet im entsprechenden
		 * Browser einfach in den Vordergrund gebracht.
		 * <p>
		 * <div align="center"> <img src="../resources/hovlet.png" width=400
		 * height=342></img> </div>
		 * <p>
		 * Ein Hovlet hat folgende Eigenschaften: <menu>
		 * <li>Es besitzt einen Titel. Siehe <a href="#title">title</a>.
		 * <p>
		 * </li>
		 * <li>Es kann eigene Menus definieren, welche im Browser neben dem
		 * globalen Menu <i>Hovlet</i> angezeigt werden. Siehe <a
		 * href="#menubar">menubar</a>.
		 * <p>
		 * </li>
		 * <li>Es besitzt ein Icon-Panel, welches im Browser neben der
		 * Hovlet-Auswahlleiste angezeigt wird. Siehe <a href="#icon">icon</a>.
		 * <p>
		 * </li>
		 * <li>Es besitzt ein Main-Panel. Dieses wird im Zentrum des Browsers
		 * angezeigt und kann frei definiert werden. Die paint(...)-Methode des
		 * Main-Panels, die standardmäßig das Hovlet-Banner zeichnet, kann dabei
		 * benutzt werden. Siehe <a href="#main">main</a>.
		 * <p>
		 * </li>
		 * <li>Es kann eine Statuszeile im Browser anzeigen. Siehe <a
		 * href="#status(java.lang.String)">status(...)</a>.
		 * <p>
		 * </li>
		 * </menu>
		 *
		 * Ein Hovlet soll sich mit dem Hoverball-Simulator verbinden können.
		 * Hierzu stehen die Methoden <a
		 * href="#connect(java.lang.String)">connect(...)</a> und <a
		 * href="#disconnect()">disconnect()</a> zur Verfügung, jedoch ist das
		 * Verbinden selbst noch nicht implementiert.
		 */
public abstract class Hovlet {
	// ---- Browser Access

	final boolean upward;
	final boolean downward;

	Hovlet parent_ = null; // internal Hovlet-Structure
	Hovlet[] hovlets_ = new Hovlet[0];

	Browser browser = null; // link to Browser (null = not fixed!)
	Point location = null; // "soft" Browser property for location

	static final int DISCONNECTED = 0; // Connection States...
	static final int CONNECTING = 1;
	static final int CONNECTED = 2;
	static final int DISCONNECTING = 3;

	int state = DISCONNECTED; // current Connection-State
	String servertext = ""; // buffer for Server-TextField
	String status = null; // Status Line

	// ---- Thread Group

//APPLET THREAD BUG//
   static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("hovlets");

	// ---- Default Panel

	private static class DefaultPanel extends JPanel {
		private final String TITLE;
		private final String CLASS;

		DefaultPanel(Hovlet hovlet) {
			super(new BorderLayout());
			setBackground((background != null) ? background : getBackground()
					.darker());
			setForeground((foreground != null) ? foreground : Color.black);

			TITLE = hovlet.title;
			String CLASS_ = "." + hovlet.getClass().getName();
			CLASS = "class " + CLASS_.substring(CLASS_.lastIndexOf('.')+1);
		}

		public void paintComponent(Graphics g) {
			super.paintComponent(g);
			Layout.drawBanner(g, getSize(), TITLE, CLASS, true);
		}
	}

	// ---- public -------------------------------------------------------------

	/**
			 * Default-Hintergrundfarbe für alle neuen Hovlets
			 */
public static Color background = null;

	/**
			 * Default-Vordergrundfarbe für alle neuen Hovlets
			 */
public static Color foreground = null;

	/**
			 * Höhe des Hovlet-Banners (in Pixeln)
			 */
public static final int BANNER = Layout.BANNER;

	/**
			 * Super-Hovlet (<b><code>null</code></b>, falls Root)
			 */
public Hovlet parent = parent_; // transparent Hovlet-Structure

	/**
			 * Sub-Hovlets
			 */
public Hovlet[] hovlets = hovlets_;

	/**
			 * Titel des Hovlets
			 */
public final String title; // Hovlet's properties...

	/**
			 * Menu-Bar des Hovlets
			 */
public final JMenuBar menubar;

	/**
			 * Icon-Panel des Hovlets
			 */
public final JPanel icon;

	/**
			 * Main-Panel des Hovlets
			 */
public final JPanel main;

	// ---- Constructor --------------------------------------------------------

	/**
			 * Erzeugt ein leeres Hovlet.
			 *
			 * @param title
			 *            Titel des Hovlets
			 */
public Hovlet(String title) {
		this(title, true, true);
	}

	/**
			 * Erzeugt ein leeres Hovlet mit angegebener Verknüpfbarkeit.
			 *
			 * @param title
			 *            Titel des Hovlets
			 * @param upward
			 *            nach oben verknüpfbar?
			 * @param downward
			 *            nach unten verknüpfbar?
			 */
public Hovlet(String title, boolean upward, boolean downward) {
		this.title = title;

		this.upward = upward;
		this.downward = downward;

		main = new DefaultPanel(this);
		icon = new JPanel(new BorderLayout());

		menubar = new JMenuBar();
		menubar.add(new JMenu("Hovlet"));
	}

	private boolean closed = false;

	/**
			 * Schließt das Hovlet.
			 */
public void close() {
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("closing: "+title+"...");
		if (closed)
			return;
		closed = true;

		for (int i = 0; i < hovlets_.length; ++i)
			hovlets_[i].close();

		disconnect();
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("closed : "+title+".");
	}

	/**
			 * Gibt den Titel des Hovlets mit Pfad zurück.
			 *
			 * @param separator
			 *            Trenn-Zeichenkette für den Pfad
			 * @return den Titel mit Pfad
			 */
public String title(String separator) {
		if (parent_ == null || separator == null)
			return title;
		else
			return parent_.title(separator) + separator + title;
	}

	/**
			 * Fügt ein Hovlet in dieses Hovlet ein.
			 * <p>
			 * Ist die Verknüpfung der beiden Hovlets nicht zulässig, oder wurde
			 * eines der beiden Hovlets bereits verknüpft oder angezeigt, kann
			 * die Verknüpfung nicht erfolgen.
			 *
			 * @param hovlet
			 *            Hovlet
			 * @return <b><code>true</code></b>, falls Verknüpfung erfolgte
			 *         <b><code>false</code></b>, falls Verknüpfung nicht
			 *         erfolgte
			 */
public boolean add(Hovlet hovlet) {
		if ((browser != null || !downward)
				|| (hovlet.browser != null || !hovlet.upward)
				|| hovlet.parent_ != null)
			return false; // addition failed.

		Hovlet[] array = hovlets_;
		hovlets_ = new Hovlet[array.length + 1];
		System.arraycopy(array, 0, hovlets_, 0, array.length);
		hovlets_[array.length] = hovlet;

		hovlet.parent_ = this;

		hovlet.parent = hovlet.parent_;
		hovlets = hovlets_;

		return true;
	}

	/**
			 * Zeigt das Hovlet an.
			 * <p>
			 * Ist dies der erste Aufruf von show() von allen Hovlets innerhalb
			 * des Hovlet-Baumes, wird ein neuer Browser für diesen Hovlet-Baum
			 * geöffnet und das Hovlet angezeigt. Ansonsten wird das Hovlet
			 * einfach in den Vordergrund gebracht.
			 */
public synchronized void show() {
		if (browser == null) // if not yet shown:
			new Browser(this); // open Browser and setup browser-links!

		browser.show(this);
	}

	/**
			 * Zeigt einen Text in der Status-Zeile des Browsers an.
			 * <p>
			 * Mit <b><code>null</code></b> wird der Standard-Status des
			 * Browsers angezeigt.
			 *
			 * @param status
			 *            Status
			 */
public void status(String status) {
		this.status = status;
		if (browser != null)
			browser.status(this, null);
	}

	/**
			 * Wünscht die Browser-Position.
			 * <p>
			 * Ist dieses Hovlet das Root-Hovlet, so wird der Browser beim
			 * Öffnen an der angegebenen Stelle positioniert.
			 *
			 * @param x
			 *            <i>x</i>-Koordinate der gewünschten Browser-Position
			 * @param y
			 *            <i>y</i>-Koordinate der gewünschten Browser-Position
			 */
public void locate(int x, int y) {
		locate(new Point(x, y));
	}

	/**
			 * Wünscht die Browser-Position.
			 * <p>
			 * Ist dieses Hovlet das Root-Hovlet, so wird der Browser beim
			 * Öffnen an der angegebenen Stelle positioniert.
			 *
			 * @param location
			 *            gewünschte Browser-Position
			 */
public void locate(Point location) {
		if (location == null)
			this.location = null;
		else
			this.location = new Point(location.x, location.y);
	}

	/**
			 * Beeinflusst die Erscheinung des Browsers, falls es sich bei
			 * diesem Hovlet um das Root-Hovlet handelt.
			 * <p>
			 * Soll ein Hovlet-Baum in einem Browser angezeigt werden, kann das
			 * Root-Hovlet über die Erscheinung des Browsers bestimmen. Ist
			 * dieses Hovlet das Root-Hovlet, so kann es in diese Methode das
			 * Frame des Browsers modifizieren. Der Browser ist zu diesem
			 * Zeitpunkt noch nicht sichtbar.
			 * <p>
			 * Diese Methode ist leer und kann überschrieben werden.
			 *
			 * @param browser
			 *            Frame des Browsers
			 */
public void browse(JInternalFrame browser) {}

	// ---- Connection Commands ------------------------------------------------

	/**
			 * Verbindet das Hovlet mit der angegebenen Adresse.
			 *
			 * @param address
			 *            Netzwerk-Adresse
			 * @return <b><code>true</code></b>, falls Verbindung erfolgte
			 */
public synchronized boolean connect(String address) {
        if (state == CONNECTED)
			disconnect(); // disconnect first!
		if (state != DISCONNECTED)
			return false;
		servertext = (address == null)? "" : "" + address;
		if (servertext.startsWith(""+Address.PORT)) servertext = "localhost" + servertext; // ":1234" >> "localhost:1234"
		
		state = CONNECTING;
		if (browser != null)
			browser.refresh(this);

////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("Hovlet.connecting: "+title+"...");
      // "AWEFUL BAD WORKAROUND"
      if (Application.servertext_with_hash != null) servertext = Application.servertext_with_hash;

		boolean success = connecting(servertext);

      // "AWEFUL BAD WORKAROUND"
      if (Application.servertext_with_hash != null) servertext = Application.servertext_with_hash;
////for(Hovlet i=this; i.parent!=null; i=i.parent) System.out.print("  ");
////System.out.println("Hovlet.connected : "+title+". "+success);

		if (success)
			state = CONNECTED;
		else
			state = DISCONNECTED; // CONNECTION_FAILED;
		if (browser != null)
			browser.refresh(this);

		return success;
	}

	/**
			 * Trennt die Verbindung des Hovlets.
			 */
public synchronized void disconnect() {
		if (state != CONNECTED)
			return;

		state = DISCONNECTING;
		if (browser != null)
			browser.refresh(this);

		disconnecting();

		state = DISCONNECTED;
		if (browser != null)
			browser.refresh(this);
	}

	// ---- Virtual Methods (to override!) -------------------------------------

	/**
			 * Frei für die Implementierung des Verbindens.
			 * <p>
			 * Diese Methode sollte <b><code>true</code></b> zurückgeben,
			 * falls die Verbindung erfolgen konnte.
			 * <p>
			 * Der bisherige Inhalt dieser Methode verbindet sukzessive alle
			 * Sub-Hovlets mit der angegebenen Adresse und gibt <b><code>true</code></b>
			 * zurück, falls mindestens eine Verbindung erfolgen konnte. Auf
			 * diesen Code kann bei Bedarf mit <b>super</b>.connecting(...)
			 * zugegriffen werden.
			 *
			 * @param address
			 *            Netzwerk-Adresse
			 * @return <b><code>true</code></b>, falls Verbindung erfolgte
			 */
protected boolean connecting(String address) {
		boolean success = false;
		for (int i = 0; i < hovlets_.length; ++i)
			success |= hovlets_[i].connect(address);
		return success;
	}

	/**
			 * Frei für die Implementierung des Trennens.
			 * <p>
			 * Der bisherige Inhalt dieser Methode trennt sukzessive alle
			 * Sub-Hovlets. Auf diesen Code kann bei Bedarf mit <b>super</b>.disconnecting()
			 * zugegriffen werden.
			 */
protected void disconnecting() {
		for (int i = 0; i < hovlets_.length; ++i)
			hovlets_[i].disconnect();
	}
}
