/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.math.*;
import hoverball.serial.DiskInfo;  // import WHATs, putString


// ---- Hoverball Object Info ------------------------------------- [Puck] ----


/**
 * Die Klasse Puck speichert alle Informationen eines Pucks, die von einer
 * Unit wahrgenommen oder erschlossen werden können.
 * Hierzu zählen:
 *
 * <ul>
 * <li>die Identität des Objekts, die sich aus der Objekt-Kennung
 *     <code>(<i>what t n</i>)</code> zusammensetzt,</li>
 * <li>die Botschaft des Objekts, falls es sich um eine Unit handelt,
 * <li>die Position und die Blickrichtung, welche durch eine orthonormale Matrix
 *     repräsentiert werden,</li>
 * <li>der Radius,</li>
 * <li>die Masse.</li>
 * </ul>
 *
 * Auch die Nodes werden durch die Klasse Puck repräsentiert, obwohl es sich
 * im strengeren Sinne bei ihnen ja um keine echten Hoverball-Pucks handelt.
 * <p>
 * Mit der Klasse Puck stehen zahlreiche Methoden zum bequemen Organisieren
 * und Vergleichen der Objekte zur Verfügung.
 */
public class Puck
{
   /**
    * Typ-Konstante: Node (Knoten)
    */
   public static final int NODE = DiskInfo.NODE;

   /**
    * Typ-Konstante: Ball
    */
   public static final int BALL = DiskInfo.BALL;

   /**
    * Typ-Konstante: Unit
    */
   public static final int UNIT = DiskInfo.UNIT;


   /**
    * Identität des Objekts (als String)
    */
   public final String id;  // "what t n"

   /**
    * Typ des Objekts
    */
   public final int what;

   /**
    * Team-Nummer des Objekts
    */
   public final int t;

   /**
    * Puck-Nummer des Objekts (bei den Units ist das die Spieler-Nummer)
    */
   public final int n;

   /**
    * Radius des Objekts
    */
   public final double r;

   /**
    * Masse des Objekts
    */
   public final double m;

   /**
    * Position des Objekts (als orthonormale Matrix)
    */
   public Matrix X;

   /**
    * Botschaft des Objekts (falls Unit)
    */
   public String message;


   /**
    * Erzeugt die Identität (als String).
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @param n Puck-Nummer
    * @return die Identität
    */
   public static String id (int what, int t, int n)
   {
      return DiskInfo.putWhat(what)+" "+t+" "+n;
   }


   /**
    * Erzeugt einen Puck gleich einem anderen Puck.
    *
    * @param puck zu kopierender Puck
    */
   public Puck (Puck puck)
   {
      this(puck == null? 0 : puck.what,
           puck == null? 0 : puck.t,
           puck == null? 0 : puck.n,
           puck == null? new Matrix() : puck.X,
           puck == null? 0 : puck.r,
           puck == null? 0 : puck.m,
           puck == null? null : puck.message);
   }

   /**
    * Erzeugt einen Puck mit angegebener Identität,
    * die restlichen Werte sind null.
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @param n Puck-Nummer
    */
   public Puck (int what, int t, int n)
   {
      this(what,t,n,new Matrix(),0,0,null);
   }

   /**
    * Erzeugt einen Puck mit den angegebenen Werten.
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @param n Puck-Nummer
    * @param X orientierte Position
    * @param r Radius
    * @param m Masse
    * @param message Message
    */
   public Puck (int what, int t, int n, Matrix X, double r, double m, String message)
   {
      this.what = what;
      this.t = t;
      this.n = n;
      this.X = (X==null)? null : new Matrix(X);
      this.r = r;
      this.m = m;
      this.message = (message==null)? null : new String(message);

      this.id = id(what,t,n);
   }


   // ---- system functions

   /**
    * Vergleicht mit einem Puck.
    * <p>
    * Zwei Puck-Objekte sind genau dann gleich, wenn ihre Identität gleich ist.
    *
    * @param puck Puck
    * @return die Gleichheit
    */
   public boolean equals (Object puck)
   {
      if (puck == null) return false;
      if (! (puck instanceof Puck)) return false;

      Puck d = (Puck)puck;
      return equals(d.what,d.t,d.n);
   }

   /**
    * Vergleicht den Typ.
    *
    * @param what Typ des Objekts
    * @return die Gleichheit
    */
   public boolean equals (int what)
   {
      return this.what == what;
   }

   /**
    * Vergleicht den Typ und das Team.
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @return die Gleichheit
    */
   public boolean equals (int what, int t)
   {
      return this.what == what && this.t == t;
   }

   /**
    * Vergleicht die Identität.
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @param n Puck-Nummer
    * @return die Gleichheit
    */
   public boolean equals (int what, int t, int n)
   {
      return this.what == what && this.t == t && this.n == n;
   }

   /**
    * Vergleicht die Identität (als String).
    *
    * @param id Identitäts-String
    * @return die Gleichheit
    */
   public boolean equals (String id)
   {
      if (id == null) return false;
      else return this.id.equals(id.trim());
   }

   /**
    * Berechnet den Hash-Code.
    *
    * @return den Hash-Code
    */
   public int hashCode ()
   {
      return (what & 0xFF << 16) + (t & 0xFF << 8) + (n & 0xFF);
   }


   /**
    * String-Repräsentation.
    *
    * @return den String
    */
   public String toString()
   {
      return getClass().getName()
         +"[(" + id + ")"
         +" X=" + (X == null? ""+X : ""+Sphere.euler(X))
         +" r=" + format(r)
         +" m=" + format(m)
         +((message == null)? "" : " " + DiskInfo.putString(message))
         +"]";
   }

   static String format (double d)
   {
      if (Complex.format == null) return ""+d;
      else return Complex.format.format(d);
   }
}
