/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import java.net.InetAddress;

import javax.swing.*;

import hoverball.net.Address;
import hoverball.simulator.Constants;


// ---- Hoverball Simulator ---------------------------------- [Simulator] ----


/**
 * Der Simulator ist das Kernstück des Hoverball-Programmpakets.
 * Er erzeugt eine virtuelle Hoverball-Welt, in der sich die Spieler, die sich
 * mit ihm verbinden, nach den Gesetzen dieser Welt bewegen können.
 * <p>
 * Da der Simulator umfassend durch das Netzwerk gesteuert werden kann,
 * stellt diese Klasse nur ein paar elementare Methoden für die Konfiguration
 * des Simulators zur Verfügung.
 * <p>
 * Für eine genauere Beschreibung des Simulators
 * sei auf das <b>Hoverball Manual</b> verwiesen.
 * <p>
 * <hr>
 * <p>
 * Die Klasse Simulator kann auch als Application gestartet werden:
 * <p>
 * <tt>java hoverball.Simulator </tt>[<tt>:</tt><i>port</i>]
 * <p>
 * <menu>
 * <li>(ohne) - Öffnet einen Simulator am Default-Port.</li>
 * <li><tt>:</tt><i>port</i> - Öffnet einen Simulator am angegebenen Port.</li>
 * </menu>
 */
public final class Simulator
{
   static Simulator headless_simulator = null; // wird bei Application.exit() extra geschlossen :-)  (AWEFUL BAD WORKAROUND)
	   
   // ---- static Launcher ----------------------------------------------------

   private static boolean VISIBLE = true;  // can globally be turned off for testing purposes

   /**
    * Startet den Simulator (Application!).
    */
   public static void main (String[] args)
   {
	  VISIBLE = ! java.awt.GraphicsEnvironment.isHeadless();

	  if (VISIBLE) {
		  Application.setTitle("Hoverball");
		  Application.setBounds();
	  }

      String arg = hoverball.layout.About.arg(args);
      Simulator simulator = new Simulator((args.length == 0)? null : arg,VISIBLE);

      Runtime.getRuntime().addShutdownHook(new Thread("Hoverball Bye Hook") {
         public void run() {
            simulator.close();
            if (!VISIBLE) System.out.println("\nBye.");
         }
      });

      if (!VISIBLE) { // headless
          String address = simulator.address;
          String canonical = ""; try { canonical = InetAddress.getLocalHost().getCanonicalHostName(); } catch (java.net.UnknownHostException e) {}
          String address_ = address.replace(new Address(address).getHostName(),canonical);

          System.out.println("Hoverball Simulator at '" + address_  + "' at " + address + "");
          System.out.println("Version "+hoverball.layout.About.VERSION_COMPLETE);
          System.out.println();
          System.out.println("Press 'Ctrl-C' to quit.");
      }
      
      if (simulator.frame != null) simulator.frame.setVisible(true); // SHOW SIMULATUR
   }


   // ---- class Simulator ----------------------------------------------------

   hoverball.simulator.Simulator simulator = null; // access by Server

   /**
    * Default-Port für den Simulator
    */
   public static final int DEFAULT_PORT = hoverball.layout.About.DEFAULT_PORT;


   /**
    * Netzwerk-Adresse des Simulators
    */
   public final String address;

   /**
    * Frame des Simulators
    */
/*private*/ final JInternalFrame frame;


   /**
    * Öffnet einen Simulator.
    */
   public Simulator () { this(-1,true); }

   /**
    * Öffnet einen Simulator am angegebenen Port.
    *
    * @param port Port des Simulators
    */
   public Simulator (int port) { this(port,true); }

   /**
    * Öffnet einen Simulator sichtbar oder unsichtbar ("headless").
    *
    * @param visible sichtbar?
    */
   public Simulator (boolean visible) { this(-1,visible); }

   /**
    * Öffnet einen Simulator am angegebenen Port sichtbar oder unsichtbar.
    *
    * @param port Port des Simulators
    * @param visible sichtbar?
    */
   public Simulator (int port, boolean visible)
   {
      this(""+port,visible);
      if (frame != null) frame.setVisible(visible);
   }

   /*
    * package-private contructor for "my" main()-methods -- does not 'show' simulator.frame
    * to locate it correctly before showing it... :-)
    */
   Simulator (String port, boolean visible)  // parse ":nnnn"-Strings (not public)
   {
	  int port_ = parse(port);
	  if (port_ < 0)
         simulator = new hoverball.simulator.Simulator(visible);
      else
         simulator = new hoverball.simulator.Simulator(visible,port_);

      address = simulator.address;
      frame = simulator.frame;
      if (frame == null) headless_simulator = this;
   }

   private static int parse (String port)
   {
      if (port != null && port.startsWith(""+hoverball.net.Address.PORT)) try {
         return Integer.parseInt(port.substring(1));
      } catch(NumberFormatException e) {
         return -1;
      } else if (port != null) try {
         return Integer.parseInt(port);
      } catch(NumberFormatException e) {
         return -1;
      } else
         return -1;
   }


   // ---- Simulator Properties (accessed by Server) --------------------------

   int getNumberOfChannels () { return simulator.getNumberOfChannels(); }
   int getNumberOfControls () { return simulator.getNumberOfControls(); }
   int getState()             { return simulator.getState(); }
   double getTime()           { return simulator.getTime(); }
   double getDuration()       { return simulator.getDuration(); }
   double getHertz()          { return simulator.getHertz(); }


   // ----

   /**
    * Setzt eine Simulator-Variable.
    * <p>
    * Mit <code>value</code> gleich <b><code>null</code></b> kann die Variable gelöscht werden.
    *
    * @param key Simulator-Variable
    * @param value Wert
    */
   public void set (String key, String value)
   {
      simulator.set(key,value);
   }

   /**
    * Setzt eine Simulator-Variable.
    * <p>
    * Der <code>double</code>-Wert wird in einen <code>String</code> konvertiert.
    *
    * @param key Simulator-Variable
    * @param value Wert
    */
   public void set (String key, double value)
   {
      simulator.set(key,(value==Math.round(value))? ""+(int)value : ""+value);
   }

   /**
    * Liefert den Wert einer Simulator-Variablen.
    *
    * @param key Simulator-Variable
    * @return den Wert
    */
   public String get (String key)
   {
      return simulator.get(key);
   }

   /**
    * Liefert den Wert einer Hoverball-Option.
    * <p>
    * Falls es sich bei <code>key</code> um eine Simulator-Variable handelt,
    * die keine Option ist, liefert die Funktion den Wert 0.
    *
    * @param key Hoverball-Option
    * @return den Wert
    */
   public double option (String key)
   {
      return simulator.option(key);
   }

   /**
    * Ändert den Zustand des Simulators.
    *
    * @param state neuer Zustand
    */
   public void state (int state)
   {
      simulator.state(state);
   }

   /**
    * Schließt den Simulator.
    */
   public void close ()
   {
      simulator.close();
   }

   /**
    * Positioniert das Simulator-Fenster.
    *
    * @param x <i>x</i>-Koordinate der Fenster-Position
    * @param y <i>y</i>-Koordinate der Fenster-Position
    */
   public void locate (int x, int y)
   {
      if (frame != null) frame.setLocation(x,y);
   }
   
   // ----

   /**
    * Gibt den Zustand des Simulators zurück.
    *
    * @return Zustand
    */
   public int state () 
   { 
	  return simulator.state(); 
   }
   
   /**
    * Gibt den Zeit des Spiels im Simulators zurück.
    *
    * @return Zeit
    */
   public double time () 
   { 
	  return simulator.time(); 
   }
   
   /**
    * Gibt zurück, ob das Spiel beendet ist ("Complete").
    *
    * @return beendet
    */
   public boolean complete () 
   { 
	  return simulator.complete(); 
   }
   
}
