/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball;

import hoverball.simulator.Constants;
import hoverball.debug.*;

import hoverball.net.*;
import hoverball.serial.*;
import hoverball.layout.*;
import hoverball.math.*;

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;

import javax.swing.*;
import java.util.Hashtable;


// ---- Unit ------------------------------------------------------ [Unit] ----


/**
 * Die Klasse Unit, basierend auf der Klasse Hovlet, stellt die Basisklasse
 * für alle Java-Units dar. Die Unit verbindet sich als Unit Client mit dem Simulator
 * und erleichtert Dir mit ihren vielen integrierten Funktionen das Erschaffen
 * Deiner eigenen Spieler.
 * <p>
 * Nach der Definition eines Konstruktors, in welchem Du Team- und Spieler-Namen
 * angeben kannst, musst Du nur noch die Denk-Schleife
 * <a href="#loop()">loop()</a>
 * implementieren. Hierzu sei gesagt:
 *
 * <menu>
 * <li>Die Schleife sollte verlassen werden, wenn das Spiel unterbrochen wird,
 *     da die Methode erneut aufgerufen wird, wenn das Spiel wieder weitergeht.
 *     Dies wird durch den richtigen Einsatz der Methode
 *     <a href="#look()">look()</a>
 *     erreicht, die auf den nächsten Look des Simulators wartet.
 *     Wird das Spiel dabei unterbrochen, liefert look() den Wert <b><code>false</code></b>,
 *     ansonsten <b><code>true</code></b>. Das bedeutet, dass die Schleifenstruktur
 *<pre>
 *public void loop()
 *{
 *   while(look())
 *   {
 *      ...
 *   }
 *}
 *</pre>
 *     das Gewünschte leistet.
 *     <p></li>
 * <li>Vor Schleifeneintritt sind die Checkin-Parameter bekannt,
 *     im Schleifenkörper - nach Aufruf von look() -
 *     dann auch die aktuellen Positionen der Objekte,
 *     welche im Array
 *     <a href="#pucks">pucks</a>
 *     abgerufen werden können.
 *     <p></li>
 * <li>Die Aktion sollte mit
 *     <a href="#action(double, double, double)">action(...)</a>
 *     am Ende des Schleifenkörpers gesendet werden.
 *     <p></li>
 * <li>Debug-Objekte können mit
 *     <a href="#debug(hoverball.debug.Debug)">debug(...)</a>
 *     jederzeit gesetzt werden.
 *     <p></li>
 * </menu>
 *
 * Das Main-Panel des Unit-Hovlets ist unbenutzt. Es steht Dir zur Darstellung
 * eigener Analysen zur vollen Verfügung.
 * <p>
 * <hr>
 * <p>
 * <b><u>Beispiel einer trivialen Java-Unit</u>:</b>
 *
 *<pre>
 *import hoverball.*;
 *import hoverball.math.*;
 *&nbsp;
 *public class Clumsy extends Unit
 *{
 *   public Clumsy ()
 *   {
 *      super(null,"Clumsy",0x8888FF);  // Name "Clumsy", Farbe hellblau
 *   }
 *&nbsp;
 *   public void loop ()
 *   {
 *      Sphere sphere = new Sphere(option("world.radius"));  // Parameter auslesen...
 *      double Qmax = option("unit.charge.max");
 *      double Fmax = option("unit.engine.max");
 *&nbsp;
 *      while (look())
 *      {
 *         Puck ball = puck(BALL,0,1);       // das ist der Ball
 *         Puck goal = puck(BALL,self.t,1);  // das ist das "Tor"
 *&nbsp;
 *         if (ball == null) action(0,-Fmax,Fmax);     // sieht keinen Ball? drehe!
 *         else {                                      // sieht Ball:
 *            Vector a = (goal != null)? Vector.vec(goal.X.c,ball.X.c) :  // berechne
 *                                       Vector.vec(ball.X.c,self.X.c);   // Drehachse
 *            Matrix X = Matrix.mul(ball.X,Matrix.rot(a,(self.r+ball.r)/sphere.rad));
 *            Complex x = sphere.warp(X.c);            // X ist die Schussposition
 *&nbsp;
 *            double l = -x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // drehe und
 *            double r =  x.arg() + Math.max(0.2,1-10*Math.abs(x.arg())); // fahre zu X
 *            double q = (x.abs() < ball.r)? 0.5 : 0;  // falls auf X: Schuss!
 *&nbsp;
 *            action(q*Qmax,l*Fmax,r*Fmax);
 *         }
 *      }
 *   }
 *}
 *</pre>
 */
public abstract class Unit extends Hovlet
{
   // ---- Puck.WHATs ---------------------------------------------------------

   /**
    * Typ-Konstante: Node
    * &nbsp;(von der Klasse <a href="Puck.html#NODE">Puck</a> übernommen)
    */
   protected static final int NODE = Puck.NODE;

   /**
    * Typ-Konstante: Ball
    * &nbsp;(von der Klasse <a href="Puck.html#NODE">Puck</a> übernommen)
    */
   protected static final int BALL = Puck.BALL;

   /**
    * Typ-Konstante: Unit
    * &nbsp;(von der Klasse <a href="Puck.html#NODE">Puck</a> übernommen)
    */
   protected static final int UNIT = Puck.UNIT;


   // ---- returned by (connect...) -------------------------------------------


   /**
    * Team-Nummer des Channels
    */
   protected int channel_t = 0;

   /**
    * Spieler-Nummer des Channels
    */
   protected int channel_n = 0;


   // ---- returned by (checkin...) -------------------------------------------

   private final Hashtable options = new Hashtable();

   private double _simulator_time = 0;
   private double _unit_radius = 0;
   private double _unit_mass   = 0;
   private double _ball_radius = 0;
   private double _ball_mass   = 0;


   // ---- returned by (look...) ----------------------------------------------

   // eigentlich 'protected'" -- nur wegen Jython (!) muss alles 'public' sein!
   
   /**
    * Zeitpunkt des letzten Looks
    */
   public double time = 0;          // -> current.time

   /**
    * Energie beim letzten Look
    */
   public double energy = 0;

   /**
    * Penalty beim letzten Look
    */
   public double penalty = 0;

   /**
    * Array der Objekte des letzten Looks,
    * die gesehen oder gehört wurden
    */
   public Puck[] pucks = null;      // -> current.pucks

   /**
    * Verweis auf sich selbst im Array der Objekte des letzten Looks
    */
   public Puck self = null;         // -> pucks[i]


   // ---- Thread Group

//APPLET THREAD BUG//
   static ThreadGroup group = Thread.currentThread().getThreadGroup(); //new ThreadGroup("units");
//   static { group.setMaxPriority(Thread.NORM_PRIORITY); }

   // ----

   private NetAgent client = null;
   private Address host = null;        // Remember host (Graphical Debugging!)

   US.Connect connect = null;          // Stores (connect...)-Infos

   private int t_ = 0;
   private int n_ = 0;

   private final JPanel icon_;

   private static final int LOOK_EVERY = 500;

   // ---- Constructors -------------------------------------------------------

   static final int hashColor (String name)
   {
      return (-name.hashCode()) & 0xFFFFFF;
   }

   /**
    * öffnet eine Unit.
    *
    * @param name Name der Unit
    */
   public Unit (String name)
   {
      this(name,name);
   }

   /**
    * öffnet eine Unit.
    *
    * @param team Name des Teams
    * @param name Name der Unit
    */
   public Unit (String team, String name)
   {
      this(team,name,hashColor(name));
   }

   /**
    * öffnet eine Unit.
    *
    * @param name Name der Unit
    * @param color Farbe
    */
   public Unit (String name, int color)
   {
      this(name,name,color);
   }

   /**
    * öffnet eine Unit.
    *
    * @param team Name des Teams
    * @param name Name der Unit
    * @param color Farbe
    */
   public Unit (String team, String name, int color)
   {
      super(name,true,false);

      if (team == null) team = ""+name;

      connect = new US.Connect("",team,name,color);  // #hash will be overwritten in 'connecting'

      icon.add(icon_ = new JPanel() {
         public void paintComponent (Graphics g) { Layout.drawIcon(g,null,getSize(),new Color(connect.color)); }
      },"Center");
      if(Layout.TOOLTIPS) icon_.setToolTipText("Debug "+name);
      icon_.addMouseListener(new MouseAdapter() {  // install Debug-Listener on icon
         public void mouseClicked (MouseEvent e) {
            Boolean state = getDebugState();
            if (state == null) return;
            setDebugState(!state.booleanValue());
         }
      });
   }



   // ---- Search Pucks -------------------------------------------------------

   /**
    * Liefert den Puck mit der angegebenen Identität aus dem Puck-Array.
    *
    * @param what Typ des Objekts
    * @param t Team-Nummer
    * @param n Puck-Nummer
    * @return den gesuchten Puck (oder <b><code>null</code></b>, falls nicht gefunden)
    */
   protected Puck puck (int what, int t, int n)
   {
      if (pucks == null) return null;
      for (int i=0; i<pucks.length; ++i) if (pucks[i].equals(what,t,n)) return pucks[i];
      return null;
   }

   /**
    * Liefert den Puck mit der angegebenen Identität aus dem Puck-Array.
    *
    * @param id Identitäts-String
    * @return den gesuchten Puck (oder <b><code>null</code></b>, falls nicht gefunden)
    */
   protected Puck puck (String id)
   {
      if (pucks == null) return null;
      for (int i=0; i<pucks.length; ++i) if (pucks[i].equals(id)) return pucks[i];
      return null;
   }


   // ---- Options ------------------------------------------------------------

   /**
    * Liefert den Wert einer Simulator-Option, die beim Checkin empfangen wurde.
    * <p>
    * Es werden nicht alle Simulator-Optionen an die Units versendet.
    *
    * @param key Simulator-Option
    * @return den Wert
    */
   protected double option (String key)
   {
      try {
         String value = (String)options.get(key);
         return Double.valueOf(value).doubleValue();
      } catch(Exception e) {  // NullPointerException / NumberFormatException
         return Double.NaN;
      }
   }


   // ---- Loop Handling... ---------------------------------------------------

   private static class TimeLook extends SU.Look
   {
      double time = 0;

      TimeLook (double time, SU.Look look) { super(look.E,look.py,look.disks); this.time = time; }
   }

   private final Object LOOP = new Object();  // loop monitor (only one Thread loops!)

   private class Loop extends Thread
   {
      long base = 0;
      TimeLook current  = null;
      TimeLook received = null;

//APPLET THREAD BUG//
//      public Loop (String name) { super(Unit.group,name); }
      public Loop (String name) { super("Hoverball Unit ["+name+"]"); }

      public void run () { synchronized(LOOP) {
////System.out.println("LOOP STARTS in "+this_.title);
    	  loop();
////System.out.println("LOOP ENDS in "+this_.title);
      } }  // only one Thread loops!
   }

   private Loop loop = null;                  // Active Loop Object


   // ---- Network Listener... ------------------------------------------------

   private synchronized void look_ (long currentTimeMillis, SU.Look look)
   {
      if (client != null) client.send(new US.Look());  // reply (look) immediately!

      if (loop == null) {  // first look after break
         loop = new Loop(connect.name);
         //loop.setPriority(Thread.MIN_PRIORITY);
         loop.base = currentTimeMillis;
         loop.received = new TimeLook(0,look);
         loop.start();
      } else synchronized(loop) {
         double time = ((double)(currentTimeMillis-loop.base)) / 1000;
         loop.received = new TimeLook(time,look);  // look.next will be ignored.
         loop.notify();
      }
   }

   private synchronized void break_ ()
   {
      if (loop == null) return;  // double-(break)
      synchronized(loop) {
         loop.received = null;   // null = interrupted!
         loop.notify();
      }
      loop = null;
   }


   // ---- Unit Commands ------------------------------------------------------

   /**
    * Erbittet einen neuen Look.
    * <p>
    * Der Methode sollte in der Denk-Schleife der Unit aufgerufen werden.
    * Sie aktualisiert die Variablen <code>time</code>, <code>pucks</code> und
    * <code>self</code>.
    * <p>
    * Der Thread wird so lange angehalten, bis die Unit vom Simulator
    * einen neuen Look erhält. Wird das Spiel unterbrochen, so liefert die
    * Methode den Wert <b><code>false</code></b> und die Schleife sollte verlassen werden.
    *
    * @return <b><code>true</code></b>, falls das Spiel weitergeht
    */
   protected boolean look ()
   {
      debug(new Debug(){});  // add null-Debug to ensure a non-empty Session
      debug();               // Close current Debug Session

      int t_ = this.t_;  // save t_
      int n_ = this.n_;  // save n_  ("disconnect" may disturb!)

      Thread.yield();

      Loop loop = (Loop)Thread.currentThread();  // local loop = real running Loop

      try {
         synchronized(loop) {
            if (loop.received != null) while (true) { // loop.received == null: (break)
            	loop.wait(LOOK_EVERY);                           // will be notified
         		if (loop.current != loop.received) break;        // new look received? quit.
            	if (client != null) client.send(new US.Look());  // else: resend (look) and continue waiting!
////System.out.println("(look) missing! "+name()+" resends (look) after "+LOOK_EVERY+" millis.");
         	}
            loop.current = loop.received;
         }
      } catch(InterruptedException e) { return false; }  // should not happen (Security exception!)
      if (loop.current == null) { return false; }

      Thread.yield();

      debug();               // Setup new Debug Session

      Puck[] pucks = new Puck[loop.current.disks.length];
      for (int l=0; l<loop.current.disks.length; ++l) {
         DiskRelative i =(DiskRelative)loop.current.disks[l];
         pucks[l] = new Puck(i.what,i.t,i.n,                   // (what t n)
            (i.x==null)? null : Sphere.euler(i.x),             // X
            (i.what == BALL)? _ball_radius : (i.what == UNIT)? _unit_radius : 0,   // r
            (i.what == BALL)? _ball_mass   : (i.what == UNIT)? _unit_mass   : 1,   // m
            (i.message==null)? null : ""+i.message             // message
         );
      }

      this.time  = loop.current.time * _simulator_time;
      this.energy = loop.current.E;
      this.penalty = loop.current.py;
      this.pucks = pucks;
      this.self = null;
      for (int i=0; i<pucks.length; ++i)
         if (pucks[i].what == UNIT &&
             pucks[i].t == t_ &&
             pucks[i].n == n_) { this.self = pucks[i]; break; }

      if (this.self == null) return false;
      return true;
   }

   /**
    * Sendet eine Aktion.
    * <p>
    * Soll ein Parameter der Aktion unverändert bleiben, kann der Wert mit <code>Double.NaN</code>
    * maskiert werden.
    *
    * @param Q Ladung
    * @param F_L Antriebskraft links
    * @param F_R Anrtiebskraft rechts
    */
   protected void action (double Q, double F_L, double F_R)
   {
      action (Q,F_L,F_R,null);
   }

   /**
    * Sendet eine Botschaft.
    *
    * @param message Message
    */
   protected void action (String message)
   {
      action (Double.NaN,Double.NaN,Double.NaN,message);
   }

   /**
    * Sendet eine Aktion und eine Botschaft.
    * <p>
    * Soll ein Parameter der Aktion unverändert bleiben, kann der Wert mit <code>Double.NaN</code>
    * maskiert werden.
    *
    * @param Q Ladung
    * @param F_L Antriebskraft links
    * @param F_R Anrtiebskraft rechts
    * @param message Message
    */
   protected void action (double Q, double F_L, double F_R, String message)
   {
      if (client == null) return;
      client.send(new US.Action(Q,F_L,F_R,message));
   }

   /**
    * Frei für die Implementierung der Denk-Schleife.
    */
   protected abstract void loop ();


   // ---- Debugging...

   private void debug ()                          // Setup new Debug Session!
   {
      if (client == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,null,null);
   }

   /**
    * Setzt ein Debug-Objekt.
    *
    * @param debug Debug-Objekt
    */
   protected void debug (Debug debug)             // Add to Debug List
   {
      if (client == null) return;
      if (debug == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,debug,null);
   }

   /**
    * Setzt ein Debug-Objekt mit angegebener Farbe.
    *
    * @param debug Debug-Objekt
    * @param color Farbe
    */
   protected void debug (Debug debug, int color)  // Add to Debug List with specified color
   {
      if (client == null) return;
      if (debug == null) return;

      if (debugger != null) debugger.addDebug(host,t_,n_,debug,new Color(color));
   }

   protected static String format (double d)
   {
      if (Complex.format == null) return ""+d;
      else return Complex.format.format(d);
   }

   protected static String format (Complex c) { return c.toString(); }
   protected static String format (Vector v)  { return v.toString(); }
   protected static String format (Matrix M)  { return M.toString(); }


   // ---- Setup Debuggers ----------------------------------------------------

   Controller debugger = null;  // Setup by Controller

   String name () { return ""+connect.name; }
   boolean owns (Address server, Channel channel)
   {
      return server.equals(host) &&                    // same Server
             channel != null && connect != null &&     // connected
             (t_ == channel.t && n_ == channel.n);     // same Channel
   }

   Boolean getDebugState ()  // by Team
   {
      if (debugger == null) return null;
      else return debugger.getDebugState(this);
   }

   void setDebugState (boolean debug)  // by Team
   {
      if (debugger == null) return;
      else debugger.debug(this,debug);
   }

   // ---- Connect & Disconnect -----------------------------------------------

   java.util.Timer pinger = null;

   /**
    * [Implementierung]
    */
   protected synchronized final boolean connecting (String server)
   {
      Address address = new Address(server);
      if (!address.isValid()) return false;

      host = address;  // Remember Host before connecting! (Follow & Graphical Debugging)
      boolean success = false;
      try
      {
         client = new UnitClient(host);  // must be defined for first received serials
         connect.hash = host.getHash();
    	   success = client.connect(host,connect);
         // host's #hash will be changed externally by 'client.connect'
      }
      catch (IOException e) {}

      if (success) {
         servertext = ""+host; // ggf. mit zugewiesenem #hash vom Server
         Application.servertext_with_hash = servertext;
      }

      if (success) { // install Pinger
    	   pinger = new java.util.Timer("Hoverball Ping (Unit)");
    	   pinger.schedule(new java.util.TimerTask() {
    	      public void run () { client.send(new US.Ping()); }
    	   },0,Socket.PING_EVERY);
      }

      if (success) {
         return true;
      } else {
         if (client!=null)
         {
            client.close();
            client = null;
         }
         host = null;
         return false;
      }
   }

   /**
    * [Implementierung]
    */
   protected synchronized final void disconnecting ()
   {
      break_();

      pinger.cancel(); // cancel Pinger
      pinger = null;

      client.close();
      client = null;
      host = null;
      channel_t = t_ = 0;
      channel_n = n_ = 0;
   }


   // ---- Network Protocol: Unit's Client to Server -------- [UnitClient] ----

   private final Unit this_ = this;

   private class UnitClient extends NetAgent
   {
      // ---- Bye on close() --------------------------------------------------

      private boolean bye = true;  // send bye on close()
      private boolean closed = false;

      public UnitClient(Address serverAddress) throws IOException
      {
    	  super(serverAddress);
      }

      public synchronized void close ()
      {
         if (closed) return;
         if (bye) this.send(new US.Bye());
         super.close();
         closed = true;
      }

      // ---- Receive ---------------------------------------------------------

      public void receive (Address sender, String message)
      {
//     	 System.out.println("Unit "+this+" got "+message);
         Serial serials = SU.deserialize(message);

         Serial serial_ = serials;
         while (serial_ != null) synchronized(this_) {
            Serial serial = serial_;
            serial_ = serial_.next;
            serial.next = null;
            if (client == null) continue;  // disconnected? ignore!

            // ---- (connect 3 2 ...) -----------------------------------------

            if (serial instanceof SU.Connect)
            {
               host.setHash(((SU.Connect)serial).hash);
               SU.Connect connect = (SU.Connect)serial;
               channel_t = t_ = connect.t;
               channel_n = n_ = connect.n;
            }

            // ---- (checkin...) ----------------------------------------------

            if (serial instanceof SU.Checkin)
            {
               SU.Checkin checkin = (SU.Checkin)serial;

               options.clear(); int k=0;
               for (int i=0; i<Constants.CONSTANTS; ++i) if (Constants.isForUnit(i))
                  options.put(Constants.getKey(i),checkin.options[k++]);

               _simulator_time = option(Constants.getKey(Constants.SIMULATOR_TIME));
               _unit_radius    = option(Constants.getKey(Constants.UNIT_RADIUS));
               _unit_mass      = option(Constants.getKey(Constants.UNIT_MASS));
               _ball_radius    = option(Constants.getKey(Constants.BALL_RADIUS));
               _ball_mass      = option(Constants.getKey(Constants.BALL_MASS));
            }

            // ---- (look 2.3 19 (ball ..) (unit ..)) --------------------------

            else if (serial instanceof SU.Look)
            {
               look_(System.currentTimeMillis(),(SU.Look)serial);
            }

            // ---- (break) ---------------------------------------------------

            else if (serial instanceof SU.Break)
            {
               break_();
            }

            // ---- (bye) -----------------------------------------------------

            else if (serial instanceof SU.Bye)
            {
               bye = false;  // don't send bye yourself!
               disconnect();
            }
         }
      }
   }
}
