/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.debug;

import hoverball.layout.Layout;  // Screen Drawing
import hoverball.math.*;

import java.awt.*;


/**
 * Die Klasse Debug stellt die Basis-Klasse für das <b>Graphical Debugging</b> dar.
 * Beim Programmieren von Java-Units hast Du die Möglichkeit, mit dem
 * <a href="../Unit.html#debug(hoverball.debug.Debug)">debug(...)</a>-Befehl
 * grafische Elemente auf dem Screen darzustellen.
 * Diese Elemente sind Objekte der Klasse Debug (oder einer abgeleiteten Klasse) und
 * können auf vielfältige Weise erzeugt und definiert werden:
 * <p>
 * <ul>
 * <li>Du benutzt die vordefinierten Debugs aus diesem Package.
 *     <p>
 * <li>Debugs sind rekursiv verschachtelbar. Jedes Debug enthält ein (zunächst leeres)
 *     Array von Sub-Debugs
 *     <a href="#debugs">debugs</a>,
 *     welches mit beliebig vielen weiteren Debugs belegt werden kann.
 *     Somit entsteht ein rekursiver Baum von Debug-Objekten. Ein Debug
 *     zeichnet zuerst sich selbst und anschließend das Array seiner Sub-Debugs.
 *     <p>
 *     Die Zeichenfarbe der Debugs wird dabei vererbt. Die Sub-Debugs eines blauen Debugs
 *     sind ebenfalls alle blau, solange sie nicht selbst eine eigene Zeichenfarbe
 *     definieren. Mit der Variablen
 *     <a href="#color">color</a>
 *     hast Du die Möglichkeit, eine Zeichenfarbe für Dein Debug anzugeben oder
 *     die durchlässige Zeichenfarbe <b><code>null</code></b> zu wählen.
 *     <p>
 *     Du kannst also Dein Debug einfach als Kombination von bekannten Debugs
 *     definieren.
 *     Folgendes Debug zeichnet ein blaues Dreieck:
 *<pre>
 *class Triangle extends Debug
 *{
 *   Triangle (Vector a, Vector b, Vector c)
 *   {
 *      debugs = new Debug[] { new Line(a,b), new Line(b,c), new Line(c,a) };
 *      color = Color.blue;
 *   }
 *}
 *</pre>
 *     <p>
 * <li>Du leitest die Klasse Debug ab und definierst Dein eigenes Debug, indem Du die Methode
 *     <a href="#paint(java.awt.Graphics, java.awt.Color, java.awt.Color, double, hoverball.math.Sphere, hoverball.math.Matrix, boolean)">paint(...)</a>
 *     überschreibst.
 *     Debugs werden relativ zur Unit gespeichert und erst im Moment des Zeichnens
 *     in absolute Koordinaten übersetzt.
 *     Folgendes Debug markiert einen Punkt mit einem kleinen Kreis:
 *<pre>
 *class O extends Debug
 *{
 *   private Vector o;                            // relative Position zur Unit
 *&nbsp;
 *   public O (Vector o) { this.o = vector(o); }  // speichere relative Position
 *&nbsp;
 *   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front)
 *   {
 *      Vector p = Vector.mul(o,base);            // berechne absolute Position
 *      if (front ^ p.z > 0) return;              // falsche Seite? return!
 *&nbsp;
 *      Point P = scale(scale,p);                 // skaliere in AWT-Koordinaten
 *      g.setColor(color(color,globe,p.z));       // setze Zeichenfarbe
 *      g.drawOval(P.x-6,P.y-6,12,12);            // zeichne den Kreis!
 *   }
 *}
 *</pre>
 * </ul>
 * <p>
 *
 * Die Farbe des Debugs wird bestimmt durch die Farbangabe im debug(...)-Befehl
 * der Unit oder durch die Farbe der Unit selbst, falls der Aufruf ohne Farbangabe
 * erfolgte.
 *
 */
public abstract class Debug
{
   // ---- static calculation tools

   /**
    * Liest und korrigiert einen Vektor.
    * <p>
    * Diese Methode wird sinnvoll im Konstruktor eines Debugs aufgerufen,
    * wenn es darum geht, Positionsvektoren zu übernehmen und zu speichern.
    * Der übergebene Vektor wird kopiert und auf die Länge 1 normiert -
    * somit ist er auf die Skalierung der
    * <a href="#paint(java.awt.Graphics, java.awt.Color, java.awt.Color, double, hoverball.math.Sphere, hoverball.math.Matrix, boolean)">paint(...)</a>-Methode
    * genau zugeschnitten.
    * Falls der Vektor den Wert (0,0,0) oder <b><code>null</code></b> hatte,
    * liefert die Methode den <i>"sphärischen Nullvektor"</i> (0,0,1).
    *
    * @param v Vektor
    * @return den korrigierten Vektor
    */
   public static Vector vector (Vector v)
   {
      if (v == null || v.zero()) return new Vector(0,0,1);
      else return Vector.norm(v);
   }

   /**
    * Ermittelt die AWT-Koordinaten eines Vektors für eine bestimmte Skalierung.
    * <p>
    * Der Hoverball-Screen ist so orientiert, dass
    * die <i>x</i>-Achse nach oben, die <i>y</i>-Achse nach links und
    * die <i>z</i>-Achse nach vorne ("aus dem Bildschirm heraus") zeigt.
    * Diese Methode projiziert das 3D-Koordinatensystem der Sphäre auf das
    * 2D-Koordinatensystem eines AWT-Graphics-Contexts. Dabei muss angegeben
    * werden, mit wie vielen Pixeln die Einheitslänge skaliert werden soll.
    * Der Koordinatenursprung wird nicht verschoben.
    *
    * @param scale Skalierung (in Pixel)
    * @param v Vektor
    * @return die AWT-Koordinaten
    */
   public static Point scale (double scale, Vector v)
   {
      return Layout.scale(scale,v);
   }

   /**
    * Ermittelt die Zeichenfarbe für eine bestimmte Tiefe.
    * <p>
    * Um einen perspektivischen Effekt bei der Darstellung der Sphäre zu erzielen,
    * werden Objekte, die weiter im Hintergrund liegen, etwas blasser gezeichnet.
    * Diese Methode ermittelt die reale Zeichenfarbe einer Farbe
    * für eine bestimmte Tiefe. Der Tiefenwert soll zwischen -1 und 1 liegen
    * und der <i>z</i>-Koordinate der normierten Positionsvektoren entsprechen.
    *
    * @param color Farbe des Debug
    * @param globe Farbe der Sphäre
    * @param z Tiefe
    * @return die reale Zeichenfarbe
    */
   public static Color color (Color color, Color globe, double z)  // -1..1
   {
      return Layout.color(color,globe,z);
   }


   // ---- class Debug

   /**
    * Farbe des Debugs
    */
   public Color color = null;

   /**
    * Array der Sub-Debugs
    */
   public Debug[] debugs = null;  // contains sub-debugs


   /**
    * Leerer Konstruktor.
    */
   public Debug () {}

   /**
    * Frei für das Zeichnen des Debugs auf einen AWT-Graphics-Context.
    * <p>
    * Diese Methode ist für das Zeichnen des Debugs mit richtiger Perspektive
    * und Farbabstufung verantwortlich. Hierfür können die Methoden
    * <a href="#scale(double, hoverball.math.Vector)">scale(...)</a>
    * und
    * <a href="#color(java.awt.Color, java.awt.Color, double)">color(...)</a>
    * sowie die Klassen aus dem Package
    * <a href="../math/package-summary.html">hoverball.math</a>
    * gute Dienste leisten.
    *
    * Beachte noch Folgendes:
    * <p>
    * <ul>
    * <li>Das Koordinatensystem des Graphics-Contexts ist bereits so verschoben,
    *     dass der (0,0)-Punkt in der Mitte der Sphäre liegt.
    *     <p>
    * <li>Das Skalierungsmaß ist so gewählt, dass es den Sphären-Radius (anstatt
    *     der Hoverball-Einheitslänge) in Pixeln angibt.
    *     Das bedeutet, dass Vektoren der Länge 1 unabhängig von der Größe der Sphäre
    *     gleich auf die Sphären-Oberfläche skaliert werden.
    *     <p>
    * <li>Die Perspektive der Unit wird mit der Matrix <code>base</code>
    *     an die Methode übergeben. Um Vektoren aus der Sicht der Unit
    *     in die richtige Perspektive zu bringen, reicht eine einfache
    *     Multiplikation mit der Matrix <code>base</code>.
    *     <p>
    * <li>Vorderseite und Rückseite der Sphäre werden getrennt gezeichnet,
    *     um den richtigen Überlappungseffekt von Vordergrund vor Hintergrund
    *     zu erzielen.
    *     Die Methode paint(...) wird jeweils zweimal aufgerufen. Dabei wird
    *     ein Parameter <code>front</code> übergeben, der angibt, ob gerade die
    *     Vorderseite oder die Rückseite der Sphäre gezeichnet wird, und der
    *     beim Zeichnen des Debugs berücksichtigt werden sollte.
    * </ul>
    *
    * @param g Graphics-Context
    * @param color Farbe des Debugs
    * @param globe Farbe der Sphäre
    * @param scale Skalierung (Sphären-Radius in Pixeln)
    * @param sphere Sphäre des Spiels
    * @param base Matrix der Perspektive
    * @param front Vorderseite?
    */
   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front) {}
}
