/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.debug;

import hoverball.layout.Layout;
import hoverball.math.*;

import java.awt.*;


/**
 * Gibt einen Text aus.
 */
public class Text extends Debug
{
   private final Vector p;
   private final String text;
   private final double line;
   private final int align;

   /**
    * Erzeugt ein Text-Debug.
    *
    * @param p Position
    * @param text Text
    */
   public Text (Vector p, String text) { this(p,text,0,0); }

   /**
    * Erzeugt ein Text-Debug mit Zeilennummer.
    * <p>
    * Die Zeilennummer verschiebt den Text unabhängig von der Perspektive
    * nach oben oder nach unten. Somit ist es möglich, an einem
    * Punkt einen mehrzeiligen Kommentar auszugeben, ohne dass sich der
    * der Text mit der Perspektive verzerrt.
    *
    * @param p Position
    * @param text Text
    * @param line Zeile
    */
   public Text (Vector p, String text, double line) { this(p,text,line,0); }

   /**
    * Erzeugt ein Text-Debug mit Zeilennummer und horizontaler Ausrichtung.
    * <p>
    * Die Zeilennummer verschiebt den Text unabhängig von der Perspektive
    * nach oben oder nach unten. Somit ist es möglich, an einem
    * Punkt einen mehrzeiligen Kommentar auszugeben, ohne dass sich der
    * der Text mit der Perspektive verzerrt.
    * <p>
    * Bei einem Ausrichtungswert <i>x</i> &lt; 0 wird der Text rechtsbündig,
    * bei <i>x</i> = 0 zentriert und bei <i>x</i> &gt; 0 linksbündig ausgegeben.
    *
    * @param p Position
    * @param text Text
    * @param line Zeile
    * @param align Ausrichtung
    */
   public Text (Vector p, String text, double line, int align)
   {
      this.p = vector(p);
      this.text = (text == null)? "" : ""+text;
      this.line = line;
      this.align = align;
   }

   /**
    * [Implementierung]
    */
   public void paint (Graphics g, Color color, Color globe, double scale, Sphere sphere, Matrix base, boolean front)
   {
      Vector v = Vector.mul(p,base);
      if (front ^ v.z > 0) return;

      Point X = scale(scale,v);
      int height = g.getFontMetrics().getAscent();
      int spacing = g.getFontMetrics().getHeight();

      g.setColor(color(color,globe,v.z));

      String[] lines = Layout.lines(text);
      for (int i=0; i<lines.length; ++i) {
         int width = (align > 0)? 0 :
                     (align < 0)? g.getFontMetrics().stringWidth(lines[i]) :
                                  g.getFontMetrics().stringWidth(lines[i])/2;
         g.drawString(lines[i],X.x-width,X.y+height/2+(int)((line+i)*spacing)-1);
      }
   }
}
