/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;


/**
 * Die Klasse Complex repräsentiert eine komplexe Zahl
 * der zweidimensionalen komplexen Zahlenebene.
 * <p>
 * Im Folgenden bedeuten: <br>
 * &nbsp; &nbsp; <i>a</i>, <i>b</i>, <i>c</i> &nbsp;-&nbsp; komplexe Zahlen<br>
 * &nbsp; &nbsp; <i>x</i>, <i>y</i>, <i>z</i> &nbsp;-&nbsp; reelle Zahlen (als Koordinaten)<br>
 * &nbsp; &nbsp; <i>r</i> &nbsp;-&nbsp; reelle Zahl<br>
 * &nbsp; &nbsp; <i>v</i> &nbsp;-&nbsp; Vektor<br>
 * &nbsp; &nbsp; <b>c</b> &nbsp;-&nbsp; die komplexe Zahl des Complex-Objekts selbst
 * <p>
 * Operator Overloading für <i>Groovy</i>, <i>Scala</i>, <i>Kotlin</i> und <i>Jython</i> wird unterstützt.
 */
public final class Complex
{
   /**
    * DecimalFormat-Objekt zur String-Konvertierung.
    * <p>
    * Dieses Objekt wird zur Formatierung aller Objekte dieses Packages benutzt.
    * Voreingestellt ist das Format <code>0.###</code> mit <code>Locale.US</code>.
    */
   public static DecimalFormat format = new DecimalFormat("0.###",new DecimalFormatSymbols(java.util.Locale.US));

   static String format (double d)
   {
      if (format == null) return ""+d;
      else return format.format(d);
   }

   /**
    * <i>x</i>-Koordinate der komplexen Zahl.
    */
   public final double x;

   /**
    * <i>y</i>-Koordinate der komplexen Zahl.
    */
   public final double y;


   // ---- constructors & bare bones ------------------------------------------

   /**
    * Erzeugt eine komplexe Zahl mit den Koordinaten (0,0).
    */
   public Complex ()                   { this.x = 0;
                                         this.y = 0; }

   /**
    * Erzeugt eine komplexe Zahl mit den Koordinaten (<i>x</i>,<i>y</i>).
    *
    * @param x <i>x</i>-Koordinate
    * @param y <i>y</i>-Koordinate
    */
   public Complex (double x, double y) { this.x = x;
                                         this.y = y; }

   /**
    * Erzeugt eine komplexe Zahl mit den Koordinaten (<i>x</i>,0).
    *
    * @param x <i>x</i>-Koordinate
    */
   public Complex (double x) { this.x = x;
                               this.y = 0; }

   /**
    * Erzeugt eine komplexe Zahl gleich einer anderen komplexen Zahl.
    *
    * @param c zu kopierende komplexe Zahl
    */
   public Complex (Complex c)          { this.x =(c==null)? 0 : c.x;
                                         this.y =(c==null)? 0 : c.y; }

   /**
    * Erzeugt eine komplexe Zahl aus einem Vektor.
    * <p>
    * Die <i>z</i>-Koordinate des Vektors wird dabei ignoriert.
    *
    * @param v Vektor
    */
   public Complex (Vector v)           { this.x =(v==null)? 0 : v.x;
                                         this.y =(v==null)? 0 : v.y; }

   /**
    * Berechnet den Hash-Code
    * &nbsp;[
    * = <i><code>####</code></i>
    * ].
    *
    * @return den Hash-Code
    */
   public int hashCode ()
   {
      return new Double(x+y).hashCode();
   }

   /**
    * String-Repräsentation
    * &nbsp;[
    * = "..."
    * ].
    *
    * @return den String
    */
   public String toString ()
   {
      return "("+format(x)+","+format(y)+")";
   }

   /**
    * Vergleicht mit einer komplexen Zahl
    * &nbsp;[
    * <b>c</b> = <i>c</i> ?
    * ].
    *
    * @param c komplexe Zahl
    * @return die Gleichheit
    */
   public boolean equals (Object c)
   {
      if (c == null) return false;
      if (! (c instanceof Complex)) return false;

      Complex c_ = (Complex)c;
      return (x == c_.x && y == c_.y);
   }

   /**
    * Vergleicht mit einer komplexen Zahl
    * &nbsp;[
    * <b>c</b> = (<i>x</i>,<i>y</i>) ?
    * ].
    *
    * @param x reelle Zahl
    * @param y reelle Zahl
    * @return die Gleichheit
    */
   public boolean equals (double x, double y)
   {
      return (this.x == x && this.y == y);
   }


   // ---- object methods -----------------------------------------------------

   /**
    * Setzt die Koordinaten
    * &nbsp;[
    * <b>c</b> = (<i>x</i>,<i>y</i>)
    * ].
    *
    * @param x reelle Zahl
    * @param y reelle Zahl
    * @return <b>this</b>
    */
//   public final Complex set (double x, double y) { this.x=x; this.y=y; return this;}

   /**
    * Setzt die Koordinaten
    * &nbsp;[
    * <b>c</b> = <i>c</i>
    * ].
    *
    * @param c komplexe Zahl
    * @return <b>this</b>
    */
//   public final Complex set (Complex c) { x =c.x; y =c.y; return this;}

   /**
    * Addiert eine komplexe Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> + <i>b</i>
    * ].
    *
    * @param b komplexe Zahl
    * @return <b>this</b>
    */
   public final Complex add (Complex b) { return new Complex(x+b.x,y+b.y); }

   /**
    * Subtrahiert eine komplexe Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> - <i>b</i>
    * ].
    *
    * @param b komplexe Zahl
    * @return <b>this</b>
    */
   public final Complex sub (Complex b) { return new Complex(x-b.x,y-b.y); }

   /**
    * Multipliziert mit einer reellen Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> <i>r</i>
    * ].
    *
    * @param r reelle Zahl
    * @return <b>this</b>
    */
   public final Complex mul (double r)  { return new Complex(x*r,y*r); }


   /**
    * Multipliziert mit einer komplexen Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> <i>b</i>
    * ].
    *
    * @param b komplexe Zahl
    * @return <b>this</b>
    */
   public final Complex mul (Complex b)
   {
	  return new Complex(x*b.x-y*b.y,x*b.y+y*b.x);
   }

   /**
    * Dividiert durch eine reelle Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> / <i>r</i>
    * ].
    *
    * @param r reelle Zahl
    * @return <b>this</b>
    */
   public final Complex div (double r)  { return new Complex(x/r,y/r); }


   /**
    * Dividiert durch eine komplexe Zahl
    * &nbsp;[
    * <b>c</b> = <b>c</b> / <i>b</i>
    * ].
    *
    * @param b komplexe Zahl
    * @return <b>this</b>
    */
   public final Complex div (Complex b)
   {
	  return this.mul(b.inv());
   }

   /**
    * Bildet die Kehrzahl
    * &nbsp;[
    * <b>c</b> = 1 / <b>c</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex inv ()
   {
      double q = x*x+y*y;
      return new Complex(x/q,-y/q);
   }

   /**
    * Bildet die Gegenzahl
    * &nbsp;[
    * <b>c</b> = -<b>c</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex neg ()  { return new Complex(-x,-y); }

   /**
    * Bildet die Konjugation
    * &nbsp;[
    * <b>c</b> = <u><b>c</b></u>
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex conj () { return new Complex(x,-y); }

   /**
    * Testet auf null
    * &nbsp;[
    * <b>c</b> = 0 ?
    * ].
    *
    * @return <b>true</b>, falls diese komplexe Zahl gleich (0,0)
    */
   public final boolean zero () { return abs2() == 0; }

   /**
    * Berechnet das Argument
    * &nbsp;[
    * = &lt;(<b>c</b>)
    * ].
    *
    * Ist die komplexe Zahl gleich null, liefert diese Funktion den Wert null.
    *
    * @return das Argument
    */
   public final double arg ()   { return Math.atan2(y,x); }  // Math.atan2(0,0) == 0

   /**
    * Berechnet den Absolutbetrag
    * &nbsp;[
    * = | <b>c</b> |
    * ].
    *
    * @return den Absolutbetrag
    */
   public final double abs ()   { return Math.sqrt(x*x+y*y); }

   /**
    * Berechnet den Absolutbetrag im Quadrat
    * &nbsp;[
    * = | <b>c</b> |^2
    * ].
    *
    * @return den Absolutbetrag im Quadrat
    */
   public final double abs2 ()  { return x*x+y*y; }

   /**
    * Normiert den Absolutbetrag auf eins
    * &nbsp;[
    * <b>c</b> = <b>c</b> / | <b>c</b> |
    * ].
    *
    * @return <b>this</b>
    */
   public final Complex norm () { double r=abs(); return new Complex(x/r,y/r); }


   // ---- class methods ------------------------------------------------------

   /**
    * Addiert zwei komplexe Zahlen
    * &nbsp;[
    * = <i>a</i> + <i>b</i>
    * ].
    *
    * @param a komplexe Zahl
    * @param b komplexe Zahl
    * @return die Summe
    */
   public final static Complex add (Complex a, Complex b)
   {
      return new Complex(a.x+b.x,a.y+b.y);
   }

   /**
    * Subtrahiert zwei komplexe Zahlen
    * &nbsp;[
    * = <i>a</i> - <i>b</i>
    * ].
    *
    * @param a komplexe Zahl
    * @param b komplexe Zahl
    * @return die Differenz
    */
   public final static Complex sub (Complex a, Complex b)
   {
      return new Complex(a.x-b.x,a.y-b.y);
   }

   /**
    * Berechnet die Gegenzahl
    * &nbsp;[
    * = -<i>c</i>
    * ].
    *
    * @param c komplexe Zahl
    * @return die Gegenzahl
    */
   public final static Complex neg (Complex c)
   {
      return new Complex(-c.x,-c.y);
   }

   /**
    * Berechnet die Kehrzahl
    * &nbsp;[
    * = 1 / <i>c</i>
    * ].
    *
    * @param c komplexe Zahl
    * @return die Kehrzahl
    */
   public final static Complex inv (Complex c)
   {
      return new Complex(c).inv();
   }

   /**
    * Berechnet die Konjugation
    * &nbsp;[
    * = <u><i>c</i></u>
    * ].
    *
    * @param c komplexe Zahl
    * @return die Konjugation
    */
   public final static Complex conj (Complex c)
   {
      return new Complex(c.x,-c.y);
   }

   /**
    * Berechnet die Normierung des Absolutbetrages auf eins
    * &nbsp;[
    * = <i>c</i> / | <i>c</i> |
    * ].
    *
    * @param c komplexe Zahl
    * @return die Normierung
    */
   public final static Complex norm (Complex c)
   {
      return new Complex(c).norm();
   }

   /**
    * Multipliziert eine komplexe Zahl mit einer reellen Zahl
    * &nbsp;[
    * = <i>c r</i>
    * ].
    *
    * @param c komplexe Zahl
    * @param r reelle Zahl
    * @return das Produkt
    */
   public final static Complex mul (Complex c, double r)  // real multiplication
   {
      return new Complex(c.x*r,c.y*r);
   }

   /**
    * Multipliziert zwei komplexe Zahlen
    * &nbsp;[
    * = <i>a b</i>
    * ].
    *
    * @param a komplexe Zahl
    * @param b komplexe Zahl
    * @return das Produkt
    */
   public final static Complex mul (Complex a, Complex b)  // complex multiplication
   {
      return new Complex(a.x*b.x-a.y*b.y,a.x*b.y+a.y*b.x);
   }

   /**
    * Berechnet das Skalarprodukt zweier komplexer Zahlen
    * &nbsp;[
    * = &lt; <i>a</i>, <i>b</i> &gt;
    * ].
    *
    * @param a komplexe Zahl
    * @param b komplexe Zahl
    * @return das Skalarprodukt
    */
   public final static double dot (Complex a, Complex b)  // dot product
   {
      return a.x*b.x+a.y*b.y;
   }

   /**
    * Berechnet das dreidimensionale Vektorprodukt zweier Vektoren
    * &nbsp;[
    * = <i>a</i> x <i>b</i>
    * ].
    * <p>
    * Formel:
    * &nbsp;
    * (<i>a.x</i>, <i>a.y</i>, 0) x (<i>b.x</i>, <i>b.y</i>, 0) = (0, 0, <i>c.z</i>)
    *
    * @param a komplexe Zahl
    * @param b komplexe Zahl
    * @return die <i>z</i>-Koordinate des Vektorprodukts
    */
   public final static double vec (Complex a, Complex b)  // vector product: (a.x,a.y,0) x (b.x,b.y,0) = (0,0,*)
   {
      return a.x*b.y-a.y*b.x;
   }

   /**
    * Berechnet das dreidimensionale Vektorprodukt zweier Vektoren
    * &nbsp;[
    * = <i>a</i> x <i>b</i>
    * ].
    * <p>
    * Formel:
    * &nbsp;
    * (0, 0, <i>a.z</i>) x (<i>b.x</i>, <i>b.y</i>, 0) = (<i>c.x</i>, <i>c.y</i>, 0)</i>
    *
    * @param a_z reelle Zahl
    * @param b komplexe Zahl
    * @return die (<i>x</i>,<i>y</i>)-Koordinaten des Vektorprodukts
    */
   public final static Complex vec (double a_z, Complex b)  // vector product: (0,0,a_z) x (b.x,b.y,0) = (*,*,0)
   {
      return new Complex(-a_z*b.y,a_z*b.x);
   }

   /**
    * Berechnet den komplexen Exponentialwert
    * &nbsp;[
    * = exp i<i>y</i>
    * ].
    *
    * @param y reelle Zahl
    * @return den Exponentialwert
    */
   public final static Complex expi (double y)  // exp(i*y)
   {
      return new Complex(Math.cos(y),Math.sin(y));
   }

   /**
    * Berechnet den komplexen Exponentialwert
    * &nbsp;[
    * = exp <i>a</i>
    * ].
    *
    * @param a komplexe Zahl
    * @return den Exponentialwert
    */
   public final static Complex exp (Complex a)  // exp(a)
   {
      return Complex.expi(a.y).mul(Math.exp(a.x));
   }

// [oo]
// =========================================================================

   public boolean equals (double r) { return equals(valueOf(r)); }

// =========================================================================

/*
 * java-oo
 *
 * [!] '==' / '!=' not overloaded, use c.equals(b) instead
 * [!] '+=' etc.   not overloaded, use c = c + b   instead
 * [!] no unary +
 *
 */

// public final Complex add            (Complex b) { return add(b); }                 // +
   public final Complex add            (double r)  { return add(valueOf(r)); }
   public final Complex addRev         (double r)  { return add(valueOf(r)); }
   //public final Complex positive       ()          { return this; }
   public final Complex subtract       (Complex b) { return sub(b); }                 // -
   public final Complex subtract       (double r)  { return sub(valueOf(r)); }
   public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex negate         ()          { return neg(); }
   public final Complex multiply       (Complex b) { return mul(b); }                 // *
   public final Complex multiply       (double r)  { return mul(r); }
   public final Complex multiplyRev    (double r)  { return mul(r); }
   public final Complex divide         (Complex b) { return div(b); }                 // /
   public final Complex divide         (double r)  { return div(r); }
   public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex not            ()          { return inv(); }                  // ~
   public static final Complex valueOf (double r)  { return new Complex(r,0); }       // c = expr

// =========================================================================

/*
 * Groovy
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Complex plus           (Complex b) { return add(b); }               // +
   public final Complex plus           (double r)  { return add(valueOf(r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex positive       ()          { return this; }
   public final Complex minus          (Complex b) { return sub(b); }               // -
   public final Complex minus          (double r)  { return sub(valueOf(r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex negative       ()          { return neg(); }
// public final Complex multiply       (Complex b) { return mul(b); }               // *
// public final Complex multiply       (double r)  { return mul(r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
// public final Complex div            (Complex b) { return div(b); }               // /
// public final Complex div            (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex bitwiseNegate  ()          { return inv(); }                // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

/*
 * Kotlin
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 * [!] ! instead of ~
 * [!] Kotlin does not support automatic type cast from int to double
 *
 */

// public final Complex plus           (Complex b) { return add(b); }               // +
// public final Complex plus           (double r)  { return add(valueOf(r)); }
   public final Complex plus           (int r)     { return add(valueOf((double)r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex unaryPlus      ()          { return this; }
// public final Complex minus          (Complex b) { return sub(b); }               // -
// public final Complex minus          (double r)  { return sub(valueOf(r)); }
   public final Complex minus          (int r)     { return sub(valueOf((double)r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex unaryMinus     ()          { return neg(); }
   public final Complex times          (Complex b) { return mul(b); }               // *
   public final Complex times          (double r)  { return mul(r); }
   public final Complex times          (int r)     { return mul((double)r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
// public final Complex div            (Complex b) { return div(b); }               // /
// public final Complex div            (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
// public final Complex not            ()          { return inv(); }                // ! statt ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

/*
 * Jython
 *
 * [?] no limitiations, everything's fine!
 *
 */

   public final Complex __add__        (Complex b) { return add(b); }                 // +
   public final Complex __add__        (double r)  { return add(valueOf(r)); }
   public final Complex __radd__       (double r)  { return add(valueOf(r)); }
   public final Complex __pos__        ()          { return this; }
   public final Complex __sub__        (Complex b) { return sub(b); }                 // -
   public final Complex __sub__        (double r)  { return sub(valueOf(r)); }
   public final Complex __rsub__       (double r)  { return neg().add(valueOf(r)); }
   public final Complex __neg__        ()          { return neg(); }
   public final Complex __mul__        (Complex b) { return mul(b); }                 // *
   public final Complex __mul__        (double r)  { return mul(r); }
   public final Complex __rmul__       (double r)  { return mul(r); }
   public final Complex __div__        (Complex b) { return div(b); }                 // /
   public final Complex __div__        (double r)  { return div(r); }
   public final Complex __rdiv__       (double r)  { return inv().mul(r); }
   public final Complex __invert__     ()          { return inv(); }                  // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }       // c = expr

// =========================================================================

/*
 * Scala
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Complex $plus          (Complex b) { return add(b); }               // +
   public final Complex $plus          (double r)  { return add(valueOf(r)); }
   //public final Complex addRev         (double r)  { return add(valueOf(r)); }
   public final Complex unary_$plus    ()          { return this; }
   public final Complex $minus         (Complex b) { return sub(b); }               // -
   public final Complex $minus         (double r)  { return sub(valueOf(r)); }
   //public final Complex subtractRev    (double r)  { return neg().add(valueOf(r)); }
   public final Complex unary_$minus   ()          { return neg(); }
   public final Complex $times         (Complex b) { return mul(b); }               // *
   public final Complex $times         (double r)  { return mul(r); }
   //public final Complex multiplyRev    (double r)  { return mul(r); }
   public final Complex $div           (Complex b) { return div(b); }               // /
   public final Complex $div           (double r)  { return div(r); }
   //public final Complex divideRev      (double r)  { return inv().mul(r); }
   public final Complex unary_$tilde   ()          { return inv(); }                // ~
   //public static final Complex valueOf (double r)  { return new Complex(r,0); }     // c = expr

// =========================================================================

}
