/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;


/**
 * Die Klasse Sphere repräsentiert die Sphäre einer dreidimensionalen Kugel.
 * <p>
 * Eine Sphäre wird lediglich durch ihren Radius definiert.
 * Sie ist im Gegensatz zu Matrix, Vektor und komplexer Zahl ein eher starres
 * Objekt. Mit ihr kann nicht viel "gerechnet" werden, jedoch stellt die Klasse
 * Sphere eine kleine Bibliothek sphärischer Operationen zu Verfügung, die
 * den Umgang mit Matrizen, Vektoren und komplexen Zahlen sehr erleichtert.
 */
public final class Sphere
{
   // ---- class Sphere -------------------------------------------------------

   /**
    * Radius der Sphäre
    */
   public final double rad;

   /**
    * Umfang der Sphäre
    */
   public final double per;


   /**
    * Erzeugt ein Sphäre mit angegebenem Radius.
    *
    * @param r Radius
    */
   public Sphere (double r) { rad = r; per = 2*Math.PI*r; }



   /**
    * Berechnet den Hash-Code.
    *
    * @return den Hash-Code
    */
   public int hashCode ()
   {
      return new Double(rad).hashCode();
   }

   /**
    * String-Repräsentation.
    *
    * @return den String
    */
   public String toString ()
   {
      return ""+getClass().getName()+"[rad="+Complex.format(rad)+"]";
   }

   /**
    * Vergleicht mit einer Sphäre
    *
    * @param S Sphäre
    * @return die Gleichheit
    */
   public boolean equals (Object S)
   {
      if (S == null) return false;
      if (! (S instanceof Sphere)) return false;

      Sphere S_ = (Sphere)S;
      return rad == S_.rad;
   }


   /**
    * Berechnet den Abstand zweier Vektoren <i>auf</i> der Sphäre.
    * <p>
    * Ist <i>R</i> der Sphären-Radius, und sind <i>a</i>, <i>b</i> != 0 die beiden
    * Vektoren, so berechnet diese Funktion also:
    * <p>
    * <i>R</i> * acos(&lt; <i>a</i>, <i>b</i> &gt; / |<i>a</i>||<i>b</i>|)
    *
    * @param a Vektor
    * @param b Vektor
    * @return den Abstand
    */
   public final double diff (Vector a, Vector b)
   {
      return rad * Math.acos(Vector.dot(a,b)/a.abs()/b.abs());
   }

   /**
    * Erzeugt aus einem gegebenen Vektor eine Orthonormalbasis.
    * <p>
    * Ist <i>v</i> != 0 der gegebene Vektor
    * und (<i>a</i>,<i>b</i>,<i>c</i>) die Orthonormalbasis,
    * so gilt:
    * <ul>
    * <li> <i>c</i> = <i>v</i> / |<i>v</i>|
    * <li> <i>b</i> = <b>z</b> x <i>c</i> &nbsp; [ <b>z</b> = (0,0,1) ]
    * <li> <i>a</i> = <i>b</i> x <i>c</i>
    * </ul>
    *
    * @param v Vektor
    * @return die Basis
    */
   public final Matrix base (Vector v)
   {
      if (v == null || v.zero()) return new Matrix();
      Vector c = Vector.norm(v);
      if (c.x == 0 && c.y == 0 && c.z ==  1) return new Matrix();
      if (c.x == 0 && c.y == 0 && c.z == -1) return new Matrix(-1,0,0, 0,1,0, 0,0,-1);
      Vector b = Vector.vec(new Vector(0,0,1),c).norm();
      Vector a = Vector.vec(b,c).norm();
      return new Matrix(a,b,c);
   }

   /**
    * Erzeugt aus einem gegebenen Vektor und einem Drehwinkel eine Orthonormalbasis.
    * <p>
    * Die Orthonormalbasis (<i>a</i>,<i>b</i>,<i>c</i>) wird zuerst aus dem gegebenen
    * Vektor erzeugt und anschließend für den gegebenen Drehwinkel um ihre
    * <i>c</i>-Achse gedreht.
    *
    * @param v Vektor
    * @param alpha Winkel
    * @return die Basis
    */
   public final Matrix base (Vector v, double alpha)
   {
      return base(v).mulL(Matrix.rotZ(alpha));
   }


   /**
    * Krümmt eine Ebenen-Position auf die Sphäre.
    *
    * @param c komplexe Zahl
    * @return den Sphären-Vektor
    */
   public final Vector warp (Complex c)
   {
      return warp(c,0).c;
   }

   /**
    * Krümmt eine orientierte Ebenen-Position auf die Sphäre.
    * <p>
    * Die Blickwinkel der Ebene sind so parametrisiert, dass Winkel
    * vom Wert 0 den Blick strahlenförmig entlang der Verbindungs-Geraden
    * von Ursprung durch Position nach außen richten.
    *
    * @param c komplexe Zahl
    * @param alpha Blickwinkel
    * @return die Sphären-Matrix
    */
   public final Matrix warp (Complex c, double alpha)
   {
      if (c == null || c.zero()) return Matrix.rotZ(alpha);
      else return euler(new Vector(c.arg(),c.abs()/rad,alpha));
   }

   /**
    * Krümmt eine Sphären-Position auf die Ebene.
    *
    * @param v Sphären-Vektor
    * @return die komplexe Zahl
    */
   public final Complex warp (Vector v)
   {
      if (v == null || v.zero()) return new Complex();
      else {
         Vector w = Vector.norm(v);
         if (w.x == 0 && w.y == 0) return new Complex(rad*Math.PI,0); // w.z = -1!
         else return new Complex(w.x,w.y).norm().mul(rad*Math.acos(w.z));
      }
   }


   /**
    * Krümmt eine orientierte Sphären-Position auf die Ebene.
    * <p>
    * Die Blickwinkel der Ebene sind so parametrisiert, dass Winkel
    * vom Wert 0 den Blick strahlenförmig entlang der Verbindungs-Geraden
    * von Ursprung durch Position nach außen richten.
    * <p>
    * Die beiden Rückgabewerte <i>Position</i> und <i>Blickrichtung</i>
    * werden aus praktischen Gründen in ein Vector-Objekt gepackt und sollten
    * etwa mit folgendem Programm-Code sofort wieder zerlegt werden:
    *<pre>
    *Vector v = Sphere.warp(M);         // "(c,a) = Sphere.warp(M);"
    *&nbsp;
    *Complex c = new Complex(cf.x,v.y);  // c = Position
    *double a = v.z;                    // a = Blickrichtung
    *</pre>
    *
    * @param M Sphären-Matrix
    * @return die komplexe Zahl und den Blickinkel als Vector-Objekt
    */
   public final Vector warp (Matrix M)  // returns (c,alpha) as Vector!
   {
      if (M == null) return new Vector();
      else {
         Vector v = euler(M);
         Complex c = Complex.expi(v.x).mul(v.y*rad);
         return new Vector(c.x,c.y,v.z);
      }
   }


   /**
    * Berechnet die Matrix zu einem Euler-Vektor.
    *
    * @param v Euler-Vektor
    * @return die Matrix
    */
   public final static Matrix euler (Vector v)
   {
      Matrix M = Matrix.rotZ(v.z);
      M = M.mul(Matrix.rotY(v.y)); /*immutable*/
      M = M.mul(Matrix.rotZ(v.x)); /*immutable*/
      return M;
   }

   /**
    * Berechnet den Euler-Vektor zu einer Matrix.
    *
    * @param M Matrix
    * @return den Euler-Vektor
    */
   public final static Vector euler (Matrix M)           // returns (-pi..pi,0..pi,-pi..pi)
   {
      Vector ay = Vector.vec(new Vector(0,0,1),M.c);     // get 2nd axis
      if (ay.abs() == 0) ay = new Vector(0,1,0);
      else ay = ay.norm(); /*immutable*/

      double vy = angle(new Vector(0,0,1),M.c,ay);       // 1st angle
      double vz = angle(ay,M.b,M.c);                     // 2nd angle
      Matrix A = new Matrix(M).mul(Matrix.rot(ay,-vy));  // undo 2nd
      double vx = angle(new Vector(1,0,0),A.a,A.c)-vz;   // 3rd angle
      while (vx  >  Math.PI) vx -= 2*Math.PI;            // fit into -pi..+pi!
      while (vx <= -Math.PI) vx += 2*Math.PI;

      return new Vector(vx,vy,vz);
   }

   private final static double angle (Vector a, Vector b, Vector orientation)  // angle between
   {
      double dot = Vector.dot(a,b); // /a.abs()/b.abs();
      if (dot > 1) dot = 1;
      if (dot <-1) dot =-1;
      if (orientation == null || Vector.dot(Vector.vec(a,b),orientation) > 0)
         return  Math.acos(dot);
      else
         return -Math.acos(dot);
   }
}
