/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.math;


/**
 * Die Klasse Vector repräsentiert einen dreidimensionalen Vektor.
 * <p>
 * Im Folgenden bedeuten: <br>
 * &nbsp; &nbsp; <i>a</i>, <i>b</i>, <i>v</i> &nbsp;-&nbsp; Vektoren<br>
 * &nbsp; &nbsp; <i>x</i>, <i>y</i>, <i>z</i> &nbsp;-&nbsp; reelle Zahlen (als Koordinaten)<br>
 * &nbsp; &nbsp; <i>r</i> &nbsp;-&nbsp; reelle Zahl<br>
 * &nbsp; &nbsp; <i>c</i> &nbsp;-&nbsp; komplexe Zahl<br>
 * &nbsp; &nbsp; <i>A</i> &nbsp;-&nbsp; Matrix<br>
 * &nbsp; &nbsp; <b>v</b> &nbsp;-&nbsp; der Vektor des Vector-Objekts selbst
 * <p>
 * Operator Overloading für <i>Groovy</i>, <i>Scala</i>, <i>Kotlin</i> und <i>Jython</i> wird unterstützt.
 */
public final class Vector
{
   /**
    * <i>x</i>-coordinate
    */
   public final double x;

   /**
    * <i>y</i>-coordinate
    */
   public final double y;

   /**
    * <i>z</i>-coordinate
    */
   public final double z;


   // ---- constructors & bare bones ------------------------------------------

   /**
    * Erzeugt einen Vektor mit den Koordinaten (0,0,0).
    */
   public Vector ()                             { this.x = 0;
                                                  this.y = 0;
                                                  this.z = 0; }

   /**
    * Erzeugt einen Vektor mit den Koordinaten (<i>x</i>,<i>y</i>,<i>z</i>).
    *
    * @param x <i>x</i>-Koordinate
    * @param y <i>y</i>-Koordinate
    * @param z <i>z</i>-Koordinate
    */
   public Vector (double x, double y, double z) { this.x = x;
                                                  this.y = y;
                                                  this.z = z; }

   /**
    * Erzeugt einen Vektor gleich einem anderen Vektor.
    *
    * @param v zu kopierender Vektor
    */
   public Vector (Vector v)            { this.x =(v==null)? 0 : v.x;
                                         this.y =(v==null)? 0 : v.y;
                                         this.z =(v==null)? 0 : v.z; }

   /**
    * Erzeugt einen Vektor aus einer komplexen Zahl.
    * <p>
    * Die <i>z</i>-Koordinate des Vektors wird dabei auf null gesetzt.
    *
    * @param c komplexe Zahl
    */
   public Vector (Complex c)           { this.x = (c == null)? 0 : c.x;
                                         this.y = (c == null)? 0 : c.y;
                                         this.z = 0; }

   /**
    * Berechnet den Hash-Code
    * &nbsp;[
    * = <i><code>####</code></i>
    * ].
    *
    * @return den Hash-Code
    */
   public int hashCode ()
   {
      return new Double(x+y+z).hashCode();
   }


   /**
    * String-Repräsentation
    * &nbsp;[
    * = "..."
    * ].
    *
    * @return den String
    */
   public String toString ()
   {
      return "("+Complex.format(x)+","+Complex.format(y)+","+Complex.format(z)+")";
   }

   /**
    * Vergleicht mit einem Vektor
    * &nbsp;[
    * <b>v</b> = <i>v</i> ?
    * ].
    *
    * @param v Vektor
    * @return die Gleichheit
    */
   public boolean equals (Object v)
   {
      if (v == null) return false;
      if (! (v instanceof Vector)) return false;

      Vector v_ = (Vector)v;
      return x == v_.x && y == v_.y && z == v_.z;
   }

   /**
    * Vergleicht mit einem Vektor
    * &nbsp;[
    * <b>v</b> = (<i>x</i>,<i>y</i>,<i>z</i>) ?
    * ].
    *
    * @param x reelle Zahl
    * @param y reelle Zahl
    * @param z reelle Zahl
    * @return die Gleichheit
    */
   public boolean equals (double x, double y, double z)
   {
      return (this.x == x && this.y == y && this.z == z);
   }


   // ---- object methods -----------------------------------------------------

   /**
    * Addiert einen Vektor
    * &nbsp;[
    * <b>v</b> = <b>v</b> + <i>b</i>
    * ].
    *
    * @param b Vektor
    * @return <b>this</b>
    */
   public final Vector add (Vector b) { return new Vector(x+b.x,y+b.y,z+b.z); }

   /**
    * Subtrahiert einen Vektor
    * &nbsp;[
    * <b>v</b> = <b>v</b> - <i>b</i>
    * ].
    *
    * @param b Vektor
    * @return <b>this</b>
    */
   public final Vector sub (Vector b) { return new Vector(x-b.x,y-b.y,z-b.z); }

   /**
    * Multipliziert mit einer reellen Zahl
    * &nbsp;[
    * <b>v</b> = <b>v</b> <i>r</i>
    * ].
    *
    * @param r reelle Zahl
    * @return <b>this</b>
    */
   public final Vector mul (double r)  { return new Vector(x*r,y*r,z*r); }

   /**
    * Multipliziert mit einer Matrix
    * &nbsp;[
    * <b>v</b> = <b>v</b> <i>A</i>
    * ].
    *
    * @param A Matrix
    * @return <b>this</b>
    */
   public final Vector mul (Matrix A)
   {
      return new Vector(x*A.a.x + y*A.b.x + z*A.c.x,   // !
                        x*A.a.y + y*A.b.y + z*A.c.y,
                        x*A.a.z + y*A.b.z + z*A.c.z);
   }

   /**
    * Dividiert durch eine reelle Zahl
    * &nbsp;[
    * <b>v</b> = <b>v</b> / <i>r</i>
    * ].
    *
    * @param r reelle Zahl
    * @return <b>this</b>
    */
   public final Vector div (double r)  { return new Vector(x/r,y/r,z/r); }

   /**
    * Bildet den Gegenvektor
    * &nbsp;[
    * <b>v</b> = -<b>v</b>
    * ].
    *
    * @return <b>this</b>
    */
   public final Vector neg ()  { return new Vector(-x,-y,-z); }

   /**
    * Testet auf null
    * &nbsp;[
    * <b>v</b> = 0 ?
    * ].
    *
    * @return <b>true</b>, falls dieser Vektor gleich (0,0,0)
    */
   public final boolean zero () { return abs()==0; }

   /**
    * Berechnet den Absolutbetrag
    * &nbsp;[
    * = | <b>v</b> |
    * ].
    *
    * @return den Absolutbetrag
    */
   public final double abs ()   { return Math.sqrt(x*x+y*y+z*z); }

   /**
    * Berechnet den Absolutbetrag im Quadrat
    * &nbsp;[
    * = | <b>v</b> |^2
    * ].
    *
    * @return den Absolutbetrag im Quadrat
    */
   public final double abs2 ()  { return x*x+y*y+z*z; }

   /**
    * Normiert den Absolutbetrag auf eins
    * &nbsp;[
    * <b>v</b> = <b>v</b> / | <b>v</b> |
    * ].
    *
    * @return <b>this</b>
    */
   public final Vector norm () { double r=abs(); return new Vector(x/r,y/r,z/r); }

   /*D
    * Normiert den Absolutbetrag auf eine reelle Zahl
    * &nbsp;[
    * <b>v</b> = r * <b>v</b> / | <b>v</b> |
    * ].
    *
    * @param r reelle Zahl
    * @return <b>this</b>
   /*E
    * Normalizes the absolute value to a real number
    * &nbsp;[
    * <b>v</b> = r * <b>v</b> / | <b>v</b> |
    * ].
    *
    * @param r real number
    * @return <b>this</b>
   public final Vector norm (double r) { double s=abs(); s = r/s; x*=s; y*=s; z*=s; return this;}
    */

   // ---- class methods ------------------------------------------------------

   /**
    * Addiert zwei Vektoren
    * &nbsp;[
    * = <i>a</i> + <i>b</i>
    * ].
    *
    * @param a Vektor
    * @param b Vektor
    * @return die Summe
    */
   public final static Vector add (Vector a, Vector b)
   {
      return new Vector(a.x+b.x,a.y+b.y,a.z+b.z);
   }

   /**
    * Subtrahiert zwei Vektoren
    * &nbsp;[
    * = <i>a</i> - <i>b</i>
    * ].
    *
    * @param a Vektor
    * @param b Vektor
    * @return die Differenz
    */
   public final static Vector sub (Vector a, Vector b)
   {
      return new Vector(a.x-b.x,a.y-b.y,a.z-b.z);
   }

   /**
    * Berechnet die Gegenzahl
    * &nbsp;[
    * = -<i>v</i>
    * ].
    *
    * @param v Vektor
    * @return die Gegenzahl
    */
   public final static Vector neg (Vector v)
   {
      return new Vector(-v.x,-v.y,-v.z);
   }

   /**
    * Berechnet die Normierung des Absolutbetrages auf eins
    * &nbsp;[
    * = <i>v</i> / | <i>v</i> |
    * ].
    *
    * @param v Vektor
    * @return die Normierung
    */
   public final static Vector norm (Vector v)
   {
      return new Vector(v).norm();
   }

   /*
    * Berechnet die Normierung des Absolutbetrages auf eine reelle Zahl
    * &nbsp;[
    * = r * <i>v</i> / |<i>v</i>|
    * ].
    *
    * @param v Vektor
    * @param r reelle Zahl
    * @return die Normierung
   public final static Vector norm (Vector v, double r)
   {
      return new Vector(v).norm(r);
   }
    */

   /**
    * Multipliziert einen Vektor mit einer reellen Zahl
    * &nbsp;[
    * = <i>v r</i>
    * ].
    *
    * @param v Vektor
    * @param r reelle Zahl
    * @return das Produkt
    */
   public final static Vector mul (Vector v, double r)  // real multiplication
   {
      return new Vector(v.x*r,v.y*r,v.z*r);
   }

   /**
    * Multipliziert einen Vektor mit einer Matrix
    * &nbsp;[
    * = <i>v</i> <i>A</i>
    * ].
    *
    * @param v Vektor
    * @param A Matrix
    * @return das Produkt
    */
   public final static Vector mul (Vector v, Matrix A)  // Matrix multiplication [***anders als Mathe!***]
   {
      return new Vector(v).mul(A);
   }

   /**
    * Berechnet das Skalarprodukt zweier Vektoren
    * &nbsp;[
    * = &lt; <i>a</i>, <i>b</i> &gt;
    * ].
    *
    * @param a Vektor
    * @param b Vektor
    * @return das Skalarprodukt
    */
   public final static double dot (Vector a, Vector b)  // dot product
   {
      return a.x*b.x+a.y*b.y+a.z*b.z;
   }

   /**
    * Berechnet das dreidimensionale Vektorprodukt zweier Vektoren
    * &nbsp;[
    * = <i>a</i> x <i>b</i>
    * ].
    *
    * @param a Vektor
    * @param b Vektor
    * @return das Vektorprodukt
    */
   public final static Vector vec (Vector a, Vector b)
   {
      return new Vector(a.y*b.z-a.z*b.y,a.z*b.x-a.x*b.z,a.x*b.y-a.y*b.x);
   }

// [oo]
// =========================================================================

/*
 * java-oo
 *
 * [!] '==' / '!=' not overloaded, use c.equals(b) instead
 * [!] '+=' etc.   not overloaded, use c = c + b   instead
 * [!] no unary +
 *
 */

// public final Vector add            (Vector b) { return add(b); }                 // +
   //public final Vector positive       ()         { return this; }
   public final Vector subtract       (Vector b) { return sub(b); }                 // -
   public final Vector negate         ()         { return neg(); }
   public final Vector multiply       (Vector b) { return Vector.vec(this,b); }     // *
   public final Vector multiply       (double r) { return mul(r); }
   public final Vector multiplyRev    (double r) { return mul(r); }
   public final Vector divide         (double r) { return mul(1/r); }               // /

// =========================================================================

/*
 * Groovy
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 *
 */

   public final Vector plus           (Vector b) { return add(b); }                 // +
   public final Vector positive       ()         { return this; }
   public final Vector minus          (Vector b) { return sub(b); }                 // -
   public final Vector negative       ()         { return neg(); }
// public final double multiply       (Vector b) { return Vector.dot(this,b); }     // *
// public final Vector multiply       (double r) { return mul(r); }
   //public final Vector multiplyRev    (double r) { return mul(r); }
// public final Vector divide         (double r) { return mul(1/r); }               // /


// =========================================================================

/*
 * Kotlin
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 * [!] Kotlin does not support automatic type cast from int to double
 *
 */

// public final Vector plus           (Vector b) { return add(b); }                 // +
   public final Vector unaryPlus      ()         { return this; }
// public final Vector minus          (Vector b) { return sub(b); }                 // -
   public final Vector unaryMinus      ()         { return neg(); }
   public final Vector times          (Vector b) { return Vector.vec(this,b); }     // *
   public final Vector times          (double r) { return mul(r); }
   public final Vector times          (int r)    { return mul((double)r); }
   //public final Vector multiplyRev    (double r) { return mul(r); }
// public final Vector div            (double r) { return mul(1/r); }               // /
   public final Vector div            (int r)    { return mul(1/(double)r); }

// =========================================================================

/*
 * Jython
 *
 * [?] no limitiations, everything's fine!
 *
 */

   public final Vector __add__        (Vector b) { return add(b); }                 // +
   public final Vector __pos__        ()         { return this; }
   public final Vector __sub__        (Vector b) { return sub(b); }                 // -
   public final Vector __neg__        ()         { return neg(); }
   public final Vector __mul__        (Vector b) { return Vector.vec(this,b); }     // *
   public final Vector __mul__        (double r) { return mul(r); }
   public final Vector __rmul__       (double r) { return mul(r); }
   public final Vector __div__        (double r) { return mul(1/r); }               // /

// =========================================================================

/*
 * Scala
 *
 * [!] '2*c' etc. not overloaded, always use 'c*2'
 * [!] 'Vector' is a key word in scala, so take care... :-)
 *
 */

   public final Vector $plus          (Vector b) { return add(b); }                 // +
   public final Vector unary_$plus    ()         { return this; }
   public final Vector $minus         (Vector b) { return sub(b); }                 // -
   public final Vector unary_$minus   ()         { return neg(); }
   public final double $times         (Vector b) { return Vector.dot(this,b); }     // *
   public final Vector $times         (double r) { return mul(r); }
   //public final Vector multiplyRev    (double r) { return mul(r); }
   public final Vector $div           (double r) { return mul(1/r); }               // /

// =========================================================================

}
