/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.net;

import java.net.*;

public class Address
{
   public static final char PORT = ':';   // port separator
   public static final char HASH = '#';   // hash separator

   InetAddress host = null;
   int port = 0;
   String hash = ""; // always without '#', never null


   public Address (Address address)
   {
      this(address.host,address.port,address.hash);
   }

   public Address (InetAddress host, int port)
   {
      this.host = host;
      this.port = port & 0xFFFF;
   }

   public Address (InetAddress host, int port, String hash)
   {
      this.host = host;
      this.port = port & 0xFFFF;
      this.hash = (hash == null)? "" : "" + hash;
   }

   public Address (String address)
   {
      if (address.indexOf(HASH)>=0) {
         int hh = address.indexOf(HASH);
         this.hash = address.substring(hh+1).trim();
         address = address.substring(0,hh);
      }
      
      if (address.indexOf(PORT)<0) address += "" + PORT + hoverball.layout.About.DEFAULT_PORT;

      int pt = address.indexOf(PORT);
      String host = address.substring(0,pt).trim().toLowerCase();
      String port = address.substring(pt+1).trim();
      if (host.equals("")) return;
      try {
         if (host.equals("localhost")) this.host = InetAddress.getLocalHost();
         else                          this.host = InetAddress.getByName(host);
         this.port = Integer.parseInt(port);
      } catch (UnknownHostException e) {
         this.host = null;
         this.port = 0;
      } catch (NumberFormatException e) {
         this.host = null;
         this.port = 0;
      }
   }


   // ---- system functions

   public boolean equals (Object a)
   {
      if (a == null) return false;
      if (! (a instanceof Address)) return false;

      Address A = (Address)a;
      return (host == A.host || host.equals(A.host)) &&
              port == A.port &&
              hash.equals(A.hash);
   }

   public int hashCode ()
   {
      return (host == null? 0 : host.hashCode()) + port + hash.hashCode();
   }

   public String toString ()
   {
      return 
         (host == null? "" : getHostName()) + 
         (getPort() == hoverball.layout.About.DEFAULT_PORT? "" : ("" + PORT + getPort())) +
         (getHash().equals("")? "" : HASH + getHash());
   }

   public String getHostName ()
   {
      if (host == null) return null;
      else return host.getHostName(); // host.getHostName() / host.getCanonicalHostName() / host.getHostAddress()
   }

   public InetAddress getHost ()
   {
      return host;
   }

   public int getPort ()
   {
      return port;
   }

   public String getHash ()
   {
      return "" + hash;
   }

   public void setHash (String hash)
   {
      this.hash = "" + hash;
   }

   public boolean isValid ()
   {
      return (host != null);
   }
}
