/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import hoverball.net.Address;

public class CS extends Protocol // Control Client to Server
{
   // Protocol

   public static Serial deserialize (String string)
   {
      Object[] objects = parse(string); if (objects == null) return null;
      Serial serials = new Serial(){};  // Start-Serial (will be discarded.)
      Serial last = serials;

      for (int i=0; i<objects.length; ++i)
      try {
         Object[] subobjects = (Object[])objects[i];
         Serial serial = null;
         if (((serial = CS.Connect.deserialize(subobjects)) != null) ||
             ((serial = CS.Set.deserialize(subobjects)) != null) ||
             ((serial = CS.Channel.deserialize(subobjects)) != null) ||
             ((serial = CS.State.deserialize(subobjects)) != null) ||
             ((serial = View.deserialize(subobjects)) != null) ||
             ((serial = CS.Ping.deserialize(subobjects)) != null) ||
             ((serial = CS.Bye.deserialize(subobjects)) != null)) last = last.next = serial;
         else throw new Serial.UnknownCommand();
      } catch (Exception e) {
////System.out.println("CS.deserialize failed ("+e+"): |"+string+"|");    	  
         return null;
      }
      return serials.next;
   }


   // (connect #hash MyController)

   public static class Connect extends Serial
   {
      public String hash;
      public String name;

      public Object clone ()
      {
         return new Connect(clone(hash),clone(name)).next(clone(next));
      }

      public Connect (String hash, String name)
      {
         this.hash = hash;
         this.name = name;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(connect ");
         buffer.append(putString(Address.HASH + hash));
         buffer.append(' ');
         buffer.append(putString(name));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("connect")) return null;
         if (objects.length != 3) throw new IllegalNumberOfArguments();

         String hash = getString(objects[1]).substring(1);
         String name = getString(objects[2]);
         return new Connect(hash,name);
      }
   }


   // (set parameter_key value)

   // ###: value==null = (set parameter_key)

   public static class Set extends Serial
   {
      public String key;
      public String value;

      public Object clone ()
      {
         return new Set(clone(key),clone(value)).next(clone(next));
      }

      public Set (String key, String value)
      {
         this.key = key;
         this.value = value;
      }

      public Set (String key)
      {
         this.key = key;
         this.value = null;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(set ");
         buffer.append(putString(key));
         if (value != null) {
            buffer.append(' ');
            buffer.append(putString(value));
         }
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("set")) return null;
         if (objects.length > 3) throw new IllegalNumberOfArguments();

         String key = getString(objects[1]);
         String value = (objects.length == 3)? getString(objects[2]) : null;
         return new Set(key,value);
      }
   }

   // (channel 3 2)

   public static class Channel extends Serial
   {
      public int t;
      public int n;

      public Object clone ()
      {
         return new Channel(t,n).next(clone(next));
      }

      public Channel (int t, int n)
      {
         this.t = t;
         this.n = n;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(channel ");
         buffer.append(putInt(t));
         buffer.append(' ');
         buffer.append(putInt(n));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("channel")) return null;
         if (objects.length != 3) throw new IllegalNumberOfArguments();

         int t = getInt(objects[1]);
         int n = getInt(objects[2]);
         return new Channel(t,n);
      }
   }

   // (state 1)

   public static class State extends Serial
   {
      public int state;

      public Object clone ()
      {
         return new State(state).next(clone(next));
      }

      public State (int state)
      {
         this.state = state;
      }

      public void serialize (StringBuffer buffer)
      {
         buffer.append("(state ");
         buffer.append(putInt(state));
         buffer.append(')');
      }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("state")) return null;
         if (objects.length != 2) throw new IllegalNumberOfArguments();

         int state = getInt(objects[1]);
         return new State(state);
      }
   }

   // (view 23.7 [<AbsoluteObject>]* [<Team>]* ...)  => View!


   // (bye)

   public static class Bye extends Serial
   {
      public Object clone () { return new Bye().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(bye)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("bye")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Bye();
      }
   }
   
   // (ping)

   public static class Ping extends Serial
   {
      public Object clone () { return new Ping().next(clone(next)); }

      public void serialize (StringBuffer buffer) { buffer.append("(ping)"); }

      public static Serial deserialize (Object[] objects)
      {
         String keyword = getString(objects[0]); if (!keyword.equals("ping")) return null;
         if (objects.length != 1) throw new IllegalNumberOfArguments();

         return new Ping();
      }
   }
}

