/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.serial;

import java.io.Serializable;
import java.text.*;

import hoverball.math.Complex;


public abstract class Serial
   implements hoverball.net.Clone, Serializable  // clone if smart-networking!
{
   public static final String JOKER = "*";
   public static NumberFormat format = new DecimalFormat("0.###",new DecimalFormatSymbols(java.util.Locale.US));

   public Serial next = null;
   public Serial next (Serial serial) { this.next = serial; return this; }
      // DANGER!  Not send( A.next(B).next(C) ) !!
      // correct:     send( A.next(B.next(C)) );

   // ---- for subclassers

   public Object clone() { return this; }

   public void serialize (StringBuffer buffer) {}  // (*)

   public static Serial deserialize (Object[] objects) { return null; }  // (*)
//      throws ClassCastException, ArrayIndexOutOfBoundsException,
//             NumberFormatException?,
//             IllegalNumberOfArguments, WrongObjectListElement, IllegalArgumentValue
//             UnknownCommand


   // ----

   public static class IllegalNumberOfArguments extends RuntimeException {}
   public static class WrongObjectListElement   extends RuntimeException {}
   public static class IllegalArgumentValue     extends RuntimeException {}
   public static class UnknownCommand           extends RuntimeException {}

   public String serialize ()
   {
      StringBuffer buffer = new StringBuffer();
      Serial serial = this;
      while (serial != null) {
         serial.serialize(buffer);
         serial = serial.next;
      }
      return buffer.toString();
   }

   public String toString () { return serialize(); }


   // ---- static HEX parsing ...

   public static int parseInt (String string)  // extends Integer.parseInt(..) by "0x.." and "$.."
   {
      string = string.trim();
      int radix = 10;
      if      (string.startsWith("0x")) { radix = 16; string = string.substring(2); }
      else if (string.startsWith("$"))  { radix = 16; string = string.substring(1); }

      return Integer.parseInt(string,radix);   // (throws NumberFormatException!)
   }

   // ---- static Clone handling ...

   public static Serial clone (Serial serial)
   {
      if (serial == null) return null;
      else return (Serial)serial.clone();
   }

   public static String clone (String string)
   {
      if (string == null) return null;
      else return ""+string;
   }


   // ---- static String handling ...

   public static String putString (String string)
   {
      if (string==null) return JOKER;  // should not be used!!
      if (string.length() == 0) return "\"\"";

      if (string.indexOf(' ') < 0 &&
          string.indexOf('(') < 0 &&
          string.indexOf(')') < 0 &&
          string.indexOf('\"') < 0 &&
          string.indexOf('\\') < 0) return string;

      StringBuffer buffer = new StringBuffer(string.length()+10);

      for (int i=0; i<string.length(); ++i) {
         char h = string.charAt(i);
         if (h == '\\') buffer.append('\\');
         if (h == '\"') buffer.append('\\');
         buffer.append(h);
      }

      return "\""+buffer.toString()+"\"";
   }

   public static String getString (Object object)
   {
      if (object.equals(JOKER)) return null;
      return (String)object;  // was already corrected by Parser. (faster!)
   }

   public static String putDouble (double x)
   {
      return (Double.isNaN(x))? JOKER : format.format(x);
   }

   public static double getDouble (Object object)
   {
      String string = (String)object;
      if (string.equals(JOKER)) return Double.NaN;
      else return Double.valueOf(string).doubleValue();
   }

   public static String putComplex (Complex c)
   {
      return putDouble(c.x)+" "+putDouble(c.y);
   }

   public static String putInt (int i)
   {
      return ""+i;
   }

   public static int getInt (Object object)
   {
      return parseInt((String)object);  // accepts HEX
   }

   public static String putInt (int i, int joker)
   {
      return (i == joker)? JOKER : ""+i;
   }

   public static int getInt (Object object, int joker)
   {
      String string = (String)object;
      if (string.equals(JOKER)) return joker;
      else return parseInt(string);  // accepts HEX
   }
}

