/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.math.*;


/*
 Hoverball Simulation Kernel (2007.04.02)

 Spezifikation:

 1. Die Disks liegen immer um ein echtes x > 0 auseinander.

 2. Dem Kernel liegt das diskrete Simulationsprinzip mit konstantem Zeitintervall T zugrunde.
    Bei jedem Simulationsschritt werden alle Impulse als konstant angenommen und
    "linear" simuliert. Nach jedem Schritt werden alle Kräfte für die Disks ausgewertet
    und "linear" auf die Impulse für das kommende Zeitintervall T "abgewirkt". Dabei wird das
    interne Zeitraster {0,T,2T,3T,...} nicht von der Wahl der "äußeren" Simulationsintervalle
    beeinflusst. Jede Disk hat ihr eigenes Zeitraster, wobei diese im allgemeinen
    (d.h. falls keine Kollisionswirkungen stattfinden) miteinander übereinstimmen.

 3. Die Aktionen und der Reibungseffekt werden ebenfalls im Zeitraster ausgewertet.

 4. Die Kollisionen sind dicht.

 5. Multi-Kollisionen mit leichtem Medium: Kollidiert eine Unit mit Ball gegen eine Unit,
    findet theoretisch eine unbeschränkt große Kollisionswiederholung des Balls zwischen den
    beiden Units statt, in welcher der Ball in kleinen Schritten den Impuls von der
    ersten Unit mit der zweiten austauscht. Um diese Multi-Kollision abzubrechen, wird
    bei einer Wiederkollision zweier Disks, in welcher die zweite Disk in der Zwischenzeit
    mit einer dritten Disk kollidiert ist, direkt die erste mit der dritten reflektiert.
    Da dies ein starker Eingriff in das Simulationsgeschehen ist, wird eine Multi-Kollision
    nur bei dem problematischen Fall, dass ein leichtes Medium zwischen zwei schweren Disks
    trillert, auf diese Weise abgebrochen.

 6. Symmetrische Kollision: Um trotz diskreter Simulation der Polarisationsfelder das
    zu erwartende symmetrische Kollisionsverhalten zu erzielen, werden die Zeitraster
    zweier Disks bei einer Kollision symmetrisch gespiegelt, damit sich die großen
    Polarisationskräfte in "Kernnähe" entsprechend wieder ausgleichen. (Dabei werden
    die Polarisationen der Disks für die "erste Auswertung danach" gespeichert.)
    Wirken bei ausreichend großer Entfernung der beiden Disks nur noch kleine Polarisations-
    kräfte, werden ihre Zeitraster wieder an ein Standard-Zeitraster angeglichen.

    Dies erzeugt das Problem, dass die Zeitintervalle zwischen den Simulationsschritten
    beliebig klein werden können, und erfordert somit zusätzlich folgende Korrektur:

    Ausdotzen an einer Disk: Damit die großen Polarisationskräfte in "Kernnähe" immer
    wieder ausgeglichen werden können, ist es wichtig, dass zwischen Kollision und
    Wieder-Kollision derselben Disks mindestens zwei Auswertungen stattfinden. Findet
    nur eine Auswertung statt, kommt die Symmetrie ins Ungleichgewicht. Gleichzeitig
    ist dann aber gewiss, dass das Andotzen schon sehr weit fortgeschritten ist.
    Die Distanz zwischen den beiden Disks ist demnach sehr klein. Es kann nun also getrennt
    und geklebt werden (separer und coller). Der viskose Reibungsfaktor für die klebenden
    Pucks wird dann vereinheitlicht.

 Mit Hilfe der beiden in 5 und 6 dargestellten Korrekturen arbeitet der diskrete Simulator
 in allgemeinen Situationen nahezu "erwartungstreu". Es kann zwar zu geringfügig längeren,
 sollte aber niemals zu unendlichen Hängenbleibern kommen.

 7. Der Energieverbrauch für die Motorenkraft ist linear: dE = -const*|F|*dt.
    Der Energieverbrauch für die Polarisation ist linear: dE = -const*|Q|*dt.
    Die Energie wird linear in der Zeit wiederaufgeladen: dE = const*dt.
    Der Aktionsschranken sind proportional an den Energievorrat gekoppelt.

 Noch zu verbessern:
  - Multi-Kollision mit mehr als 3 Pucks
  - Multi-Kollision mit Unit-Ball-Ball
  - Starke Fehler bei grobem Zeitraster und starker Parameterabweichung
  - Viskositätsmodell für Reibung (Translation und Rotation)
*/


class Kernel
{
   // ---- Kernel-Konstanten --------------------------------------------------


   private final double X;     // Strecke   (Skalierung der Kernel-Konstanten)
   private final double M;     // Masse
//   private final double Q;     // Ladung
   private final double T;     // Zeit

   private final double PRECISION;
   private final double EPSILON_IMPULSION;
   private final double EPSILON_COLLISION;
   private final double EPSILON_COULOMB;

   private final double ADAPTER_COULOMB;
   private final double MULTI_COLLISION_TEMPS;
   private final double MULTI_COLLISION_MASSE;
   private final double SEPARER_ECART;
   private final int    SEPARER_ITERATION;
   private final double COLLER_PROPORTION_Y;
   private final double VISCOSITE_EGALITE;


   // ---- Kernel's public ----------------------------------------------------


   public Kernel (Terrain terrain, double precision)
   {
      X = terrain.BALLON_RAYON;                        // Scale Constants
      M = terrain.BALLON_MASSE;
//      Q = terrain.BALLON_CHARGE;
      T = terrain.TEMPS_QUOTIENT;

      PRECISION             =  0.001; // ehemals 0.020
      EPSILON_IMPULSION     =  0.001 *M*X/T;           // Setup Constants
      EPSILON_COLLISION     =  0.001 *X;
      EPSILON_COULOMB       =  0.001 *M*X/T/T;

      ADAPTER_COULOMB       = 50     *M*X/T/T;
      MULTI_COLLISION_TEMPS =  1;
      MULTI_COLLISION_MASSE =  2;                      // medium.m*MCM < d1.m && medium.m*MCM < d2.m !
      SEPARER_ITERATION     = terrain.disques.length;
      SEPARER_ECART         =  0.02  *X;               // oder: ECART * PROP_Y = 0.03 * 0.5
      COLLER_PROPORTION_Y   =  0.8;                    // Verhältnis beim Angleichen der Impulse in y
      VISCOSITE_EGALITE     =  5;                      // Bremsreibung angleichen (Vorwärtsfahr-Error!)

      this.terrain = terrain;
      this.trame = PRECISION / precision;
   }

   public synchronized double simulate (double moment) { return simuler(moment); }


   // ---- Interne Simulation -------------------------------------------------


   private final Terrain terrain;            // das Terrain, das simuliert werden soll
   private final double trame;               // die Dichte des Zeitrasters


   private double simuler (double moment)
   // Simuliert einen Moment
   {
      double progres = 0;       // sum up simulated time!

      while (terrain.simulate && moment > 0)
         synchronized(terrain)  // synchronize: no interaction during one loop!
      {
         // ---- Wie lange passiert nichts? - Bis nächsten trame oder Collision!

         double avance = moment;                                 // avance = T_max
         for (int i=0; i<terrain.disques.length; ++i)
            avance = Math.min(terrain.disques[i].trame,avance);  // avance = prochaine trame

         Collision collision = collision(avance);                // Collision?

         if (collision != null) avance = collision.t;            // avance = prochaine collision

         // ---- Bewegen!

         for (int i=0; i<terrain.disques.length; ++i)            // avancer!
            avancer(terrain.disques[i],avance);

         moment  -= avance;
         progres += avance;

         // ---- Kollision auswerten

         if (collision != null) {                                // Collisions
            Disque d1 = collision.A; Disque D1 = d1;
            Disque d2 = collision.B; Disque D2 = d2;

            if (d1.collision == d2 &&                            // Mulit-Collision:
                d2.collision != d1 &&                            // d2 Medium?
                d2.temps < trame*MULTI_COLLISION_TEMPS &&        // innerhalb dt?
                d2.m * MULTI_COLLISION_MASSE < d1.m &&           // medium sehr leicht?
                d2.m * MULTI_COLLISION_MASSE < d2.collision.m &&
                approchent(d1,d2.collision))                     // nähern sich?
            {
               separer(d1,d2);
               D2 = d2.collision;
            }

            if (d2.collision == d1 &&                            // Multi-Collision:
                d1.collision != d2 &&                            // d1 Medium?
                d1.temps < trame*MULTI_COLLISION_TEMPS &&        // ...
                d1.m * MULTI_COLLISION_MASSE < d2.m &&
                d1.m * MULTI_COLLISION_MASSE < d1.collision.m &&
                approchent(d1.collision,d2))
            {
               separer(d1,d2);
               D1 = d1.collision;
            }

            reflecter(D1,D2);                                    // Reflektieren

            if (d1.collision == d2 &&                            // 1-Repeated-Collision
                d2.collision == d1 &&                            // (zerstört Symmetrische Kollision!)
                d1.temps - d1.trame < trame &&
                d2.temps - d2.trame < trame)
            {
               coller(d1,d2);
               separer(d1,d2);
               d1.trame = trame;
               d2.trame = trame;
            }

            d1.collision = d2;                                   // Speichere Collision-Info
            d2.collision = d1;
            d1.collision_Q = d2.Q;
            d2.collision_Q = d1.Q;
            d1.temps = 0;
            d2.temps = 0;

            terrain.events.addElement(new Event(Event.COLLISION, // Collision in Event-Liste
               terrain.getIndex(d1),
               terrain.getIndex(d2)));

            // Penalty?
            if      (d1.quoi == d1.BALLON && d1.t != 0 && d2.quoi == d2.UNITE) punir((Unite)d2);
            else if (d2.quoi == d2.BALLON && d2.t != 0 && d1.quoi == d1.UNITE) punir((Unite)d1);

            // Goal?
            if      (d1.quoi == d1.BALLON && d1.t == 0 && d2.quoi == d2.BALLON && d2.t != 0) d2.Q = terrain.BALLON_CHARGE;
            else if (d2.quoi == d2.BALLON && d2.t == 0 && d1.quoi == d1.BALLON && d1.t != 0) d1.Q = terrain.BALLON_CHARGE;
         }

         // ---- Actions auswerten, falls im Zeitraster und nicht nach Collision (-> Symmetrische Kollision)

         for (int i=0; i<terrain.disques.length; ++i)
            if (terrain.disques[i].trame == 0 &&
                terrain.disques[i].quoi == Disque.UNITE) actions((Unite)terrain.disques[i]);

         // ---- Disks auswerten, falls im Zeitraster

         for (int i=0; i<terrain.disques.length; ++i)
            if (terrain.disques[i].trame == 0) evaluer(terrain.disques[i]);

      }

      return progres;
   }


   // ---- Auswerten und Wirken -----------------------------------------------

   private static double bounds (double a, double x, double b)
   {
      if(!Double.isNaN(a) && x<a) return a;
      if(!Double.isNaN(b) && x>b) return b;
      return x;
   }

   private static boolean approchent (Disque d1, Disque d2)
   // true, falls sich d1 und d2 annähern
   {
      double d1d2 = angle(d1,d2);
      double d2d1 = angle(d2,d1);
      Complex p1 = Complex.mul(d1.p,Complex.expi(-d1d2));
      Complex p2 = Complex.mul(d2.p,Complex.expi(-d2d1).neg());
      return Complex.sub(p1,p2).x > 0;
   }

   private static double angle (Disque d1, Disque d2)
   // Berechnet den Winkel zur Position von d2 aus der Sicht von d1
   {
      return angle(d1.X,d2.X);
   }

   private static double angle (Matrix X1, Matrix X2)
   // Berechnet den Winkel zur Position von X2 aus der Sicht von X1
   {
      return new Complex(Vector.mul(X2.c,Matrix.inv(X1))).arg();
   }


   private static void force (Disque d, Complex x, Complex F)
   // Setzt eine Kraft an der Disk an
   {
      d.F = d.F.add(F); /*immutable*/
      d.T += Complex.vec(x,F);
   }

   private void tramer (Disque d, double dt)
   // Wirkt die Kräfte einer Disk ab
   {
      d.p = d.p.add(Complex.mul(d.F,dt)); d.F = new Complex(); // modifier impulsion de translation   /*immutable*/ /*immutable*/
      d.L += d.T*dt;                d.T = 0;                    // modifier impulsion de rotation
   }

   private void actions (Unite unite)
   // Wertet die Aktionen aus, berechnet diese neu
   // und übernimmt sie für die Units
   {
      do {
         Action action = null;                        // cherche Action in liste (si Unite)
         for (int i=0; i<terrain.actions.size(); ++i) {
            Action a = (Action)terrain.actions.elementAt(i);
            if (unite.t == a.t && unite.n == a.n) { action = a; break; }
         }
         if (action == null) break;                   // il n'y en a plus? quitter!

         terrain.actions.removeElement(action);

         if (!Double.isNaN(action.polarise)) unite.aQ = action.polarise; // prends les valeurs!
         if (!Double.isNaN(action.gauche))   unite.amL = action.gauche;
         if (!Double.isNaN(action.droite))   unite.amR = action.droite;
         if (action.message != null)         unite.message = action.message;

      } while(true);

      // Aktionen übernehmen und einschränken
      appliquerActions(unite);
   }

   // Energiesystem: Proportionsformel für Aktion und Energiereserve
   //
   private void appliquerActions (Unite u)
   {
      double ratio = u.E/terrain.UNITE_ENERGIE_MAX;
      u.Q  = bounds(ratio*terrain.UNITE_CHARGE_MIN,u.aQ ,ratio*terrain.UNITE_CHARGE_MAX);
      u.mL = bounds(ratio*terrain.UNITE_MOTEUR_MIN,u.amL,ratio*terrain.UNITE_MOTEUR_MAX);
      u.mR = bounds(ratio*terrain.UNITE_MOTEUR_MIN,u.amR,ratio*terrain.UNITE_MOTEUR_MAX);
   }

   private void evaluer (Disque d)
   // Wertet die Polarisation und Bremsreibung aus
   {
      polarisation(d);  // polarisation

      if (d.quoi == Disque.UNITE) {  // moteurs
         Unite u = (Unite)d;
         force(u,u.xmL(),u.smL().mul(u.mL));
         force(u,u.xmR(),u.smR().mul(u.mR));
      }

      ralentir(d);      // ralentir /*(wichtig: hinter force!)*/

      tramer(d,trame);  // tramer

      // trame

      if (d.trame == 0)  // falls nicht angeglichen in polarisation()
         d.trame = trame;
   }

   private void punir (Unite u)  // Penalty ausführen
   {
      u.py = terrain.PENALTY;
      u.E = 0;
   }


   // ---- Phänomen der viskosen Reibung --------------------------------------


   private void ralentir (Disque d)
   // Setzt Bremskräfte an eine Disk an
   {
      double stokes = stokes(d);  // nach dem Modell der Luftreibung (Gesetz von Stokes für Kugeln)
      Complex F = Complex.mul(d.p,stokes).neg(); // Gegenkraft

      if (d.temps < VISCOSITE_EGALITE * trame) { // Andotz? Gemilderte Reibung!
         stokes = Math.min(stokes,stokes(d.collision));  // die kleinere Kraft (also die des Größeren)
         Complex G = Complex.mul(d.p,stokes).neg();      // Gegenkraft_2!

         double d_dc = angle(d,d.collision);
         Complex turn = Complex.expi(-d_dc);
         F = F.mul(turn);                            // Stokes   für y-Komponente; /*immutable*/
         G = G.mul(turn);                            // Stokes_2 für x-Komponente! /*immutable*/
         Complex R = new Complex(G.x,F.y);
         F = R.mul(Complex.conj(turn));
      }
      d.F = d.F.add(F); /*immutable*/

      double newton = newton(d);  // frei nach der Prandtlschen Grenzschicht = 1 und der Viskosität (Newtonsches Reibungsgesetz)
      double T = - d.L * newton;                 // Gegen-Drehmoment
      d.T += T;
   }

   private double stokes (Disque d)
   {
      return 6 * Math.PI * terrain.VISCOSITE * d.r / d.m;
   }

   private double newton (Disque d)
   {
      return 4 * Math.PI * terrain.VISCOSITE / terrain.BOUNDARY / d.m;
   }


   // ---- Phänomen der beschleunigten Bewegung -------------------------------


   private void avancer (Disque d, double dt)
   // Bewegt eine Disk für die Zeit dt
   {
      double rotate    = dt * (d.L/d.J);
      double translate = dt * (d.p.abs()/d.m)/terrain.sphere.rad;

      Vector direction = new Vector(d.p).mul(d.X);
      Vector axis = Vector.vec(d.X.c,direction);  // axis == 0? -> Matrix.rot(...) = 1!

      d.X = d.X.mulL(Matrix.rotZ(rotate)); /*immutable*/
      d.X = d.X.mulR(Matrix.rot(axis,translate)); /*immutable*/
      d.p = d.p.mul(Complex.expi(-rotate)); /*immutable*/

      d.trame -= dt;
      d.temps += dt;
      if(d.quoi == d.UNITE) { // für Units:
         Unite u = (Unite)d;
         // Energie verbrauchen und aufladen
         u.E -= payerCharge(u,dt);
         u.E -= payerMoteur(u,dt);
         if(u.py==0) u.E += dt*terrain.RECHARGER;
         u.E = bounds(0,u.E,terrain.UNITE_ENERGIE_MAX);
         // Penalty herunterzählen
         u.py -= dt;
         u.py = bounds(0,u.py,Double.POSITIVE_INFINITY);
      } else if(d.quoi == d.BALLON && d.t != 0) { // für Team-Bälle
      	// Charge Halflife (nach einem Goal...)
      	d.Q *= Math.exp(dt/terrain.BALLON_HALFLIFE * Math.log(0.5));
      }

      if (d.p.abs() < EPSILON_IMPULSION) d.p = new Complex();  // Ekel-Error(!): ONBs gehen sonst kaputt.  /*immutable*/
      d.X = reparer(d.X);                                      // Zusätzliche Sicherheit: X "reparieren".
   }

   // Energiesystem: Kostenformel für Polarisation und Antrieb
   //
   private double payerCharge (Unite u, double dt)
   {
      return dt * terrain.UNITE_CHARGE_PAYER
         * Math.abs(u.Q);
   }
   private double payerMoteur (Unite u, double dt)
   {
      return dt * terrain.UNITE_MOTEUR_PAYER
         * (Math.abs(u.mL) + Math.abs(u.mR));
   }

   private static Matrix reparer (Matrix X) /*immutable*/
   // Stellt ONB wieder präzise her
   {
      Vector X_c = X.c.norm(); /*immutable*/
      Vector X_b = Vector.vec(X.c,X.a).norm();
      Vector X_a = Vector.vec(X.b,X.c);
      return new Matrix(X_a,X_b,X_c); /*immutable*/
   }


   // ---- Phänomen der Reflexion ---------------------------------------------


   // Suchen der Kollision (dicht!)


   private Collision collision (double t)
   // Schaut, ob bis zum Zeitpunkt t eine Kollision stattfindet
   // und gibt die frühste Kollision zurück
   {
      Collision collision = null;
      double x = t;

      for (int i=0; i<terrain.disques.length; ++i)
         for (int j=i+1; j<terrain.disques.length; j++) {
            Disque d1 = terrain.disques[i];
            Disque d2 = terrain.disques[j];

            double c = collision(d1,d2,x);
            if (c < x) collision = new Collision(d1,d2,x = c);
         }
      return collision;
   }

   private double collision (Disque d1, Disque d2, double t)
   // Schaut, ob zwischen zwei Disks bis zum Zeitpunkt t eine Kollision stattfindet
   // und gibt den genauen Zeitpunkt zurück
   {
      double t1 = d1.p.zero()? t : d1.r * d1.m / d1.p.abs();  // Zeit zum überschreiten
      double t2 = d2.p.zero()? t : d2.r * d2.m / d2.p.abs();  // des eigenen Radius
      double pas = Math.min(t1,t2);                 // Schritt-Zeit

      double R = d1.r+d2.r;
      double a = 0;                                 // Startzeitpunkt des Schritts

      while (a < t) {
         double b = Math.min(a+pas,t);              // Endzeitpunkt des Schritts
         Vector pos1 = position(d1,b);
         Vector pos2 = position(d2,b);

         if (terrain.sphere.diff(pos1,pos2) <= R) { // Collision! Feinuntersuchung...
            double x = a;                           // Collision-Zeitpunkt
            double z = b-a;
            do {
               z /= 2;
       	      pos1 = position(d1,x+z);
     	         pos2 = position(d2,x+z);
               if (terrain.sphere.diff(pos1,pos2) > R) x += z;  // noch dichter...
            } while (z >= EPSILON_COLLISION);
            return x;
         }
         a = b;                                    // Nächster Schritt!
      }
      return t;                                    // keine Collision.
   }

   private Vector position (Disque d, double dt)
   // Gibt die voraussichtliche Position eine Disk zurück unter der Annahme,
   // dass nichts passiert
   {
      double translate = dt * (d.p.abs()/d.m)/terrain.sphere.rad;

      Vector direction = new Vector(d.p).mul(d.X);
      Vector axis = Vector.vec(d.X.c,direction);  // axis == 0? -> Matrix.rot(...) = 1!

      return Vector.mul(d.X.c,Matrix.rot(axis,translate));
   }


   // Reflexion


   private void reflecter (Disque d1, Disque d2)
   // Reflektiert zwei Disks
   {
   /************
     1) Energieerhaltung: p1^2/m1+p2^2/m2+L1^2/J1+L2^2/J2 = p1_neu^2/m1+p2_neu^2/m2+L1_neu^2/J1+L2_neu^2/J2
     2) Impulserhaltung: p1+p2 = p1_neu+p2_neu

     Wir gehen ins Schwerkraftsystem. Die Schwerkraft zweier Massenpunkte bewegt sich
     vs = (p1+p2)/(m1+m2).
     Diese Geschwindigkeit müssen wir also am Anfang der Rechnung abziehen, und danach wieder dazuzählen.
     Im Schwerkraftsystem gilt zusätzlich
     3) Drehimpuls: Delta_L1 = R1 x Delta_p1 ( = r1*Delta_p1_tang)
     4) Drehimpuls: Delta_L2 = R2 x Delta_p2 ( = r2*Delta_p1_tang) [beachte 2)]

     Wir setzen "willkürlich"
     5) Delta_p1_tang/Delta_p1_normal = g (fest), und zwar g so, dass
     6) Delta_p1_normal = -2*p1_normal (volles Abprallen des Pucks)

     Rechnung:
     Setze 2), 3), 4), 5) in 1) ein und löse nach Delta_p1_normal auf. Dann folgt mit
     x:=p1_normal, y:=p1_tang, M:= 1/m1+1/m2

     Delta_p1_normal = -2*(M*x + g*(M*y + L1*r1/J1 + L2*r2/J2)/(M+3*M*g^2) = -2*x   [6)]

     ergibt nun g = 0 oder g = (M*y + L1*r1/J1 + L2*r2/J2)/(3*M*x)

     Also
     Delta_p1_normal = -2*x
     Delta_p1_tang = g*Delta_p1_normal = -2*(M*y + L1*r1/J1 + L2*r2/J2)/(3*M)
     Delta_L1 = r1*Delta_p1_tang.
   *****************/

      double d1d2 = angle(d1,d2);       // Bezugswinkel
      double d2d1 = angle(d2,d1);

      double r1 = d1.r;             // Radius
      double r2 = d2.r;

      double m1 = d1.m;             // Masse & Impuls
      double m2 = d2.m;
      Complex p1 = Complex.mul(d1.p,Complex.expi(-d1d2));
      Complex p2 = Complex.mul(d2.p,Complex.expi(-d2d1).neg());

      double J1 = d1.J;    // Drehmoment & Drehimpuls
      double J2 = d2.J;
      double L1 = d1.L;
      double L2 = d2.L;

      Complex vs = Complex.add(p1,p2).mul(1/(m1+m2));  // Schwerpunktsystem
      Complex p1s = Complex.mul(vs,m1);
      Complex p2s = Complex.mul(vs,m2);

      Complex P1 = Complex.sub(p1,p1s); // Jetzt sind wir im Schwerpunktsystem
      Complex P2 = Complex.sub(p2,p2s);

      double x = P1.x;                  // Rechnen!
      double y = P1.y;

      double M = 1/m1+1/m2;
      double Lr1 = L1*r1/J1;
      double Lr2 = L2*r2/J2;

      double X = M*y+Lr1+Lr2;

      double dx = -2*x;
      double dy = -2*X/(3*M);

      Complex Delta_P = new Complex(dx,dy);
      double Delta_L1 = r1*dy;
      double Delta_L2 = r2*dy; // hier war der Error. Delta_L2 != Delta_L1.

      // Resultat:

      Complex P1_neu = Complex.add(P1,Delta_P);
      Complex P2_neu = Complex.sub(P2,Delta_P);

      double L1_neu = L1 + Delta_L1;
      double L2_neu = L2 + Delta_L2;

      Complex p1_neu = Complex.add(P1_neu,p1s).mul(Complex.expi(d1d2));  // zurück aus dem Schwerkraftsystem!
      Complex p2_neu = Complex.add(P2_neu,p2s).mul(Complex.expi(d2d1).neg());

      // Speichern!

      d1.p = p1_neu;  /*immutable*/
      d2.p = p2_neu;  /*immutable*/

      d1.L = L1_neu;
      d2.L = L2_neu;

      // trame

      if (d1.trame < d2.trame) d1.trame = d2.trame;        // Angleichen des Zeitrasters
      if (d2.trame < d1.trame) d2.trame = d1.trame;        // (Rückwärtsflieg-Error!)

      d1.trame = trame - d1.trame;                         // Symmetrische Kollision
      d2.trame = trame - d2.trame;
   }


   // Minimal-Problem


   private boolean move (Disque d, Matrix move, int iteration)
   // Verschiebt eine Disk um die Drehmatrix move und verschiebt rekursiv alle
   // Disks mit, die sich anschließend überlappen würden.
   {
      if (iteration == SEPARER_ITERATION) return false;  // Absolute Ausnahme-Abbruchbedingung!

      Disque d1 = d;
      Matrix X = Matrix.mul(d1.X,move);

      boolean success = true;
      for (int i=0; i<terrain.disques.length; i++) {
         Disque d2 = terrain.disques[i]; if (d2 == d1) continue;

         if (terrain.sphere.diff(X.c,d2.X.c) < d1.r + d2.r)
            success &= move(d2,move,iteration+1);
      }

      if (success) d1.X = X;
      return success;
   }

   private void separer (Disque d1, Disque d2)
   // Setzt zwei Pucks d1 und d2 um ein SEPARER_ECART-stel ihrer Radien voneinander weg
   {
      double ratio = d2.m / (d1.m+d2.m);

      double diff = terrain.sphere.diff(d1.X.c,d2.X.c)-d1.r-d2.r;
      Vector rotation = Vector.vec(d1.X.c,d2.X.c);
      Matrix move1 = Matrix.rot(rotation,- ratio  *(-diff+SEPARER_ECART)/terrain.sphere.rad);
      Matrix move2 = Matrix.rot(rotation,(1-ratio)*(-diff+SEPARER_ECART)/terrain.sphere.rad);

      move(d1,move1,0);
      move(d2,move2,0);
   }

   private void coller (Disque d1, Disque d2)
   // Hält zwei Pucks d1 und d2 aneinander (gleicht ihre Impulse aus)
   {
      double d1d2 = angle(d1,d2);
      double d2d1 = angle(d2,d1);

      Complex p1 = Complex.mul(d1.p,Complex.expi(-d1d2));
      Complex p2 = Complex.mul(d2.p,Complex.expi(-d2d1).neg());

      Complex v = Complex.add(p1,p2).mul(1/(d1.m+d2.m));       // "mittlere" Geschwindigkeit

      final double theta = COLLER_PROPORTION_Y;

      double p1_neu_x = v.x * d1.m;                            // Angleichen des x-Impulses
      double p2_neu_x = v.x * d2.m;
      double p1_neu_y = theta * p1.y + (1-theta) * v.y * d1.m; // Angleichen des y-Impulses (verzögert)
      double p2_neu_y = theta * p2.y + (1-theta) * v.y * d2.m;

      Complex p1_neu = new Complex(p1_neu_x,p1_neu_y).mul(Complex.expi(d1d2));
      Complex p2_neu = new Complex(p2_neu_x,p2_neu_y).mul(Complex.expi(d2d1).neg());

      // Speichern!

      d1.p = p1_neu;  /*immutable*/
      d2.p = p2_neu;  /*immutable*/
   }


   private static class Collision  // ---- class Collision --------------------
   {
      final Disque A;  // Disk A
      final Disque B;  // Disk B
      final double t;  // Zeitpunkt

      Collision (Disque A, Disque B, double t)
      {
         this.A = A;
         this.B = B;
         this.t = t;
      }
   }


   // ---- Phänomen der Polarisation ------------------------------------------


   private void polarisation (Disque d)
   // Wertet die Anziehungskräfte aus, die auf eine Disk wirken
   {
      Complex force = new Complex();

      Disque d1 = d;
      if (d1.Q != 0)
         for (int i=0; i<terrain.disques.length; ++i) {
            Disque d2 = terrain.disques[i]; if (d1 == d2) continue;
            double d2_Q = d2.Q;
            if (d1.collision == d2 && d1.temps < trame)
               d2_Q = d1.collision_Q;  // apres-collision? collision-polarisation!
            if (d2_Q == 0) continue;

            double d1d2 = angle(d1,d2);
            double d2d1 = angle(d2,d1);

            Complex D = Complex.expi(d1d2).mul(terrain.sphere.diff(d2.X.c,d1.X.c))
               .sub(d1.xQ())
               .add(Complex.mul(d2.xQ(),Complex.expi(d1d2-d2d1).neg()));

            double d_abs = D.abs();  // != 0
            D = D.mul(-coulomb(d1.Q,d2_Q,d_abs)/d_abs); /*immutable*/

            force = force.add(D); /*immutable*/
         }

      if (force.abs() < ADAPTER_COULOMB)           // Angleichen des Zeitrasters,
         d.trame = terrain.disques[0].trame;       // wenn nur kleine Kräfte wirken

      force(d,d.xQ(),force);
   }

   private double coulomb (double q1, double q2, double d)
   // Sphärische Coulomb-Kraft (bricht die unendliche Reihe ab, sobald die Kraft zu gering wird)
   {
      double factor = q1 * q2 / terrain.PI4EPSILON_0;
      while (d <-terrain.sphere.per/2) d += terrain.sphere.per;
      while (d > terrain.sphere.per/2) d -= terrain.sphere.per;

      double x = 0;
      double n = 0;
      double t = 0;
      do {
          t += 1/(d+n*terrain.sphere.per)/(d+n*terrain.sphere.per); ++ n;
          t += 1/(d+n*terrain.sphere.per)/(d+n*terrain.sphere.per); ++ n;
          t += 1/(d+n*terrain.sphere.per)/(d+n*terrain.sphere.per); ++ n;
          x  = 1/(d+n*terrain.sphere.per)/(d+n*terrain.sphere.per); ++ n;
          t += x;
      } while (factor * x >= EPSILON_COULOMB);

      n = 1;
      double s = 0;
      do {
          s += 1/(d-n*terrain.sphere.per)/(d-n*terrain.sphere.per); ++ n;
          s += 1/(d-n*terrain.sphere.per)/(d-n*terrain.sphere.per); ++ n;
          s += 1/(d-n*terrain.sphere.per)/(d-n*terrain.sphere.per); ++ n;
          x  = 1/(d-n*terrain.sphere.per)/(d-n*terrain.sphere.per); ++ n;
          s += x;
      } while (factor * x >= EPSILON_COULOMB);

      return factor * (t-s);
   }
}
