/*
   Hoverball
   Copyright (c) 2002-2021 Stefan Bornhofen, Matthias Bornhofen

   This file is part of Hoverball.

   Hoverball is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   Hoverball is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Hoverball. If not, see <http://www.gnu.org/licenses/>.
 */
package hoverball.simulator;

import hoverball.math.*;
import hoverball.serial.*;


// ---- Terrain ------------------------------------------------ [Terrain] ----


class Terrain  // public: Access by Simulator, <default>: Access by Kernel
{
   final double PI4EPSILON_0;
   final double VISCOSITE;
   final double BOUNDARY;

   final double TEMPS_QUOTIENT;

   final double BALLON_RAYON;
   final double BALLON_MASSE;
   final double BALLON_CHARGE;
   final double BALLON_HALFLIFE;

   final double SCORE_TIMEOUT;
   final double PENALTY;
   final double RECHARGER;

   private final double UNITE_RAYON;
   final double UNITE_CHARGE_MIN;
   final double UNITE_CHARGE_MAX;
   private final double UNITE_CHARGE_POS;
   final double UNITE_CHARGE_PAYER;
   final double UNITE_MOTEUR_MIN;
   final double UNITE_MOTEUR_MAX;
   private final double UNITE_MOTEUR_POS;
   final double UNITE_MOTEUR_PAYER;
   final double UNITE_ENERGIE_MAX;
   private final double UNITE_VISION;
   private final int UNITE_MESSAGE;

   // ---- Knoten

   private static final Matrix[] nodes = new Matrix[] {  // 1 ggü. 4, 2 ggü. 5, 3 ggü. 6
      new Matrix( 0,+1, 0, 0, 0,+1,+1, 0, 0),  // c:+x (a:+y) 0->1
      new Matrix( 0, 0,+1,+1, 0, 0, 0,+1, 0),  // c:+y (a:+z) 1->2
      new Matrix(-1, 0, 0, 0,-1, 0, 0, 0,+1),  // c:+z (a:-x) 2->3
      new Matrix( 0,-1, 0, 0, 0, 1,-1, 0, 0),  // c:-x (a:-y) 3->4
      new Matrix( 0, 0,-1,+1, 0, 0, 0,-1, 0),  // c:-y (a:-z) 4->5
      new Matrix(+1, 0, 0, 0,-1, 0, 0, 0,-1)   // c:-z (a:+x) 5->0
   };

   // ----

   final Sphere sphere;
   Disque[] disques = new Disque[0];          // liste des Disques!

   private final Kernel kernel;               // Simulator Kernel

   private double temps = 0;


   // ---- Constructor --------------------------------------------------------

   public Terrain (Constants constants, Channel[] channels)
   {
      // Constants

      double RAYON = constants.getConstant(Constants.WORLD_RADIUS);
      VISCOSITE = constants.getConstant(Constants.WORLD_VISCOSITY);
      BOUNDARY  = constants.getConstant(Constants.WORLD_BOUNDARY);

      sphere = new Sphere(RAYON);

      PI4EPSILON_0 = 4*Math.PI*constants.getConstant(Constants.WORLD_PERMITTIVITY);

      TEMPS_QUOTIENT = constants.getConstant(Constants.SIMULATOR_TIME);

      BALLON_RAYON = constants.getConstant(Constants.BALL_RADIUS);
      BALLON_MASSE = constants.getConstant(Constants.BALL_MASS);
      BALLON_CHARGE = constants.getConstant(Constants.BALL_CHARGE);
      BALLON_HALFLIFE = constants.getConstant(Constants.BALL_HALFLIFE);

      UNITE_RAYON = constants.getConstant(Constants.UNIT_RADIUS);
      UNITE_CHARGE_MIN = constants.getConstant(Constants.UNIT_CHARGE_MIN);;
      UNITE_CHARGE_MAX = constants.getConstant(Constants.UNIT_CHARGE_MAX);
      UNITE_CHARGE_POS = constants.getConstant(Constants.UNIT_CHARGE_POS);
      UNITE_CHARGE_PAYER = constants.getConstant(Constants.UNIT_CHARGE_COST);
      UNITE_MOTEUR_MIN = constants.getConstant(Constants.UNIT_ENGINE_MIN);
      UNITE_MOTEUR_MAX = constants.getConstant(Constants.UNIT_ENGINE_MAX);
      UNITE_MOTEUR_POS = constants.getConstant(Constants.UNIT_ENGINE_POS);
      UNITE_MOTEUR_PAYER = constants.getConstant(Constants.UNIT_ENGINE_COST);
      UNITE_ENERGIE_MAX = constants.getConstant(Constants.UNIT_ENERGY_MAX);
      UNITE_VISION = constants.getConstant(Constants.UNIT_VISION);
      UNITE_MESSAGE = (int)Math.round(constants.getConstant(Constants.UNIT_MESSAGE));

      SCORE_TIMEOUT = constants.getConstant(Constants.GAME_TIMEOUT);
      PENALTY = constants.getConstant(Constants.GAME_PENALTY);
      RECHARGER = constants.getConstant(Constants.GAME_RECHARGE);

      double UNITE_MASSE = constants.getConstant(Constants.UNIT_MASS);

      int teamballs = (int)constants.getConstant(Constants.GAME_TEAMBALLS);
      int shotballs = (int)constants.getConstant(Constants.GAME_SHOTBALLS);

      // checkin

      java.util.Vector teams = new java.util.Vector();

      for (int k=1; k<=shotballs; ++k)  // "White" Balls...
         checkin(new Ballon(0,k,BALLON_MASSE,BALLON_RAYON,BALLON_CHARGE,null,new Complex(),0));

      for (int i=0; i<channels.length; ++i) {  // Units...
         Channel ch = channels[i];
         if (!ch.active) continue;

         if (!teams.contains(""+ch.t)) {  // New team? Team-Balls...
            teams.addElement(""+ch.t);

            for (int k=1; k<=teamballs; ++k)
               checkin(new Ballon(ch.t,k,BALLON_MASSE,BALLON_RAYON,0,null,new Complex(),0));
         }

         checkin(new Unite(ch.t,ch.n,UNITE_MASSE,UNITE_RAYON,null,new Complex(),0,UNITE_CHARGE_POS,UNITE_MOTEUR_POS,UNITE_ENERGIE_MAX,0));
      }

      // Setup Kernel

      double precision = constants.getConstant(Constants.SIMULATOR_PRECISION);
      kernel = new Kernel(this,precision);
   }

   Disque getDisque (int quoi, int t, int n)
   {
      for (int i=0; i<disques.length; ++i)
         if (disques[i].quoi == quoi &&
             disques[i].t == t &&
             disques[i].n == n) return disques[i];
      return null;
   }

   int getIndex (Disque disque)
   {
      for (int i=0; i<disques.length; ++i)
         if (disques[i] == disque) return i;
      return -1;
   }


   // ---- In & Out -----------------------------------------------------------

   private synchronized boolean checkin (Disque disque)
   {
      if (disque.X == null) {  // null = random
         disque.X = getFreePlace(disque.r);
         if (disque.X == null) return false;
      } else if (!isFreePlace(disque.X,disque.r)) return false;

      Disque[] disques_ = new Disque[disques.length+1];
      System.arraycopy(disques,0,disques_,0,disques.length);
      disques_[disques.length] = disque;
      disques = disques_;

      return true;
   }

   private synchronized Disque checkout (int quoi, int t, int n)
   {
      Disque disque = getDisque(quoi,t,n);
      if (disque == null) return null;

      Disque[] disques_ = new Disque[disques.length-1];
      int j=0;
      for (int i=0; i<disques.length; ++i)
         if (disques[i] != disque) disques_[j++] = disques[i];
      disques = disques_;

      return disque;
   }


   // ---- Methodes transparentes ---------------------------------------------

   java.util.Vector actions = new java.util.Vector();
   java.util.Vector events = new java.util.Vector();

   public synchronized void addAction (Action action)
   {
      if (getDisque(Disque.UNITE,action.t,action.n) == null) return;

      action.polarise = bounds(action.polarise,UNITE_CHARGE_MIN,UNITE_CHARGE_MAX);
      action.gauche = bounds(action.gauche,UNITE_MOTEUR_MIN,UNITE_MOTEUR_MAX);  // NaN rests.
      action.droite = bounds(action.droite,UNITE_MOTEUR_MIN,UNITE_MOTEUR_MAX);
      if (action.message!=null) {
        int l = action.message.length();
        if (l>UNITE_MESSAGE) l = UNITE_MESSAGE;
        action.message = action.message.substring(0,l);
      }
      actions.addElement(action);
   }

   public synchronized Event[] popEvents ()
   {
      Event[] e = new Event[events.size()];
      events.copyInto(e);
      events.removeAllElements();
      return e;
   }

   public synchronized double getTime () { return temps; }

   public synchronized void setTime (double time) { temps = time; }


   // ---- Get DiskRelatives

   public synchronized DiskRelative[] getDiskRelatives (int t, int n)
   {
      // search self info

      Disque me = getDisque(Disque.UNITE,t,n);
      if (me == null) return new DiskRelative[0];  // outside? looks nothing!
      Matrix me_ = Matrix.inv(me.X);

      // pucks: affine transformation, view filter

      java.util.Vector relatives = new java.util.Vector(disques.length);

      for (int k=0; k<disques.length; ++k) {
         Disque i = disques[k];

         Vector euler = null;
         if (me == i)                  // if (me == i) take euler = (0,0,0)
            euler = new Vector();      //   [ erase "(unit ... x=-0 y=-y ...)"
         else {                        // else { if (see(me,...)) take euler = ... }
            Matrix relative = Matrix.mul(i.X,me_);
            if (!see(me,relative.c,i.r)) euler = null;
            else euler = Sphere.euler(relative);
         }

         if (euler == null && i.message == null) continue;  // ENABLE_HEAR_ONLY?

         relatives.addElement(new DiskRelative(
            (i.quoi == Disque.UNITE?   DiskInfo.UNIT :
             i.quoi == Disque.BALLON?  DiskInfo.BALL : DiskInfo.UNKNOWN),i.t,i.n,
            euler,
            i.message  // ENABLE_HEAR_OPPONENTS?
         ));
      }

      // nodes: affine transformation, view filter

      for (int k=0; k<nodes.length; ++k) {
         Matrix relative = Matrix.mul(nodes[k],me_);
         if (!see(me,relative.c,0)) continue;
         Vector euler = Sphere.euler(relative);

         relatives.addElement(new DiskRelative(DiskInfo.NODE,0,k+1,euler,null));
      }

      // copy to array

      DiskRelative[] r = new DiskRelative[relatives.size()];
      relatives.copyInto(r);
      return r;
   }

   private boolean see (Disque me, Vector position, double radius)  // position != me
   {
      if (UNITE_VISION == 2) return true;   // 2 Extremfälle
      if (UNITE_VISION == 0) return false;

      if (sphere.diff(new Vector(0,0,-1),position)<radius) return true;

      Vector pos = new Vector(position);
      Complex pos_ = new Complex(pos);

      if (pos_.zero()) return true;  // sieht den Südpol
      double turn = pos_.arg();
      pos = pos.mul(Matrix.rotZ(-turn)); /*immutable*/  // Drehung der Disk auf die x-z-Ebene -> pos.c.y = 0 -> pos.c.x != 0
      Vector p = pos;

      double vision = UNITE_VISION*Math.PI /2;  // ="phi/2"
      double border = 0;

      if (radius > 0) {  // Berechne zusätzlichen "Rand" der Objekte, falls Radius > 0
         double C = Math.cos(radius/sphere.rad);
         if (C==0) return true;         // Extremfall.
         double vx = (C*C-p.z*p.z)/C/p.x;
         double vy = Math.sqrt(1-(p.z*p.z/C/C)-vx*vx);
         // vz = p.z/c
         border = new Complex(vx,vy).arg();
      }

      return ((turn + border >= -vision) && (turn - border <= vision));
   }


   // ---- Get DiskAbsolutes

   public synchronized DiskAbsolute[] getDiskAbsolutes ()
   {
      DiskAbsolute[] objects = new DiskAbsolute[disques.length];

      for (int k=0; k<objects.length; ++k) {
         Disque i = disques[k];

         objects[k] = new DiskAbsolute(
            (i.quoi == Disque.UNITE?   DiskInfo.UNIT :
             i.quoi == Disque.BALLON?  DiskInfo.BALL : DiskInfo.UNKNOWN),i.t,i.n,
            Sphere.euler(i.X),
            impulsion2euler(i),  // convert impulsion
            i.r,
            i.m,
            i.Q,
            (i.quoi == Disque.UNITE)? ((Unite)i).mL : 0,
            (i.quoi == Disque.UNITE)? ((Unite)i).mR : 0,
            (i.quoi == Disque.UNITE)? ((Unite)i).E : 0,
            (i.quoi == Disque.UNITE)? ((Unite)i).py : 0,
            i.message
         );
      }
      return objects;
   }


   // ---- Set DiskAbsolutes

   public synchronized void setDiskAbsolutes (DiskAbsolute[] disks)
   {
      Disque[] outside = new Disque[disks.length];

      for (int i=0; i<disks.length; ++i) {  // checkout
         DiskAbsolute o = disks[i];
         int quoi = (o.what == DiskInfo.UNIT? Disque.UNITE :
                     o.what == DiskInfo.BALL? Disque.BALLON : Disque.INCONNU);
         Disque disque = checkout(quoi,disks[i].t,disks[i].n);
         if (disque == null) disks[i] = null;
         else outside[i] = disque;
      }

      for (int i=0; i<disks.length; ++i) {  // checkin with modified parameters
         DiskAbsolute o = disks[i];
         Disque disque = outside[i];
         if (disque == null) continue;  // exists?

         Matrix remember = new Matrix(disque.X);
         /*
         Vector x = Sphere.euler(disque.X);
         if (!Double.isNaN(o.x.x)) x.x = o.x.x;
         if (!Double.isNaN(o.x.y)) x.y = o.x.y;
         if (!Double.isNaN(o.x.z)) x.z = o.x.z;
         */
         Vector x_ = Sphere.euler(disque.X);
         Vector x = new Vector(
            !Double.isNaN(o.x.x)? o.x.x : x_.x,
            !Double.isNaN(o.x.y)? o.x.y : x_.y,
            !Double.isNaN(o.x.z)? o.x.z : x_.z
         ); /*immutable*/
         Matrix position = Sphere.euler(x);

      do {
         disque.X = position;
         if (checkin(disque)) break;  // try new position!
         disque.X = remember;
         if (checkin(disque)) break;  // try old position!
         disque.X = null;
         if (checkin(disque)) break;  // try random!!
         // full => rests outside.
      } while(false);

         /*
         Vector euler = impulsion2euler(disque);      // Impulsion-Converter
         if (!Double.isNaN(o.v.x)) euler.x = o.v.x;
         if (!Double.isNaN(o.v.y)) euler.y = o.v.y;
         if (!Double.isNaN(o.v.z)) euler.z = o.v.z;
                                                      // euler2impulsion
         */
         Vector euler_ = impulsion2euler(disque);
         Vector euler = new Vector(
            !Double.isNaN(o.v.x)? o.v.x : euler_.x,
            !Double.isNaN(o.v.y)? o.v.y : euler_.y,
            !Double.isNaN(o.v.z)? o.v.z : euler_.z
         ); /*immutable*/

         disque.p = Complex.expi(euler.x).mul(euler.y*sphere.rad * disque.m); /*immutable*/
         disque.L = euler.z * disque.J;

         if (disque.quoi == Disque.BALLON) {
            Ballon ballon = (Ballon)disque;
            if (!Double.isNaN(o.Q)) ballon.Q = (o.t == 0)? BALLON_CHARGE : 0;
         } else if (disque.quoi == Disque.UNITE) {
            Unite unite = (Unite)disque;
            if (!Double.isNaN(o.Q)) unite.Q  = bounds(o.Q,UNITE_CHARGE_MIN,UNITE_CHARGE_MAX);
            if (!Double.isNaN(o.L)) unite.mL = bounds(o.L,UNITE_MOTEUR_MIN,UNITE_MOTEUR_MAX);
            if (!Double.isNaN(o.R)) unite.mR = bounds(o.R,UNITE_MOTEUR_MIN,UNITE_MOTEUR_MAX);
            if (!Double.isNaN(o.E)) unite.E  = bounds(o.E,0,UNITE_ENERGIE_MAX);
            if (!Double.isNaN(o.py))unite.py = bounds(o.py,0,PENALTY);
         }
      }
   }

   private Vector impulsion2euler (Disque disque)
   {                                                                    // Impulsion-Converter:
      double y = disque.p.abs()/sphere.rad / disque.m;      // Impuls-Länge
      double x = disque.p.arg();                            // Impuls-Winkel
      double z = disque.L / disque.J;                       // Drehimpuls
      return new Vector(x,y,z);
   }


   // ---- Methodes privées ---------------------------------------------------

   private static double bounds (double value, double min, double max)
   {
      if (value < min) return min;
      if (value > max) return max;
      return value;
   }

   private static Vector random (double range)
   {
      return new Vector((1-2*Math.random())*range,(1-2*Math.random())*range,(1-2*Math.random())*range);
   }

   private Matrix getFreePlace (double r)
   {
      final int LOOP = 100;

      Matrix place = null;

      for (int i=0; i<LOOP; ++i) {
         place = Sphere.euler(Terrain.random(Math.PI));
         if (isFreePlace(place,r)) return place;
      }
      return null;
   }

   private boolean isFreePlace (Matrix place, double r)
   {
      for (int i=0; i<disques.length; i++) {
         double d = sphere.diff(disques[i].X.c,place.c);
         if (d <= disques[i].r+r) return false;
      }
      return true;
   }


   // ---- simulate!

   boolean simulate = false;  // Simulation flag ("false" interrupts Kernel)

   public void simulate (double newtime)  // Interaction-Threads must be synchronized
   {                                      // in Kernel.simulate()!!
      if (newtime <= temps) return;             // straight forward only!

      simulate = true;
      temps += kernel.simulate(newtime-temps);  // synchronizes access to "this"
      simulate = false;
   }

   public void interrupt ()
   {
      simulate = false;
   }
}
