/**
 * 
 */
package hoverball.team.TribalQueens_;

import hoverball.math.Complex;
import hoverball.math.Vector;

/**
 * Un role est une manière de jouer
 * 
 */
public abstract class Role {

	/**
	 * Le joueur dont on est l'état
	 */
	protected TribalQueen joueur;

	/**
	 * Le terrain dans lequel on evolue
	 */
	protected Environnement terrain;

	/**
	 * Le puissance du moteur de gauche
	 */
	protected double moteurGauche;

	/**
	 * Le puissance du moteur de droite
	 */
	protected double moteurDroit;

	/**
	 * Le niveau d'attraction de la balle
	 */
	protected double polarisation;

	/**
	 * La distance à partir de laquelle on attire la balle
	 */
	protected double distanceAttraction = 20;

	/**
	 * La distance à partir de laquelle on concidère possèder la balle
	 */
	protected double distancePossession = 5;

	/**
	 * La distance à partir de laquelle on concidère que quelquechose est devant
	 * ET dangereuse
	 */
	protected double distanceDevant = 20;

	/**
	 * La distance en dessous de laquelle doit etre ce qu'on veut eviter
	 */
	protected double distanceEvitementMin;

	/**
	 * L'angle a partir duquel on concidère qu'on se dirige vers la cible
	 */
	protected double angleMouvement = Math.PI / 8;

	/**
	 * L'angle à partir du quel on attire la balle
	 */
	protected double angleAttraction = Math.PI;

	/**
	 * L'angle à partir du quel on concidère qu'on possède la balle
	 */
	protected double anglePossession = Math.PI / 2;

	/**
	 * L'angle à partir duquel on concidère que quelquechose est devant ET
	 * dangereuse
	 */
	protected double angleDevant = Math.PI / 2;

	/**
	 * L'angle avec lequel on considère qu'on evite quelquechose
	 */
	protected double angleEvitement = Math.PI / 2;

	/**
	 * L'energie qu'il doit rester après un tir
	 */
	protected double energiePourTir;
	
	/**
	 * Distance minimale en deça de laquelle un éléments est jugé trop proche d'une trajectoire.
	 */
	protected double distanceTrajectoireMinimum = 15;
	
	/**
	 * Vrai si le role est en situation critique, il ne faut alors surtout pas changer de role
	 */
	protected boolean critique;
	
	/**
	 * @param joueur
	 * @param terrain
	 */
	public Role(TribalQueen joueur, Environnement terrain) {
		super();
		this.joueur = joueur;
		this.terrain = terrain;
		energiePourTir =  0.1;
		critique = false;
	}

	/**
	 * @param cible
	 *            Le point ciblé
	 * @return La distance qui sépare le joueur de la cible
	 */
	protected double distance(Vector cible) {
		return (terrain.distance(cible, joueur.getJoueur()));
	}

	/**
	 * Retourne l'angle sous lequel est vu un point du monde compris entre -Pi
	 * et Pi. 0 => en face -Pi/2 => à droite PI/2 => à gauche Pi ou -Pi =>
	 * derrière
	 * 
	 * @param cible
	 *            Le point ciblé
	 * @return L'angle qu'il y a entre la cible et le joueur (entre -Pi et PI)
	 */
	protected double angle(Vector cible) {
		return (terrain.angle(cible));
	}

	/**
	 * @return the moteurGauche
	 */
	public double getMoteurGauche() {
		return moteurGauche;
	}

	/**
	 * @return the moteurDroit
	 */
	public double getMoteurDroit() {
		return moteurDroit;
	}

	/**
	 * @return the polarisation
	 */
	public double getPolarisation() {
		return polarisation;
	}

	/**
	 * @return vrai si on possède la balle
	 */
	protected boolean possedeBalle() {
		return (distance(joueur.getBalle()) < distancePossession && Math
				.abs(angle(joueur.getBalle())) < anglePossession);
	}

	/**
	 * Permet de tourner d'un certain angle
	 * 
	 * @param angle
	 *            L'angle duquel on veut tourner
	 */
	protected void tourner(double angle) {
		// si on tourne vers le but advers alors qu'il est proche et qu'on
		// possède la balle
		if (possedeBalle()
				&& Math.signum(angle) == Math.signum(angle(joueur
						.getButAdverse()))
				&& distance(joueur.getButAdverse()) < distancePossession) {
			// alors on tourne à l'opposé
			angle *= -1;
		}
		// on calcule la puissance du moteur de droite
		// plus on est proche de PI/4 plus on doit etre precis
		moteurDroit = joueur.getPuissanceMax() * (angle / (Math.PI / 6));
		// le moteur gauche est à l'opposé du moteur droit
		moteurGauche = -moteurDroit;

	}

	/**
	 * Permet d'avancer vers un point situé à une certaine distance
	 * 
	 * @param distance
	 *            La distance au point qu'on veut atteindre
	 */
	protected void avancer(double distance) {
		// on calcule la puissance du moteur de droite
		moteurDroit = joueur.getPuissanceMax() * (distance / 20);
		// le moteur gauche est à la même puissance que le moteur droit
		moteurGauche = moteurDroit;
	}

	/**
	 * Permet d'attirer quelquechose situé à une certaine distance
	 * 
	 * @param distance
	 *            La distance où est situé la balle
	 */
	protected void attraper(double distance) {
		double energie;

		// s'il n'y a pas de danger pour les buts
		if (!(Math
				.abs(angle(joueur.getButAdverse()) - angle(joueur.getBalle())) < angleMouvement && distance(joueur
				.getButAdverse()) < distance)) {
			// on attire la balle avec une puissance proportionnelle à la
			// distance
			if (!possedeBalle()) {
				energie = joueur.getChargeMin() * 0.02;	
			} else {
				energie = joueur.getChargeMin() * 0.015;
			}
			// si on a assez d'energie
			//if (energieSuffisante(energie)) {
				// alors on polarise
				polarisation = energie;
			//} else {
			//	polarisation = 0;
			//}
		} else {
			// si on arrive ici c'est que les buts sont en danger, on degage
			// donc
			envoyer(100);
		}
	}

	/**
	 * Permet d'envoyer quelquechose à une certaine distance
	 * 
	 * @param distance
	 *            La distance où est situé la cible
	 */
	protected void envoyer(double distance) {
		double energie;

		// si l'angle avec le but est sans risque et que le but est assez loin
		if (Math.abs(angle(joueur.getButAdverse()) - angle(joueur.getBalle())) > angleMouvement
				|| distance(joueur.getButAdverse()) > distance) {
			// on repousse la balle avec une puissance proportionnelle à la
			// distance
			energie = joueur.getPuissanceMax() * (distance / 12500);
			//energie = joueur.getPuissanceMax() * ( (25 * Math.log(distance) + 33) / 1000);
		//	System.out.println("Envoie : "+energie);
			// si on a assez d'energie
			//if (energieSuffisante(energie)) {
				// on polarise
				polarisation = energie;
			//} else {
			//	polarisation = 0;
			//}
		}
	}

	/**
	 * @param energie
	 *            L'energie requise
	 * @return Vrai si on possède l'energie suffisante
	 */
	protected boolean energieSuffisante(double energie) {
		return ((joueur.getEnergy() - energiePourTir)
				- (joueur.getCoutCharge() * Math.abs(energie) * joueur.getFrequence()) > 0);
	}

	protected void allerA(Vector point, boolean force) {
		allerA(point, force, angleMouvement, false);
	}

	/**
	 * Permet d'aller à un point donnée, en force... ou pas
	 * 
	 * @param point
	 *            Le point de destination
	 * @param force
	 *            Vrai si on veut jouer aux bourrins
	 * @param anglePrecision
	 *            l'angle minimal pour avancer
	 */
	protected void allerA(Vector point, boolean force, double anglePrecision,
			boolean eviterJoueurs) {

		//joueur.tag(point, "Cible");
		
		// on recupère l'angle du point
		double angleCible = angle(point);
		// et sa distance
		double distanceCible = distance(point);

		// si on doit etre fin
		if (!force) {
			// on calcul l'angle d'evitement
			angleCible = angleEvitement(angleCible, eviterJoueurs);
		}

		// si l'angle est suffisament petit
		if (Math.abs(angleCible) < anglePrecision) {
			// on avance vers la cible
			avancer(distanceCible);
			moteurDroit += joueur.getPuissanceMax() * (angleCible / (Math.PI/6));
			moteurGauche -= joueur.getPuissanceMax() * (angleCible / (Math.PI/6));
		} else {
			// sinon on tourne vers la cible
			tourner(angleCible);
		}

	}

	/**
	 * @param angleDepart
	 *            l'angle où on veut aller
	 * @return
	 */
	private double angleEvitement(double angleDepart, boolean eviterJoueurs) {
		double angle = angleDepart;
		Vector[] adversaires = joueur.getAdversaires();
		Vector[] equipiers = joueur.getEquipiers();
		distanceEvitementMin = distanceDevant-5;

		// si on veut eviter les joueurs
		if (eviterJoueurs) {
			
			// pour tous les joueurs
			for (int i = 0; i < equipiers.length; i++) {
				// on les evites si c 'est pas moi
				if (!equipiers[i].equals(joueur.getJoueur())) {
					angle = angleEvitementPoint(angle, equipiers[i], true);
				}
				angle = angleEvitementPoint(angle, adversaires[i], true);
			}
		}

		// si on a un but devant nous on l'evite
		angle = angleEvitementPoint(angle, joueur.getButAdverse(), false);

		// sauf si c'est notre but et qu'on ne possède pas la balle
		if (!possedeBalle()) {
			angle = angleEvitementPoint(angle, joueur.getButEquipe(), false);
		}

		// on retourne l'angle
		return angle;
	}

	/**
	 * Permet d'eviter un point
	 * 
	 * @param angleDepart
	 *            La direction vers laquelle on va
	 * @param point
	 *            Le point à eviter
	 * @return L'angle de direction à prendre
	 */
	private double angleEvitementPoint(double angleDepart, Vector point, boolean mobile) {
		double angle = angleDepart;
		double angleTmp;

		// on recupère l'angle du point à eviter
		angleTmp = angle(point);
		

		// si on doit eviter le point
		if (distance(point) < distanceEvitementMin //&& devant(point)
				&& (Math.abs(angleDepart - angleTmp) < angleEvitement || (mobile && Math.abs(angleTmp) < angleEvitement))) {
			if(!mobile) {
			// on doit finir de l'eviter
			critique = true;
			}
			distanceEvitementMin = distance(point);
			// si l'obstacle n'est pas mobile et et que la cible est à droite 
			// ou si l'angle de l'obstacle est négatif
			if ((!mobile && angleDepart - angleTmp > 0)|| angleTmp < 0) {
				// on evite avec en ajoutant l'angle d'evitement (par la droite)
				angle = angleTmp + angleEvitement;
			} else {
				// sinon on evite avec en enlevant l'angle d'evitement (par la gauche)
				angle = angleTmp - angleEvitement;
			}
			joueur.tag(point, "Eviter");
			//System.out.println(joueur.title+" evite qqch devant a un angle de "+angleTmp+" donc nouvel angle : "+angle);
		}

		return angle;
	}

	/**
	 * Un point est devant
	 * 
	 * @param point
	 *            Le point qu'on regarde
	 * @return Vrai si le point est devant
	 */
	protected boolean devant(Vector point) {
		// le point est devant si son angle est inferieur à PI/2
		return Math.abs(angle(point)) < angleDevant
				&& distance(point) < distanceDevant;
	}

	/**
	 * Reflechir au actions à réaliser
	 */
	public abstract void reflechir();
	
	/**
	 * Renvoie true si aucun obstacle n'est trop proche des deux points
	 * @param point1
	 * @param point2
	 * @return boolean
	 */
	public boolean trajectoireOk(Vector point1, Vector point2) {
		Vector[] adversaires = this.joueur.getAdversaires();
		Vector[] equipiers = this.joueur.getEquipiers();
		Vector butEquipe = this.joueur.getButEquipe();
		Vector butAdverse = this.joueur.getButAdverse();
		
		//on charge tous les points pouvant poser problème
		Vector[] points = new Vector[adversaires.length+equipiers.length+2];
		points[0] = butEquipe;
		points[1] = butAdverse;
		int i = 2;
		for (int j = 0 ; j < adversaires.length ; j ++) {
			points[i] = adversaires[j];
			i++;
		}
		for (int j = 0 ; j < equipiers.length ; j ++) {
			points[i] = equipiers[j];
			i++;
		}
		
		//on calcul les coordonnées de la droite
		Complex c1 = this.terrain.warp(point1);
		Complex c2 = this.terrain.warp(point2);
		Complex c;
		double a = 1000000;
		if ((c2.x - c1.x) != 0) {
			a = (c2.y - c1.y) / (c2.x - c1.x);
		}
		double b = c1.y - (a * c1.x);
		
		//on regarde si la distance de ces points à la droite est acceptable
		for (i = 0; i < points.length; i++) {
			c = this.terrain.warp(points[i]);
			//on vérifie qu'on ne check pas un des points de la droite
			if (!(c.equals(c1) || c.equals(c2))) {
				//on vérifie que le points peut être concerné
				if ( ( (c.x < c1.x && c.x > c2.x) || (c.x > c1.x && c.x < c2.x) ) ||
					( (c.y < c1.y && c.y > c2.y) || (c.y > c1.y && c.y < c2.y) ) ) {
					//distance du point à la droite
					double distance = Math.abs((a*c.x+c.y + b)/Math.sqrt(a*a + 1));
					if (distance  < this.distanceTrajectoireMinimum ) {
						//System.out.println(this.joueur.title+" : un obstacle est trop proche "+distance);
						return false;
					}
				}
				
			}
		}
		return true;
	}

	/**
	 * @return the critique
	 */
	public boolean isCritique() {
		return critique;
	}
	
}
